//
// Created by koerberm on 09.03.18.
//

#include <random>
#include <bitset>
#include <thread>
#include "ep/operators.h"
#include "ep/cpu_operators.h"
#include "ep/dispatch_operators.h"
#include "ep/hsa_operators.h"
#include "ep/event.h"
#include "util/eval_util.h"
#include "ep/cpu_parallel_operators.h"

#define EATER_MIN_BATCH 1

#define RAND_LOWER 0.0f
#define RAND_UPPER 100.0f

std::random_device rd;  //Will be used to obtain a seed for the random number engine
auto seed = rd();

////////////////////////////////////////////////////////////////
//
// JOIN
//
////////////////////////////////////////////////////////////////

template <class TI, class TO>
void aggregate_test(uint64_t num_events, uint64_t window_size,
               const hsa::hsa_platform &platform, const hsa::hsa_agent &agent,
               std::function<std::unique_ptr<ep::event_sink<TI>>(
                       const std::shared_ptr<ep::event_queue<TI>> &in,
                       const std::shared_ptr<ep::event_queue<TO>> &out,
                       ep::count_window<event> &&window)> create_op,
               std::vector<double> &results, std::vector<uint64_t> &latency) {

    uint32_t queue_size = 1;
    while ( queue_size < num_events )
        queue_size = queue_size << 1;

    auto qi = std::make_shared<ep::event_queue<TI>>(queue_size, platform, agent);
    auto qo = std::make_shared<ep::event_queue<TO>>(queue_size, platform, agent);

    std::mt19937 gen(seed); //Standard mersenne_twister_engine seeded with rd()
    std::uniform_real_distribution<float> dis(RAND_LOWER,RAND_UPPER);

    ep::event_eater<TO> p(qo,EATER_MIN_BATCH,queue_size);

    // Pre-Fill Window
    ep::count_window<TI> window(window_size, agent);

    uint64_t ts = 1;
    for ( ; ts <= window_size; ts++) {
        window.add(create_event(ts,dis,gen));
    }

    for ( ; ts <= window_size+num_events; ts++) {
        qi->add(create_event(ts,dis,gen));
    }
    auto wi = qi->get_write_idx();
    qi->set_write_index(0);

    auto agg = create_op(qi,qo,std::move(window));

    std::this_thread::sleep_for(std::chrono::milliseconds(100));

    auto begin= std::chrono::steady_clock::now();
    qi->set_write_index(wi);
    qi->wait_for_space(queue_size, 1000000000000);
    std::chrono::duration<double> duration =  std::chrono::steady_clock::now() - begin;
    agg->stop();
    std::this_thread::sleep_for(std::chrono::milliseconds(100));
    qo->wait_for_space(queue_size, 6000000000);
    results.push_back(duration.count());
    latency.push_back(agg->get_processing_latency());
    std::cout << ", (" << p.count << "," << p.unique_val << ")" << std::flush;
}

template <class TI, class TO>
void run() {
    hsa::hsa_platform platform;

    for ( auto &agent : platform.get_agents() ) {
        if (agent.features != HSA_AGENT_FEATURE_KERNEL_DISPATCH)
            continue;

        std::cout << "Using agent " << agent.name << std::endl;

        hsa::hsa_executable exe;

        std::vector<std::string> cos = {
                "src/rocm/cl/out/hells_aggregate_normal.co",
                "src/rocm/cl/out/hsa_aggregate_normal.co"
        };

        for (auto &file : cos) {
            auto co = hsa::code_object_from_file(file,agent);
            try {
                exe.load_code_object(co, agent);
            } catch (const hsa::hsa_exception &e) {
                std::cout << "Error loading module (" << file << "): " << e.what() << std::endl;
                return exit(1);
            }
        }

        exe.freeze();


        // Operator create functions

        std::function<std::unique_ptr<ep::event_sink<TI>>(
                const std::shared_ptr<ep::event_queue<TI>> &in,
                const std::shared_ptr<ep::event_queue<TO>> &out,
                ep::count_window<TI> &&window)> create_cpu = [&](const std::shared_ptr<ep::event_queue<TI>> &in,
                                                                 const std::shared_ptr<ep::event_queue<TO>> &out,
                                                                 ep::count_window<TI> &&window) {
            return std::make_unique<ep::aggregate_cpu<TI, TO, float>>(
                    in, out, std::move(window),
                    [](const TI &e) { return agg_init(&e); },
                    [](float &accum, const float &ev) { agg_merge(&accum, ev); },
                    [](const TI &e, const float &accum) { return agg_finish(&e, accum); });
        };

        std::function<std::unique_ptr<ep::event_sink<TI>>(
                const std::shared_ptr<ep::event_queue<TI>> &in,
                const std::shared_ptr<ep::event_queue<TO>> &out,
                ep::count_window<TI> &&window)> create_cpu_mt = [&](const std::shared_ptr<ep::event_queue<TI>> &in,
                                                                    const std::shared_ptr<ep::event_queue<TO>> &out,
                                                                    ep::count_window<TI> &&window) {

            uint16_t num_threads = 4;

            return std::make_unique<ep::parallel_aggregate<TI, TO, float>>(
                    in, out, std::move(window), num_threads,
                    platform, agent,
                    [](const TI &e) { return agg_init(&e); },
                    [](float &accum, const float &ev) { agg_merge(&accum, ev); },
                    [](const TI &e, const float &accum) { return agg_finish(&e, accum); });
        };

        std::function<std::unique_ptr<ep::event_sink<TI>>(
                const std::shared_ptr<ep::event_queue<TI>> &in,
                const std::shared_ptr<ep::event_queue<TO>> &out,
                ep::count_window<TI> &&window)> create_hells = [&](const std::shared_ptr<ep::event_queue<TI>> &in,
                                                                   const std::shared_ptr<ep::event_queue<TO>> &out,
                                                                   ep::count_window<TI> &&window) {

            uint32_t elements_per_item = 32;

            return std::make_unique<ep::hells_aggregate<TI, TO, float>>(
                    in, out, std::move(window), elements_per_item,
                    platform, agent, exe,
                    [](const TI &e, const float &accum) { return agg_finish(&e, accum); });
        };

        std::function<std::unique_ptr<ep::event_sink<TI>>(
                const std::shared_ptr<ep::event_queue<TI>> &in,
                const std::shared_ptr<ep::event_queue<TO>> &out,
                ep::count_window<TI> &&window)> create_hsa = [&](const std::shared_ptr<ep::event_queue<TI>> &in,
                                                                 const std::shared_ptr<ep::event_queue<TO>> &out,
                                                                 ep::count_window<TI> &&window) {

            return std::make_unique<ep::hsa_aggregate<TI, TO, float>>(
                    in, out, std::move(window),
                    platform, agent, exe);
        };

        std::vector<double> cpu;
        std::vector<double> cpu_mt;
        std::vector<double> hells;
        std::vector<double> hsa;

        std::vector<uint64_t> cpu_l;
        std::vector<uint64_t> cpu_mt_l;
        std::vector<uint64_t> hells_l;
        std::vector<uint64_t> hsa_l;

        std::vector<uint32_t> window_sizes = {4096, 4096 << 1, 4096 << 2, 4096 << 3, 4096 << 4, 4096 << 5, 4096 << 6,
                                              4096 << 7, 4096 << 8};
        cpu.clear();
        cpu_l.clear();
        cpu_mt.clear();
        cpu_mt_l.clear();
        hells.clear();
        hells_l.clear();
        hsa.clear();
        hsa_l.clear();

        for (auto window : window_sizes) {
//            uint32_t factor = (*(--window_sizes.end())/window);
//            uint64_t num_events = 5000 * factor;
            uint64_t num_events = 10000;

            std::cout << window << std::flush;
            aggregate_test(num_events, window, platform, agent, create_cpu, cpu, cpu_l );
            aggregate_test(num_events, window, platform, agent, create_cpu_mt, cpu_mt, cpu_mt_l );
            aggregate_test(num_events, window, platform, agent, create_hells, hells, hells_l );
            aggregate_test(num_events, window, platform, agent, create_hsa, hsa, hsa_l);
            std::cout << std::endl;
        }


        std::cout << "Window sizes; " << window_sizes << std::endl;
        std::cout << "CPU     ; " << cpu << std::endl;
        std::cout << "CPU-MT  ; " << cpu_mt << std::endl;
        std::cout << "Dispatch; " << hells << std::endl;
        std::cout << "HSA     ; " << hsa << std::endl;

        std::cout << "Latency CPU   : " << cpu_l << std::endl;
        std::cout << "Latency CPU-MT: " << cpu_mt_l << std::endl;
        std::cout << "Latency HELLS : " << hells_l << std::endl;
        std::cout << "Latency HSA   : " << hsa_l << std::endl;

        std::cout << "==========================================================" << std::endl;
    }
}

int main(int argc, char **argv) {
    (void) argc;
    (void) argv;
    run<event,aggregated_event>();
    return 0;
}
