//
// Created by koerberm on 30.05.18.
//

#include "cpu_operators.h"
#include "ep_common.h"
#include "event.h"
#include "latency/event_latency.h"

namespace ep {

    /////////////////////////////////////////////////////////////////
    //
    // UNARY SINK
    //
    /////////////////////////////////////////////////////////////////

    template<class T>
    event_sink_cpu<T>::event_sink_cpu(const std::shared_ptr<ep::event_queue<T>> &in_queue, uint32_t min_batch_size,
                                      uint32_t max_batch_size) : event_sink<T>(in_queue),
                                                                 base(in_queue->get_base()),
                                                                 mask(in_queue->mask),
                                                                 min_batch_size(min_batch_size),
                                                                 max_batch_size(max_batch_size),
                                                                 last_batch_time(0),
                                                                 latency_accum(0),
                                                                 latency_count(0),
                                                                 next_read(0),
                                                                 stopped(false),
                                                                 processor(&event_sink_cpu<T>::run, this) {
    }

    template<class T>
    event_sink_cpu<T>::~event_sink_cpu() {
        debug_print("%s\n", "Destroying sink cpu.");
        stop();
        debug_print("%s\n", "Finished destroying sink cpu.");
    }

    template<class T>
    void event_sink_cpu<T>::run() {
        while (!stopped) {
            auto w_idx = this->in_queue->wait_for(min_batch_size, 100);
            if (w_idx >= 0) {
                auto end = std::min(w_idx, next_read + max_batch_size);

                process_batch(next_read,end);
                next_read = end;
                this->in_queue->consume_until(end);

                uint64_t now = static_cast<uint64_t >(std::chrono::steady_clock::now().time_since_epoch().count());
                if ( last_batch_time > 0 ) {
                    latency_accum += (now-last_batch_time);
                    latency_count++;
                }
                last_batch_time = now;
            }
        }
    }

    template<class T>
    void event_sink_cpu<T>::process_batch(hsa_signal_value_t begin, hsa_signal_value_t end) {
        for (hsa_signal_value_t i = begin; i < end; i++) {
            process_element(base[i & mask]);
        }
    }


    template<class T>
    uint64_t event_sink_cpu<T>::get_processing_latency() {
        return (latency_count > 0) ? latency_accum/latency_count : 0;
    }

    template<class T>
    void event_sink_cpu<T>::stop() {
        if (stopped)
            return;
        debug_print("%s\n", "Stopping processor thread.");
        stopped = true;
        processor.join();
        debug_print("%s\n", "Processor thread stopped.");
    }

    /////////////////////////////////////////////////////////////////
    //
    // BINARY SINK
    //
    /////////////////////////////////////////////////////////////////

    template<class T1, class T2>
    binary_event_sink_cpu<T1, T2>::binary_event_sink_cpu(const std::shared_ptr<ep::event_queue<T1>> &in_queue1,
                                                         const std::shared_ptr<ep::event_queue<T2>> &in_queue2)
            : binary_event_sink<T1, T2>(in_queue1, in_queue2),
              next_read_l(0),
              next_read_r(0),
              last_batch_time(0),
              latency_accum(0),
              latency_count(0),
              stopped(false),
              processor(&binary_event_sink_cpu<T1, T2>::run, this) {
    }

    template<class T1, class T2>
    binary_event_sink_cpu<T1, T2>::~binary_event_sink_cpu() {
        debug_print("%s\n", "Destroying bi-sink.");
        stop();
        debug_print("%s\n", "Finished destroying bi-sink.");
    }

    template<class T1, class T2>
    void binary_event_sink_cpu<T1, T2>::run() {

        T1 *base_l = this->in_queue1->get_base();
        uint32_t mask_l = this->in_queue1->mask;

        T2 *base_r = this->in_queue2->get_base();
        uint32_t mask_r = this->in_queue2->mask;

        hsa_signal_value_t w_idx_l = 0;
        hsa_signal_value_t w_idx_r = 0;

        while (!stopped) {
            w_idx_l = this->in_queue1->wait_not_empty(100);
            w_idx_r = this->in_queue2->wait_not_empty(100);

            if (w_idx_l >= 0 && w_idx_r >= 0) {
                T1 &l = base_l[next_read_l & mask_l];
                T2 &r = base_r[next_read_r & mask_r];

                while (!stopped) {
                    if (l.timestamp <= r.timestamp) {
                        process_element_left(l);
                        this->in_queue1->consume_until(++next_read_l);

                        while (!stopped) {
                            w_idx_l = this->in_queue1->wait_not_empty(100);
                            if (w_idx_l >= 0) {
                                l = base_l[next_read_l & mask_l];
                                break;
                            }
                        }

                    } else {
                        process_element_right(r);
                        this->in_queue2->consume_until(++next_read_r);

                        while (!stopped) {
                            w_idx_r = this->in_queue2->wait_not_empty(100);
                            if (w_idx_r >= 0) {
                                r = base_r[next_read_r & mask_r];
                                break;
                            }
                        }
                    }

                    uint64_t now = static_cast<uint64_t >(std::chrono::steady_clock::now().time_since_epoch().count());
                    if ( last_batch_time > 0 ) {
                        latency_accum += (now-last_batch_time);
                        latency_count++;
                    }
                    last_batch_time = now;
                }
            }
        }
    }

    template<class T1, class T2>
    uint64_t binary_event_sink_cpu<T1, T2>::get_processing_latency() {
        return latency_count > 0 ? latency_accum/latency_count : 0;
    }

    template<class T1, class T2>
    void binary_event_sink_cpu<T1, T2>::stop() {
        if (stopped)
            return;
        debug_print("%s\n", "Stopping processor thread.");
        stopped = true;
        processor.join();
        debug_print("%s\n", "Processor thread stopped.");
    }

    /////////////////////////////////////////////////////////////////
    //
    // UNARY OPERATOR
    //
    /////////////////////////////////////////////////////////////////


    template<class TIN, class TOUT>
    unary_operator_cpu<TIN, TOUT>::unary_operator_cpu(const std::shared_ptr<ep::event_queue<TIN>> &in_queue,
                                                      const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                                                      uint32_t min_batch_size,
                                                      uint32_t max_batch_size) : event_source<TOUT>(out_queue),
                                                                                 event_sink_cpu<TIN>(in_queue,
                                                                                                     min_batch_size,
                                                                                                     max_batch_size) {
    }

    template<class TIN, class TOUT>
    unary_operator_cpu<TIN, TOUT>::~unary_operator_cpu() {
        debug_print("%s\n", "Destroying op cpu.");
    }

    template<class TI1, class TI2, class TOUT>
    binary_operator_cpu<TI1, TI2, TOUT>::binary_operator_cpu(
            const std::shared_ptr<ep::event_queue<TI1>> &in_queue1, const std::shared_ptr<ep::event_queue<TI2>> &in_queue2,
            const std::shared_ptr<ep::event_queue<TOUT>> &out_queue) : event_source<TOUT>(out_queue),
                                                                   binary_event_sink_cpu<TI1, TI2>(in_queue1,
                                                                                                   in_queue2) {
    }

    template<class TI1, class TI2, class TOUT>
    binary_operator_cpu<TI1, TI2, TOUT>::~binary_operator_cpu() {
        debug_print("%s\n", "Destroying binary op cpu.");
    }

    /////////////////////////////////////////////////////////////////
    //
    // printer
    //
    /////////////////////////////////////////////////////////////////

    template<class T>
    event_printer<T>::event_printer(const std::shared_ptr<ep::event_queue<T>> &in_queue, uint32_t min_batch_size,
                                    uint32_t max_batch_size) :
            event_sink_cpu<T>(in_queue, min_batch_size, max_batch_size) {

    }

    template<class T>
    void event_printer<T>::process_element(const T &event) {
        std::cout << "Out: " << event << std::endl;
    }

    template<class T>
    event_printer<T>::~event_printer() {
        debug_print("%s\n", "Destroying printer cpu.");
    }

    template<class T>
    event_eater<T>::event_eater(const std::shared_ptr<ep::event_queue<T>> &in_queue, uint32_t min_batch_size,
                                uint32_t max_batch_size)
            :event_sink_cpu<T>(in_queue, min_batch_size, max_batch_size), count(0), unique_val(0.0) {
    }

    /////////////////////////////////////////////////////////////////
    //
    // EATER
    //
    /////////////////////////////////////////////////////////////////

    template<class T>
    void event_eater<T>::process_batch(hsa_signal_value_t begin, hsa_signal_value_t end) {
        count += (end-begin);
    }


//    template<class T>
//    void event_eater<T>::process_element(const T &event) {
//        (void) event;
//        count++;
//    }
//
//    template<>
//    void event_eater<joined_event>::process_element(const joined_event &event) {
//        (void) event;
//        count++;
//        unique_val += (event.values[0]+event.values[NUM_VALUES]);
//    }
//
//    template<>
//    void event_eater<event>::process_element(const event &event) {
//        (void) event;
//        count++;
//        unique_val += (event.values[0]);
//    }
//
//    template<>
//    void event_eater<aggregated_event>::process_element(const aggregated_event &event) {
//        (void) event;
//        count++;
//        unique_val += (event.values[0]);
//    }

    template<class T>
    event_eater<T>::~event_eater() {
        debug_print("%s\n", "Destroying eater.");
    }

    /////////////////////////////////////////////////////////////////
    //
    // FILTER
    //
    /////////////////////////////////////////////////////////////////

    template<class T>
    filter_cpu<T>::filter_cpu(const std::shared_ptr<ep::event_queue<T>> &in_queue,
                              const std::shared_ptr<ep::event_queue<T>> &out_queue, uint32_t min_batch_size,
                              uint32_t max_batch_size,
                              std::function<bool(const T&)> predicate) : unary_operator_cpu<T, T>(in_queue, out_queue,
                                                                                           min_batch_size,
                                                                                           max_batch_size),
                                                                  predicate(predicate), writer(out_queue) {
    }

    template<class T>
    void filter_cpu<T>::process_batch(hsa_signal_value_t begin, hsa_signal_value_t end) {
        event_sink_cpu<T>::process_batch(begin, end);
        writer.write_batch(write_buffer,true);
    }

    template<class T>
    void filter_cpu<T>::process_element(const T &event) {
        if (predicate(event)) {
//            this->out_queue->add(event);
            write_buffer.push_back(event);
            writer.write_batch(write_buffer);
        }
    }

    /////////////////////////////////////////////////////////////////
    //
    // JOIN
    //
    /////////////////////////////////////////////////////////////////

    template<class T1, class T2, class TOUT>
    join_cpu<T1, T2, TOUT>::join_cpu(const std::shared_ptr<ep::event_queue<T1>> &in_queue1,
                                     const std::shared_ptr<ep::event_queue<T2>> &in_queue2,
                                     const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                                     uint64_t window_size, const hsa::hsa_agent &agent,
                                     std::function<bool(const T1 &, const T2 &)> predicate,
                                     std::function<TOUT(const T1 &, const T2 &)> assemble) :
            join_cpu(in_queue1,in_queue2,out_queue,count_window<T1>(window_size,agent),count_window<T2>(window_size,agent),predicate,assemble) {
    }

    template<class T1, class T2, class TOUT>
    join_cpu<T1, T2, TOUT>::join_cpu(const std::shared_ptr<ep::event_queue<T1>> &in_queue1,
                                     const std::shared_ptr<ep::event_queue<T2>> &in_queue2,
                                     const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                                     count_window<T1> &&window_l, count_window<T2> &&window_r,
                                     std::function<bool(const T1 &, const T2 &)> predicate,
                                     std::function<TOUT(const T1 &, const T2 &)> assemble) :
    binary_operator_cpu<T1, T2, TOUT>(in_queue1, in_queue2, out_queue),
    window_l(std::move(window_l)), window_r(std::move(window_r)), predicate(predicate), assemble(assemble), writer(out_queue) {

    }

    template<class T1, class T2, class TOUT>
    void join_cpu<T1, T2, TOUT>::process_element_left(const T1 &event) {
        window_l.add(event);
        for (uint64_t i = 0; i < std::min(window_r.get_inserts(), window_r.get_capacity()); i++) {
            if (predicate(event, window_r.get_base()[i])) {
//                this->out_queue->add(assemble(event, window_r.get_base()[i]));
                write_buffer.push_back(assemble(event, window_r.get_base()[i]));
                writer.write_batch(write_buffer);
            }
        }
        writer.write_batch(write_buffer,true);
    }

    template<class T1, class T2, class TOUT>
    void join_cpu<T1, T2, TOUT>::process_element_right(const T2 &event) {
        window_r.add(event);
        for (uint64_t i = 0; i < std::min(window_l.get_inserts(), window_l.get_capacity()); i++) {
            if (predicate(window_l.get_base()[i], event)) {
//                this->out_queue->add(assemble(window_l.get_base()[i], event));
                write_buffer.push_back(assemble(window_l.get_base()[i], event));
                writer.write_batch(write_buffer);
            }
        }
        writer.write_batch(write_buffer,true);
    }

    template<class T1, class T2, class TOUT>
    join_cpu<T1, T2, TOUT>::~join_cpu() {
        debug_print("%s\n", "Destroying join cpu.");
        binary_event_sink_cpu<T1, T2>::stop();
        debug_print("%s\n", "Finished destroying join cpu.");
    }


    /////////////////////////////////////////////////////////////////
    //
    // AGGREGATE
    //
    /////////////////////////////////////////////////////////////////

    template<class TIN, class TOUT, class I_TYPE>
    aggregate_cpu<TIN, TOUT, I_TYPE>::aggregate_cpu(const std::shared_ptr<ep::event_queue<TIN>> &in_queue, const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
            uint64_t  window_size, const hsa::hsa_agent &agent,
            std::function<I_TYPE(const TIN&)> init,
            std::function<void(I_TYPE&, const I_TYPE&)> merge,
            std::function<TOUT(const TIN&, const I_TYPE&)> out) :
            aggregate_cpu(in_queue,out_queue,count_window<TIN>(window_size,agent),init,merge,out) {
    }

    template<class TIN, class TOUT, class I_TYPE>
    aggregate_cpu<TIN, TOUT, I_TYPE>::aggregate_cpu(const std::shared_ptr<ep::event_queue<TIN>> &in_queue, const std::shared_ptr<ep::event_queue<TOUT>> &out_queue, count_window<TIN> &&window,
                                                    std::function<I_TYPE(const TIN&)> init,
                                                    std::function<void(I_TYPE&, const I_TYPE&)> merge,
                                                    std::function<TOUT(const TIN&, const I_TYPE&)> out) :
            unary_operator_cpu<TIN,TOUT>(in_queue,out_queue,1,1),
            window(std::move(window)),
            init(init),
            merge(merge),
            out(out) {

    }

    template<class TIN, class TOUT, class I_TYPE>
    void aggregate_cpu<TIN, TOUT, I_TYPE>::process_element(const TIN &event) {
        window.add(event);

        const TIN* base = window.get_base();
        I_TYPE accum = init(base[0]);
        for ( uint64_t i = 1; i < std::min(window.get_inserts(), window.get_capacity()); i++ ) {
            merge(accum, init(base[i]));
        }
        TOUT result = out(event,accum);
        this->out_queue->add(result);
    }

    /////////////////////////////////////////////////////////////////
    //
    // INSTANTIATION
    //
    /////////////////////////////////////////////////////////////////

    template class event_sink_cpu<event>;
    template class event_sink_cpu<joined_event>;
    template class event_sink_cpu<aggregated_event>;

    template class binary_event_sink_cpu<event,event>;

    template class event_printer<event>;
    template class event_printer<joined_event>;
    template class event_printer<aggregated_event>;

    template class event_eater<event>;
    template class event_eater<joined_event>;
    template class event_eater<aggregated_event>;

    template class filter_cpu<event>;

    template class join_cpu<event,event,joined_event>;
    template class aggregate_cpu<event,aggregated_event,float>;


    // Latency
    template class event_sink_cpu<event_latency>;
    template class event_sink_cpu<joined_event_latency>;

    template class binary_event_sink_cpu<event_latency,event_latency>;

    template class event_printer<event_latency>;
    template class event_printer<joined_event_latency>;

    template class event_eater<event_latency>;
    template class event_eater<joined_event_latency>;

    template class filter_cpu<event_latency>;

    template class join_cpu<event_latency,event_latency,joined_event_latency>;



}