//
// Created by koerberm on 30.05.18.
//

#ifndef EPHSA_CPU_OPERATORS_H
#define EPHSA_CPU_OPERATORS_H

#include <thread>
#include <functional>

#include "operators.h"

namespace ep {
    template<class T>
    class event_sink_cpu : public event_sink<T> {
    public:
        explicit event_sink_cpu(const std::shared_ptr<ep::event_queue<T>> &in_queue, uint32_t min_batch_size, uint32_t max_batch_size);
        virtual ~event_sink_cpu();
        void stop();
        uint64_t get_processing_latency();
    protected:
        virtual void process_batch( hsa_signal_value_t begin, hsa_signal_value_t end );
        virtual void process_element( const T &event) { (void) event; };
        const T* base;
        const hsa_signal_value_t mask;
    private:
        const uint32_t min_batch_size;
        const uint32_t max_batch_size;

        uint64_t last_batch_time;
        uint64_t latency_accum;
        uint64_t latency_count;

        hsa_signal_value_t next_read;
        bool stopped;
        std::thread processor;
        void run();
    };

    template<class T1, class T2>
    class binary_event_sink_cpu : public binary_event_sink<T1,T2> {
    public:
        explicit binary_event_sink_cpu(const std::shared_ptr<event_queue<T1>> &in_queue1, const std::shared_ptr<event_queue<T2>> &in_queue2);
        virtual ~binary_event_sink_cpu();
        void stop();
        uint64_t get_processing_latency();
    protected:
        virtual void process_element_left( const T1 &event) { (void) event; };
        virtual void process_element_right( const T2 &event) { (void) event; };
    private:
        hsa_signal_value_t next_read_l;
        hsa_signal_value_t next_read_r;

        uint64_t last_batch_time;
        uint64_t latency_accum;
        uint64_t latency_count;

        bool stopped;
        std::thread processor;
        void run();
    };



    template<class TIN, class TOUT>
    class unary_operator_cpu : public event_source<TOUT>, public event_sink_cpu<TIN> {
    public:
        explicit unary_operator_cpu(const std::shared_ptr<ep::event_queue<TIN>> &in_queue, const std::shared_ptr<ep::event_queue<TOUT>> &out_queue, uint32_t min_batch_size, uint32_t max_batch_size);
        virtual ~unary_operator_cpu();
    };

    template<class TI1, class TI2, class TOUT>
    class binary_operator_cpu : public event_source<TOUT>, public binary_event_sink_cpu<TI1, TI2> {
    public:
        explicit binary_operator_cpu( const std::shared_ptr<ep::event_queue<TI1>> &in_queue1,
                                      const std::shared_ptr<ep::event_queue<TI2>> &in_queue2,
                                      const std::shared_ptr<ep::event_queue<TOUT>> &out_queue);
        virtual ~binary_operator_cpu();
    };

    /*
     * Prints out every received event
     */
    template<class T>
    class event_printer : public event_sink_cpu<T> {
    public:
        explicit event_printer(const std::shared_ptr<ep::event_queue<T>> &in_queue, uint32_t min_batch_size, uint32_t max_batch_size);
        ~event_printer();
    protected:
        void process_element( const T &event);
    };

    /*
     * /dev/null event sink
     */
    template<class T>
    class event_eater : public event_sink_cpu<T> {
    public:
        event_eater(const std::shared_ptr<ep::event_queue<T>> &in_queue, uint32_t min_batch_size, uint32_t max_batch_size);
        ~event_eater();
        uint64_t count;
        double unique_val;
    protected:
        void process_batch( hsa_signal_value_t begin, hsa_signal_value_t end );
//        void process_element( const T &event);
    };


    /*
     * CPU Filter
     */
    template<class T>
    class filter_cpu : public unary_operator_cpu<T,T> {
    public:
        filter_cpu(const std::shared_ptr<ep::event_queue<T>> &in_queue, const std::shared_ptr<ep::event_queue<T>> &out_queue, uint32_t min_batch_size, uint32_t max_batch_size, std::function<bool(const T&)> predicate);
    protected:
        void process_batch( hsa_signal_value_t begin, hsa_signal_value_t end );
        void process_element( const T &event);
    private:
        std::function<bool(const T&)> predicate;
        std::vector<T> write_buffer;
        batch_writer<T> writer;
    };

    /*
     * CPU Aggregate
     */
    template<class TIN, class TOUT, class I_TYPE>
    class aggregate_cpu : public unary_operator_cpu<TIN,TOUT> {
    public:
        aggregate_cpu(const std::shared_ptr<ep::event_queue<TIN>> &in_queue, const std::shared_ptr<ep::event_queue<TOUT>> &out_queue, uint64_t  window_size,
                const hsa::hsa_agent &agent,
                std::function<I_TYPE(const TIN&)> init,
                std::function<void(I_TYPE&, const I_TYPE&)> merge,
                std::function<TOUT(const TIN&, const I_TYPE&)> out);
        aggregate_cpu(const std::shared_ptr<ep::event_queue<TIN>> &in_queue, const std::shared_ptr<ep::event_queue<TOUT>> &out_queue, count_window<TIN> &&window,
                std::function<I_TYPE(const TIN&)> init,
                std::function<void(I_TYPE&, const I_TYPE&)> merge,
                std::function<TOUT(const TIN&, const I_TYPE&)> out);
    protected:
        void process_element( const TIN &event);
    private:
        count_window<TIN> window;
        std::function<I_TYPE(const TIN&)> init;
        std::function<void(I_TYPE&, const I_TYPE&)> merge;
        std::function<TOUT(const TIN&, const I_TYPE&)> out;
    };

    /*
     * CPU Join
     */
    template<class T1, class T2, class TOUT>
    class join_cpu : public binary_operator_cpu<T1,T2,TOUT> {
    public:
        join_cpu(const std::shared_ptr<ep::event_queue<T1>> &in_queue1, const std::shared_ptr<ep::event_queue<T2>> &in_queue2,
                 const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                 uint64_t window_size, const hsa::hsa_agent &agent,
                 std::function<bool(const T1&,const T2&)> predicate,
                 std::function<TOUT(const T1&,const T2&)> assemble);

        join_cpu(const std::shared_ptr<ep::event_queue<T1>> &in_queue1, const std::shared_ptr<ep::event_queue<T2>> &in_queue2,
                 const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                 count_window<T1> &&window_l, count_window<T2> &&window_r,
                 std::function<bool(const T1&,const T2&)> predicate,
                 std::function<TOUT(const T1&,const T2&)> assemble);

        ~join_cpu();

    protected:
        void process_element_left( const T1 &event);
        void process_element_right( const T2 &event);
    private:
        count_window<T1> window_l;
        count_window<T2> window_r;
        std::function<bool(const T1&,const T2&)> predicate;
        std::function<TOUT(const T1&,const T2&)> assemble;
        std::vector<TOUT> write_buffer;
        batch_writer<TOUT> writer;
    };
}

#endif //EPHSA_CPU_OPERATORS_H
