//
// Created by koerberm on 30.05.18.
//

#include <bitset>
#include "cpu_parallel_operators.h"
#include "ep_common.h"
#include "event.h"
#include "latency/event_latency.h"

namespace ep {

    sync_context::sync_context(uint16_t num_workers, const hsa::hsa_platform &platform) :
      num_workers(num_workers),
      sync1(platform.create_signal(0)),
      sync2(platform.create_signal(0)) {
    }

    void sync_context::worker_ready() {
        sync1->add_screlease(1);
    }

    void sync_context::worker_wait(bool s) {
        auto &sig = s ? *sync1 : *sync2;
        sig.wait_scacquire(HSA_SIGNAL_CONDITION_EQ,0);
    }

    void sync_context::worker_done(bool s) {
        auto &sig = s ? *sync2 : *sync1;
        sig.add_screlease(1);
    }

    void sync_context::coord_wait(bool s) {
        auto &sig = s ? *sync1 : *sync2;
        sig.wait_scacquire(HSA_SIGNAL_CONDITION_EQ, num_workers);
    }

    void sync_context::coord_fire(bool s) {
        auto &sig = s ? *sync1 : *sync2;
        sig.store_screlease(0);
    }


    ///////////////////////////////////////////////7
    filter_context::filter_context(uint16_t num_workers, const hsa::hsa_platform &platform) : sync_context(num_workers,platform) {
    }


    /////////////////////////////////////////////////////////////////
    //
    // PARALLEL FILTER
    //
    /////////////////////////////////////////////////////////////////


    template<class T>
    parallel_filter<T>::parallel_filter(const std::shared_ptr<event_queue<T>> &in_queue,
                                  const std::shared_ptr<event_queue<T>> &out_queue, uint32_t min_batch_size,
                                  uint32_t max_batch_size, uint16_t num_threads,
                                  std::function<bool(const T&)> predicate, const hsa::hsa_platform &platform,
                                  const hsa::hsa_agent &agent) :
            event_source<T>(out_queue),
            event_sink<T>(in_queue),
            platform(platform),
            agent(agent),

            min_batch_size(min_batch_size),
            max_batch_size(max_batch_size),
            num_threads(num_threads),
            predicate(predicate),
            s_ctx(num_threads,platform),
            writer(out_queue),

            last_batch_time(0),
            latency_accum(0),
            latency_count(0),

            stopped(false),
            processor(&parallel_filter<T>::run,this) {
    }

    template<class T>
    parallel_filter<T>::~parallel_filter() {
        stop();
    }

    template<class T>
    void parallel_filter<T>::stop() {
        if ( stopped )
            return;
        debug_print("%s\n", "Stopping processor thread.");
        stopped = true;
        processor.join();
        debug_print("%s\n", "Processor thread stopped.");
    }

    template<class T>
    void parallel_filter<T>::run() {

        const uint32_t events_per_element = sizeof(HELLS_FILTER_BITFIELD_TYPE) << 3;
        const uint32_t result_length = (max_batch_size+events_per_element-1)/events_per_element;
        results[0] = agent.get_fine_grained_region().allocate(sizeof(HELLS_FILTER_BITFIELD_TYPE)*result_length);
        results[1] = agent.get_fine_grained_region().allocate(sizeof(HELLS_FILTER_BITFIELD_TYPE)*result_length);


        hsa_signal_value_t  next_read = this->in_queue->get_read_idx();
        hsa_signal_value_t  last_read = next_read;

        std::vector<std::thread> workers;
        for ( uint16_t i = 0; i < num_threads; i++ ) {
            workers.emplace_back( &parallel_filter<T>::work,this,i,result_length);
        }

        bool flag = false;
        while ( !stopped ) {
            auto w_idx = this->in_queue->wait_for(min_batch_size, 100);

            if ( w_idx >= 0  ) {
                flag = !flag;

                // Wait for kernel to finish
                s_ctx.coord_wait(flag);

                int64_t batch_size = std::min(w_idx-next_read,static_cast<int64_t >(max_batch_size));

                s_ctx.results = results[flag]->template address<HELLS_FILTER_BITFIELD_TYPE>();
                s_ctx.begin   = next_read;
                s_ctx.end     = next_read+batch_size;

                s_ctx.coord_fire(flag);

                // Probe
                probe(results[!flag]->template address<HELLS_FILTER_BITFIELD_TYPE>(), result_length, last_read, next_read);
                results[flag^1]->clear();

                last_read = next_read;
                next_read += batch_size;
            }
        }
        s_ctx.coord_fire(false);
        s_ctx.coord_fire(true);
        for ( auto &t : workers )
            t.join();
    }

    template<class T>
    void parallel_filter<T>::probe(const HELLS_FILTER_BITFIELD_TYPE *results, uint32_t result_length, hsa_signal_value_t begin, hsa_signal_value_t end) {
        for ( uint32_t block_idx = 0; block_idx < result_length; block_idx++ ) {
            HELLS_FILTER_BITFIELD_TYPE block = results[block_idx];

            // Rough check
            if ( block == 0 )
                continue;

            // Fine check
            for ( uint32_t bit = 0; bit < (sizeof(HELLS_FILTER_BITFIELD_TYPE)<<3); bit++ ) {
                if ( (block & (1<<bit)) != 0 ) {
                    hsa_signal_value_t idx = begin + block_idx * sizeof(HELLS_FILTER_BITFIELD_TYPE) + bit;
                    if ( idx < end ) {
//                        this->out_queue->add(this->in_queue->get(idx));
                        write_buffer.push_back(this->in_queue->get(idx));
                        writer.write_batch(write_buffer);
                    }
                }
            }
        }
        writer.write_batch(write_buffer, true);

        // Consume processed events
        this->in_queue->consume_until(end);

        uint64_t now = static_cast<uint64_t >(std::chrono::steady_clock::now().time_since_epoch().count());
        if ( last_batch_time > 0 ) {
            latency_accum += (now-last_batch_time);
            latency_count++;
        }
        last_batch_time = now;
    }

    template<class T>
    uint64_t parallel_filter<T>::get_processing_latency() {
        return latency_count > 0 ? latency_accum/latency_count : 0;
    }

    template<class T>
    void parallel_filter<T>::work(uint16_t tid, uint32_t result_length) {

        T *base   = this->in_queue->get_base();
        auto mask = this->in_queue->mask;

        auto           blocks_per_thread = (result_length + s_ctx.num_workers-1) / s_ctx.num_workers;
        const uint32_t events_per_block  = sizeof(HELLS_FILTER_BITFIELD_TYPE) << 3;
        const uint64_t events_per_thread = blocks_per_thread * events_per_block;


        bool flag = false;
        s_ctx.worker_ready();
        while ( !stopped ) {
            flag = !flag;
            s_ctx.worker_wait(flag);
            if ( stopped )
                break;

            auto begin = s_ctx.begin + tid*events_per_thread;

            for ( uint64_t block_num = 0; block_num < blocks_per_thread; block_num++ ) {
                hsa_signal_value_t block_idx = begin + (block_num*events_per_block);
                HELLS_FILTER_BITFIELD_TYPE &block = s_ctx.results[blocks_per_thread*tid+block_num];
                for ( uint32_t bit = 0; bit < events_per_block; bit++ ) {
                    auto  idx = block_idx+bit;
                    if ( idx < s_ctx.end && predicate(base[idx&mask]) ) {
                        block |= (1<<bit);
                    }
                }
            }
            s_ctx.worker_done(flag);
        }
    }





    /////////////////////////////////////////////////////////////////
    //
    // PARALLEL AGGREGATE
    //
    /////////////////////////////////////////////////////////////////

    template<class TIN, class TOUT, class I_TYPE>
    parallel_aggregate<TIN, TOUT, I_TYPE>::parallel_aggregate(
            const std::shared_ptr<event_queue<TIN>> &in_queue,
            const std::shared_ptr<event_queue<TOUT>> &out_queue,
            uint64_t window_size,
            uint16_t  num_threads,
            const hsa::hsa_platform &platform,
            const hsa::hsa_agent &agent,
            std::function<I_TYPE(const TIN&)> init,
            std::function<void(I_TYPE&, const I_TYPE&)> merge,
            std::function<TOUT(const TIN&, const I_TYPE&)> out
    ) : parallel_aggregate<TIN,TOUT,I_TYPE>(
            in_queue,
            out_queue,
            count_window<TIN>(window_size,agent),
            num_threads,
            platform,
            agent,
            init,
            merge,
            out) {
    }


    template<class TIN, class TOUT, class I_TYPE>
    parallel_aggregate<TIN, TOUT, I_TYPE>::parallel_aggregate(
            const std::shared_ptr<event_queue<TIN>> &in_queue,
            const std::shared_ptr<event_queue<TOUT>> &out_queue,
            count_window<TIN> &&window,
            uint16_t  num_threads,
            const hsa::hsa_platform &platform,
            const hsa::hsa_agent &agent,
            std::function<I_TYPE(const TIN&)> init,
            std::function<void(I_TYPE&, const I_TYPE&)> merge,
            std::function<TOUT(const TIN&, const I_TYPE&)> out ) :

            event_source<TOUT>(out_queue),
            event_sink<TIN>(in_queue),

            platform(platform),
            agent(agent),

            window(std::move(window)),
            num_threads(num_threads),

            init(init),
            merge(merge),
            out(out),

            results(agent.get_fine_grained_region().allocate(sizeof(I_TYPE)*num_threads)),
            s_ctx(num_threads,platform),

            last_batch_time(0),
            latency_accum(0),
            latency_count(0),

            stopped(false),
            processor(&parallel_aggregate<TIN, TOUT, I_TYPE>::run, this) {
    }




    template<class TIN, class TOUT, class I_TYPE>
    void parallel_aggregate<TIN, TOUT, I_TYPE>::run() {

        hsa_signal_value_t next_read = this->in_queue->get_read_idx();

        std::vector<std::thread> workers;
        for ( uint16_t i = 0; i < num_threads; i++ ) {
            workers.emplace_back( &parallel_aggregate<TIN, TOUT, I_TYPE>::work,this,i);
        }


        bool flag = true;
        s_ctx.coord_wait(true);

        while ( !stopped ) {
            auto w_idx = this->in_queue->wait_not_empty(100);
            if ( w_idx >= 0  ) {

                // Initial step;
                auto &event = this->in_queue->get(next_read++);
                this->window.add(event);

                s_ctx.coord_fire(flag);
                flag = !flag;


                s_ctx.coord_wait(flag);




                I_TYPE accum = results->template address<I_TYPE>()[0];

                for ( uint16_t i = 1; i < num_threads; i++ ) {
                    merge(accum,results->template address<I_TYPE>()[i]);
                }
                this->out_queue->add( agg_finish(&event,accum) );
                this->in_queue->consume_until(next_read);

                // Latency measure
                uint64_t now = static_cast<uint64_t >(std::chrono::steady_clock::now().time_since_epoch().count());
                if ( last_batch_time > 0 ) {
                    latency_accum += (now-last_batch_time);
                    latency_count++;
                }
                last_batch_time = now;
            }
        }

        s_ctx.coord_fire(false);
        s_ctx.coord_fire(true);
        for ( auto &t : workers )
            t.join();

    }

    template<class TIN, class TOUT, class I_TYPE>
    parallel_aggregate<TIN, TOUT, I_TYPE>::~parallel_aggregate() {
        stop();
    }

    template<class TIN, class TOUT, class I_TYPE>
    uint64_t  parallel_aggregate<TIN, TOUT, I_TYPE>::get_processing_latency() {
        return (latency_count != 0) ? latency_accum/latency_count : 0;
    }

    template<class TIN, class TOUT, class I_TYPE>
    void parallel_aggregate<TIN, TOUT, I_TYPE>::stop() {
        if ( stopped )
            return;
        debug_print("%s\n", "Stopping processor thread.");
        stopped = true;
        processor.join();
        debug_print("%s\n", "Processor thread stopped.");
    }

    template<class TIN, class TOUT, class I_TYPE>
    void parallel_aggregate<TIN, TOUT, I_TYPE>::work(uint16_t t_id) {
        auto base   = this->window.get_base();

        auto elements_per_thread = (window.get_capacity() + num_threads - 1) / num_threads;

        hsa_signal_value_t my_begin = t_id * elements_per_thread;
        hsa_signal_value_t my_end   = std::min(my_begin + elements_per_thread, window.get_capacity());

        bool flag = false;
        s_ctx.worker_ready();
        while ( !stopped ) {
            flag = !flag;
            s_ctx.worker_wait(flag);
            if ( stopped )
                break;

            hsa_signal_value_t max = std::min(this->window.get_inserts(),this->window.get_capacity());
            if ( my_begin < max ) {
                I_TYPE accum = init(base[my_begin]);
                for ( hsa_signal_value_t i = my_begin+1; i < my_end; i++ ) {
                    merge(accum, init(base[i]) );
                }
                results->template address<I_TYPE>()[t_id] = accum;
            }
            s_ctx.worker_done(flag);
        }


    }



    /////////////////////////////////////////////////////////////////
    //
    // PARALLEL JOIN
    //
    /////////////////////////////////////////////////////////////////

    template<class TI1, class TI2, class TOUT>
    parallel_join<TI1, TI2, TOUT>::parallel_join(const std::shared_ptr<ep::event_queue<TI1>> &in_queue1,
                                           const std::shared_ptr<ep::event_queue<TI2>> &in_queue2,
                                           const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                                           uint32_t window_size,
                                           uint16_t num_threads,
                                           std::function<bool(const TI1&,const TI2&)> predicate,
                                           std::function<TOUT(const TI1&,const TI2&)> assemble,
                                           const hsa::hsa_platform &platform,
                                           const hsa::hsa_agent &agent) :
            parallel_join<TI1,TI2,TOUT>(
            in_queue1, in_queue2, out_queue, count_window<TI1>(window_size,agent), count_window<TI2>(window_size,agent),
            num_threads, predicate, assemble, platform, agent ){
    }

    template<class TI1, class TI2, class TOUT>
    parallel_join<TI1, TI2, TOUT>::parallel_join(const std::shared_ptr<event_queue<TI1>> &in_queue1,
                                           const std::shared_ptr<event_queue<TI2>> &in_queue2,
                                           const std::shared_ptr<event_queue<TOUT>> &out_queue,
                                           count_window<TI1> &&window_l, count_window<TI2> &&window_r,
                                           uint16_t num_threads,
                                           std::function<bool(const TI1&,const TI2&)> predicate,
                                           std::function<TOUT(const TI1&,const TI2&)> assemble,
                                           const hsa::hsa_platform &platform, const hsa::hsa_agent &agent) :
            event_source<TOUT>(out_queue),
            binary_event_sink<TI1, TI2>(in_queue1, in_queue2),
            num_threads(num_threads),
            window_l(std::move(window_l)),
            window_r(std::move(window_r)),
            predicate(predicate),
            assemble(assemble),

            s_ctx(num_threads,platform),
            writer(out_queue),

            last_batch_time(0),
            latency_accum(0),
            latency_count(0),
            stopped(false),
            processor(&parallel_join<TI1, TI2, TOUT>::run, this) {
        (void) agent;
    }


    template<class TI1, class TI2, class TOUT>
    parallel_join<TI1, TI2, TOUT>::~parallel_join() {
        stop();
    }

    template<class TI1, class TI2, class TOUT>
    void parallel_join<TI1, TI2, TOUT>::run() {
        hsa_signal_value_t next_read_l = this->in_queue1->get_read_idx();
        hsa_signal_value_t next_read_r = this->in_queue2->get_read_idx();

        hsa_signal_value_t w_idx_l = -1;
        hsa_signal_value_t w_idx_r = -1;

        TI1 l;
        TI2 r;

        // Initial gathering of events
        while ( !stopped && (w_idx_l < 0 || w_idx_r < 0) ) {
            if ( w_idx_l < 0 )
                w_idx_l = this->in_queue1->wait_not_empty(100);
            if ( w_idx_r < 0 )
                w_idx_r = this->in_queue2->wait_not_empty(100);
        }

        if ( stopped )
            return;

        l = this->in_queue1->get(next_read_l);
        r = this->in_queue2->get(next_read_r);

        std::vector<std::thread> workers;
        for ( uint16_t i = 0; i < num_threads; i++ ) {
            workers.emplace_back( &parallel_join<TI1, TI2, TOUT>::work,this,i);
        }


        bool flag = false;


        while ( !stopped ) {
            flag = !flag;

            s_ctx.coord_wait(flag);

            uint64_t now = static_cast<uint64_t >(std::chrono::steady_clock::now().time_since_epoch().count());
            if ( last_batch_time > 0 ) {
                latency_accum += (now-last_batch_time);
                latency_count++;
            }
            last_batch_time = now;

            if (l.timestamp <= r.timestamp) {
                s_ctx.probe_left = false;
                s_ctx.left = l;
                s_ctx.coord_fire(flag);

                window_l.add(l);
                this->in_queue1->consume_until(++next_read_l);

                while ( !stopped ) {
                    w_idx_l = this->in_queue1->wait_not_empty(100);
                    if ( w_idx_l >= 0 ) {
                        l = this->in_queue1->get(next_read_l);
                        break;
                    }
                }

            }
            else {
                s_ctx.probe_left = true;
                s_ctx.right = r;
                s_ctx.coord_fire(flag);

                window_r.add(r);
                this->in_queue2->consume_until(++next_read_r);

                while ( !stopped ) {
                    w_idx_r = this->in_queue2->wait_not_empty(100);
                    if ( w_idx_r >= 0 ) {
                        r = this->in_queue2->get(next_read_r);
                        break;
                    }
                }
            }
        }
        s_ctx.coord_fire(false);
        s_ctx.coord_fire(true);
        for ( auto &t : workers )
            t.join();
    }


    template<class TI1, class TI2, class TOUT>
    void parallel_join<TI1, TI2, TOUT>::work(uint16_t t_id) {

        auto elements_per_thread_l = (window_l.get_capacity() + num_threads - 1) / num_threads;
        hsa_signal_value_t my_begin_l = t_id * elements_per_thread_l;
        hsa_signal_value_t my_end_l   = std::min(my_begin_l + elements_per_thread_l, window_l.get_capacity());

        auto elements_per_thread_r = (window_r.get_capacity() + num_threads - 1) / num_threads;
        hsa_signal_value_t my_begin_r = t_id * elements_per_thread_r;
        hsa_signal_value_t my_end_r   = std::min(my_begin_r + elements_per_thread_r, window_r.get_capacity());

        bool flag = false;
        s_ctx.worker_ready();

        std::vector<TOUT> write_buffer;

        while ( !stopped ) {
            flag = !flag;
            s_ctx.worker_wait(flag);
            if ( stopped )
                break;

            if ( s_ctx.probe_left ) {
                hsa_signal_value_t max = std::min(this->window_l.get_inserts(),this->window_l.get_capacity());

                for ( hsa_signal_value_t i = my_begin_l; i < std::min(my_end_l,max); i++ ) {
                    if ( predicate(window_l.get_base()[i],s_ctx.right) ) {
//                        this->out_queue->add( assemble(window_l.get_base()[i],s_ctx.right) );
                        write_buffer.push_back(assemble(window_l.get_base()[i], s_ctx.right));
                        writer.write_batch(write_buffer);
                    }
                }
            }
            else {
                hsa_signal_value_t max = std::min(this->window_r.get_inserts(),this->window_r.get_capacity());

                for ( hsa_signal_value_t i = my_begin_r; i < std::min(my_end_r,max); i++ ) {
                    if ( predicate(s_ctx.left, window_r.get_base()[i]) ) {
//                        this->out_queue->add( assemble(s_ctx.left, window_r.get_base()[i]) );
                        write_buffer.push_back(assemble(s_ctx.left, window_r.get_base()[i]));
                        writer.write_batch(write_buffer);
                    }
                }
            }

            writer.write_batch(write_buffer, true);

            s_ctx.worker_done(flag);
        }
    }

    template<class TI1, class TI2, class TOUT>
    uint64_t parallel_join<TI1, TI2, TOUT>::get_processing_latency() {
        return (latency_count > 0) ? latency_accum/latency_count : 0;
    }

    template<class TI1, class TI2, class TOUT>
    void parallel_join<TI1, TI2, TOUT>::stop() {
        if ( stopped )
            return;
        debug_print("%s\n", "Stopping processor thread.");
        stopped = true;
        processor.join();
        debug_print("%s\n", "Processor thread stopped.");
    }

    /////////////////////////////////////////////////////////////////
    //
    // INSTANTIATION
    //
    /////////////////////////////////////////////////////////////////


    template class join_context<event, event>;
    template class parallel_join<event, event, joined_event>;
    template class parallel_filter<event>;
    template class parallel_aggregate<event,aggregated_event,float>;

    // Latency
    template class join_context<event_latency, event_latency>;
    template class parallel_join<event_latency, event_latency, joined_event_latency>;
    template class parallel_filter<event_latency>;



}