//
// Created by koerberm on 30.05.18.
//

#ifndef EPHSA_CPU_PARALLEL_OPERATORS_H
#define EPHSA_CPU_PARALLEL_OPERATORS_H

#include <thread>
#include <functional>
#include <memory>
#include <atomic>

#include "event_queue.h"
#include "operators.h"

namespace ep {

    /////////////////////////////////////////////////////////////////
    //
    // PARALLEL FILTER
    //
    /////////////////////////////////////////////////////////////////

    class sync_context {
    public:
        sync_context(uint16_t num_workers, const hsa::hsa_platform &platform);
        void worker_ready();
        void worker_wait(bool s);
        void worker_done(bool s);
        void coord_wait(bool s);
        void coord_fire(bool s);
        const uint16_t num_workers;
    private:
        std::shared_ptr<hsa::hsa_signal> sync1;
        std::shared_ptr<hsa::hsa_signal> sync2;
    };


    class filter_context : public sync_context {
    public:
        filter_context(uint16_t num_workers, const hsa::hsa_platform &platform);
        hsa_signal_value_t begin;
        hsa_signal_value_t end;
        HELLS_FILTER_BITFIELD_TYPE *results;
    };



    template<class T>
    class parallel_filter : public event_source<T>, public event_sink<T> {
    public:
        parallel_filter(const std::shared_ptr <ep::event_queue<T>> &in_queue,
                        const std::shared_ptr <ep::event_queue<T>> &out_queue,
                        uint32_t min_batch_size,
                        uint32_t max_batch_size,
                        uint16_t num_threads,
                        std::function<bool(const T&)> predicate,
                        const hsa::hsa_platform &platform,
                        const hsa::hsa_agent &agent);
        ~parallel_filter();
        void stop();

        uint64_t get_processing_latency();
    private:
        const hsa::hsa_platform &platform;
        const hsa::hsa_agent &agent;

        const uint32_t min_batch_size;
        const uint32_t max_batch_size;
        const uint16_t num_threads;
        std::function<bool(const T&)> predicate;

        filter_context s_ctx;
        std::unique_ptr<hsa::hsa_buffer> results[2];
        std::vector<T> write_buffer;
        batch_writer<T> writer;

        uint64_t last_batch_time;
        uint64_t latency_accum;
        uint64_t latency_count;

        bool stopped;

        std::thread processor;

        void run();
        void probe( const HELLS_FILTER_BITFIELD_TYPE *results, uint32_t result_length, hsa_signal_value_t begin, hsa_signal_value_t end);
        void work(uint16_t tid, uint32_t result_length);
    };


    /////////////////////////////////////////////////////////////////
    //
    // PARALLEL AGGREGATE
    //
    /////////////////////////////////////////////////////////////////

    template<class TIN, class TOUT, class I_TYPE>
    class parallel_aggregate : public event_source<TOUT>, public event_sink<TIN> {
    public:
        parallel_aggregate(
                const std::shared_ptr<ep::event_queue<TIN>> &in_queue,
                const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                uint64_t  window_size,
                uint16_t  num_threads,
                const hsa::hsa_platform &platform,
                const hsa::hsa_agent &agent,
                std::function<I_TYPE(const TIN&)> init,
                std::function<void(I_TYPE&, const I_TYPE&)> merge,
                std::function<TOUT(const TIN&, const I_TYPE&)> out);

        parallel_aggregate(
                const std::shared_ptr<ep::event_queue<TIN>> &in_queue,
                const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                count_window<TIN> &&window,
                uint16_t  num_threads,
                const hsa::hsa_platform &platform,
                const hsa::hsa_agent &agent,
                std::function<I_TYPE(const TIN&)> init,
                std::function<void(I_TYPE&, const I_TYPE&)> merge,
                std::function<TOUT(const TIN&, const I_TYPE&)> out);

        ~parallel_aggregate();
        void stop();

        uint64_t get_processing_latency();
    private:
        const hsa::hsa_platform      &platform;
        const hsa::hsa_agent         &agent;

        count_window<TIN> window;
        const uint16_t num_threads;

        const std::function<I_TYPE(const TIN&)> init;
        const std::function<void(I_TYPE&, const I_TYPE&)> merge;
        const std::function<TOUT(const TIN&, const I_TYPE&)> out;

        std::unique_ptr<hsa::hsa_buffer> results;
        sync_context s_ctx;



        uint64_t last_batch_time;
        uint64_t latency_accum;
        uint64_t latency_count;

        bool stopped;
        std::thread processor;

        void run();
        void work(uint16_t t_id);
    };

    /////////////////////////////////////////////////////////////////
    //
    // PARALLEL JOIN
    //
    /////////////////////////////////////////////////////////////////

    template<class TI1, class TI2>
    class join_context : public sync_context {
    public:
        join_context(uint16_t num_workers, const hsa::hsa_platform &platform);
        TI1 left;
        TI2 right;
        bool probe_left;
    };

    template<class TI1, class TI2>
    join_context<TI1, TI2>::join_context(uint16_t num_workers, const hsa::hsa_platform &platform):sync_context(
            num_workers, platform) , probe_left(false) {
    }


    template<class TI1, class TI2, class TOUT>
    class parallel_join : public event_source<TOUT>, public binary_event_sink<TI1,TI2> {
    public:
        parallel_join(const std::shared_ptr<ep::event_queue<TI1>> &in_queue1,
                   const std::shared_ptr<ep::event_queue<TI2>> &in_queue2,
                   const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                   uint32_t window_size,
                   uint16_t num_threads,
                   std::function<bool(const TI1&,const TI2&)> predicate,
                   std::function<TOUT(const TI1&,const TI2&)> assemble,
                   const hsa::hsa_platform &platform,
                   const hsa::hsa_agent &agent);

        parallel_join(const std::shared_ptr<ep::event_queue<TI1>> &in_queue1,
                   const std::shared_ptr<ep::event_queue<TI2>> &in_queue2,
                   const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                   count_window<TI1> &&window_l, count_window<TI2> &&window_r,
                   uint16_t num_threads,
                   std::function<bool(const TI1&,const TI2&)> predicate,
                   std::function<TOUT(const TI1&,const TI2&)> assemble,
                   const hsa::hsa_platform &platform,
                   const hsa::hsa_agent &agent);

        ~parallel_join();
        void stop();

        uint64_t get_processing_latency();
    private:
        const uint32_t num_threads;
        count_window<TI1> window_l;
        count_window<TI2> window_r;

        const std::function<bool(const TI1&,const TI2&)> predicate;
        const std::function<TOUT(const TI1&,const TI2&)> assemble;

        join_context<TI1,TI2> s_ctx;
        batch_writer<TOUT> writer;

        uint64_t last_batch_time;
        uint64_t latency_accum;
        uint64_t latency_count;

        bool stopped;
        std::thread processor;

        void run();
        void work(uint16_t t_id);
    };

}
#endif //EPHSA_CPU_PARALLEL_OPERATORS_H
