//
// Created by koerberm on 30.05.18.
//

#include <bitset>
#include "dispatch_operators.h"
#include "ep_common.h"
#include "event.h"
#include "latency/event_latency.h"

namespace ep {

    /////////////////////////////////////////////////////////////////
    //
    // HELLS FILTER
    //
    /////////////////////////////////////////////////////////////////


    template<class T>
    hells_filter<T>::hells_filter(const std::shared_ptr<event_queue<T>> &in_queue,
                                  const std::shared_ptr<event_queue<T>> &out_queue, uint32_t min_batch_size,
                                  uint32_t max_batch_size, const hsa::hsa_platform &platform,
                                  const hsa::hsa_agent &agent, const hsa::hsa_executable &exe) :
            event_source<T>(out_queue),
            event_sink<T>(in_queue),
            platform(platform),
            agent(agent),
            kernel_symbol(exe.kernel_by_name("hells_filter")),

            writer(out_queue),

            min_batch_size(min_batch_size),
            max_batch_size(max_batch_size),
            last_batch_time(0),
            latency_accum(0),
            latency_count(0),

            stopped(false),
            processor(&hells_filter<T>::run,this) {
    }

    template<class T>
    hells_filter<T>::~hells_filter() {
        stop();
    }

    template<class T>
    void hells_filter<T>::stop() {
        if ( stopped )
            return;
        debug_print("%s\n", "Stopping processor thread.");
        stopped = true;
        processor.join();
        debug_print("%s\n", "Processor thread stopped.");
    }

    template<class T>
    void hells_filter<T>::run() {
        auto completion_signal = platform.create_signal(0);
        auto queue             = agent.create_queue(agent.queue_max_size,HSA_QUEUE_TYPE_SINGLE,UINT32_MAX,UINT32_MAX);

        auto kernargs          = agent.get_kernarg_region().allocate(kernel_symbol.kernarg_segment_size,kernel_symbol.kernarg_segment_alignment);
        auto aux_data          = agent.get_fine_grained_region().allocate(sizeof(hells_filter_aux_data),64);

        std::unique_ptr<hsa::hsa_buffer> results[2];
        const uint32_t events_per_element = sizeof(HELLS_FILTER_BITFIELD_TYPE) << 3;
        const uint32_t result_length = (max_batch_size+events_per_element-1)/events_per_element;
        results[0] = agent.get_fine_grained_region().allocate(sizeof(HELLS_FILTER_BITFIELD_TYPE)*result_length);
        results[1] = agent.get_fine_grained_region().allocate(sizeof(HELLS_FILTER_BITFIELD_TYPE)*result_length);

        auto aux_addr = aux_data->template address<hells_filter_aux_data>();
        auto arg_addr = kernargs->template address<hells_filter_args>();

        // Prepare aux data
        aux_addr->in_q = this->in_queue->to_struct();

        // Prepare args
        arg_addr->in_q = &aux_addr->in_q;

        hsa_signal_value_t next_read = this->in_queue->get_read_idx();
        hsa_signal_value_t last_read = next_read;

        HELLS_FILTER_BITFIELD_TYPE *next_results = nullptr;

        uint8_t flag = 0;
        while ( !stopped ) {
            auto w_idx = this->in_queue->wait_for(min_batch_size, 100);

            if ( w_idx >= 0  ) {
                // Wait for kernel to finish
                completion_signal->wait_scacquire(HSA_SIGNAL_CONDITION_EQ,0);

                // Schedule next batch
                flag ^= 1;
                completion_signal->store_screlease(1);

                // Compute group size
                int64_t batch_size = std::min(w_idx-next_read,static_cast<int64_t >(max_batch_size));
                int64_t grid_size  = (batch_size+events_per_element-1) / events_per_element;
                int64_t group_size = std::min(grid_size, static_cast<int64_t >(agent.workgroup_max_size));

                // Prepare kernel args
                arg_addr->batch_size = static_cast<uint32_t >(batch_size);
                arg_addr->in_start   = next_read;
                arg_addr->results    = results[flag]->template address<HELLS_FILTER_BITFIELD_TYPE>();

                // Dispatch kernel
                hsa::hsa_kernel_dispatch_packet dispatch_packet(1,kernel_symbol,*completion_signal,*kernargs);
                dispatch_packet.set_grid_size(static_cast<uint32_t >(grid_size));
                dispatch_packet.set_workgroup_size(static_cast<uint16_t >(group_size));
                queue->dispatch_kernel_packet(dispatch_packet);

                // Processing took longer than waiting for next batch
                if ( next_results != nullptr ) {
                    // Probe
                    probe(next_results, result_length, last_read, next_read);
                    results[flag^1]->clear();
                }
                next_results = results[flag]->template address<HELLS_FILTER_BITFIELD_TYPE>();

                last_read = next_read;
                next_read += batch_size;
            }
            // Processing was finished before next batch is ready
            else if ( next_results != nullptr && completion_signal->load_scacquire() == 0 ) {
                // Probe
                probe(next_results, result_length, last_read, next_read);
                results[flag]->clear();
                next_results = nullptr;
            }
        }
    }

    template<class T>
    void hells_filter<T>::probe(const HELLS_FILTER_BITFIELD_TYPE *results, uint32_t result_length, hsa_signal_value_t begin, hsa_signal_value_t end) {
        for ( uint32_t block_idx = 0; block_idx < result_length; block_idx++ ) {
            HELLS_FILTER_BITFIELD_TYPE block = results[block_idx];

            // Rough check
            if ( block == 0 )
                continue;

            // Fine check
            for ( uint32_t bit = 0; bit < (sizeof(HELLS_FILTER_BITFIELD_TYPE)<<3); bit++ ) {
                if ( (block & (1<<bit)) != 0 ) {
                    hsa_signal_value_t idx = begin + block_idx * sizeof(HELLS_FILTER_BITFIELD_TYPE) + bit;
                    if ( idx < end ) {
//                        this->out_queue->add(this->in_queue->get(idx));
                        write_buffer.push_back(this->in_queue->get(idx));
                        writer.write_batch(write_buffer);
                    }
                }
            }
        }

        writer.write_batch(write_buffer,true);

        // Consume processed events
        this->in_queue->consume_until(end);

        uint64_t now = static_cast<uint64_t >(std::chrono::steady_clock::now().time_since_epoch().count());
        if ( last_batch_time > 0 ) {
            latency_accum += (now-last_batch_time);
            latency_count++;
        }
        last_batch_time = now;
    }

    template<class T>
    uint64_t hells_filter<T>::get_processing_latency() {
        return (latency_count > 0) ? latency_accum/latency_count : 0;
    }

    /////////////////////////////////////////////////////////////////
    //
    // HELLS AGGREGATE
    //
    /////////////////////////////////////////////////////////////////

    template<class TIN, class TOUT, class I_TYPE>
    hells_aggregate<TIN, TOUT, I_TYPE>::hells_aggregate(
            const std::shared_ptr<event_queue<TIN>> &in_queue,
            const std::shared_ptr<event_queue<TOUT>> &out_queue,
            uint64_t window_size,
            uint32_t  elements_per_item,
            const hsa::hsa_platform &platform,
            const hsa::hsa_agent &agent,
            const hsa::hsa_executable &exe,
            std::function<TOUT(const TIN&, const I_TYPE&)> out
    ) : hells_aggregate<TIN,TOUT,I_TYPE>(
            in_queue,
            out_queue,
            count_window<TIN>(window_size,agent),
            elements_per_item,
            platform,
            agent,
            exe,
            out) {
    }


    template<class TIN, class TOUT, class I_TYPE>
    hells_aggregate<TIN, TOUT, I_TYPE>::hells_aggregate(
            const std::shared_ptr<event_queue<TIN>> &in_queue,
            const std::shared_ptr<event_queue<TOUT>> &out_queue,
            count_window<TIN> &&window,
            uint32_t  elements_per_item,
            const hsa::hsa_platform &platform,
            const hsa::hsa_agent &agent,
            const hsa::hsa_executable &exe,
            std::function<TOUT(const TIN&, const I_TYPE&)> out ) :

            event_source<TOUT>(out_queue),
            event_sink<TIN>(in_queue),

            platform(platform),
            agent(agent),
            kernel_symbol_init(exe.kernel_by_name("hells_aggregate_init")),
            kernel_symbol_run(exe.kernel_by_name("hells_aggregate")),

            window(std::move(window)),
            out(out),
            elements_per_item(elements_per_item),

            last_batch_time(0),
            latency_accum(0),
            latency_count(0),

            stopped(false),
            processor(&hells_aggregate<TIN, TOUT, I_TYPE>::run, this) {

        if ( elements_per_item < 2 )
            throw new ep::ep_exception("Aggregation must process at least 2 inputs per item");
    }




    template<class TIN, class TOUT, class I_TYPE>
    void hells_aggregate<TIN, TOUT, I_TYPE>::run() {
        auto completion_signal = platform.create_signal(0);
        auto queue             = agent.create_queue(agent.queue_max_size,HSA_QUEUE_TYPE_SINGLE,UINT32_MAX,UINT32_MAX);

        auto kernargs_init     = agent.get_kernarg_region().allocate(kernel_symbol_init.kernarg_segment_size,kernel_symbol_init.kernarg_segment_alignment);
        auto kernargs_run      = agent.get_kernarg_region().allocate(kernel_symbol_run.kernarg_segment_size,kernel_symbol_run.kernarg_segment_alignment);

        std::unique_ptr<hsa::hsa_buffer> results[2];
        const uint64_t accum_size = (window.get_capacity()+elements_per_item-1)/elements_per_item;
        results[0] = agent.get_fine_grained_region().allocate(sizeof(I_TYPE)*accum_size);
        results[1] = agent.get_fine_grained_region().allocate(sizeof(I_TYPE)*accum_size);

        auto arg_addr_init = kernargs_init->template address<hells_agg_init_args>();
        auto arg_addr_run  = kernargs_run->template address<hells_agg_run_args>();


        arg_addr_init->elements_per_item = elements_per_item;
        arg_addr_init->in  = window.get_base();
        arg_addr_init->out = results[0]->template address<I_TYPE>();

        arg_addr_run->elements_per_item  = elements_per_item;

        hsa_signal_value_t next_read = this->in_queue->get_read_idx();

        while ( !stopped ) {
            auto w_idx = this->in_queue->wait_not_empty(100);
            if ( w_idx >= 0  ) {
                // Initial step;
                auto &event = this->in_queue->get(next_read++);
                this->window.add(event);

                uint64_t num_events  = std::min(window.get_inserts(),window.get_capacity());
                uint64_t num_threads = (num_events + elements_per_item - 1) / elements_per_item;
                uint64_t group_size  = std::min(num_threads, static_cast<uint64_t>(agent.workgroup_max_size));

                arg_addr_init->element_count = num_events;

                hsa::hsa_kernel_dispatch_packet init_dispatch(1,kernel_symbol_init,*completion_signal,*kernargs_init);
                init_dispatch.set_grid_size(static_cast<uint32_t >(num_threads));
                init_dispatch.set_workgroup_size(static_cast<uint16_t >(group_size));

                completion_signal->store_screlease(1);
                queue->dispatch_kernel_packet(init_dispatch);
                completion_signal->wait_scacquire(HSA_SIGNAL_CONDITION_EQ,0);

                // Reduce further
                uint8_t flag = 0;
                while ( num_threads > 1 ) {
                    num_events = num_threads;
                    num_threads = (num_events + elements_per_item - 1) / elements_per_item;
                    group_size  = std::min(num_threads, static_cast<uint64_t >(agent.workgroup_max_size));

                    arg_addr_run->in  = results[flag]->template address<I_TYPE>();
                    arg_addr_run->out = results[flag^1]->template address<I_TYPE>();
                    arg_addr_run->element_count = num_events;

                    hsa::hsa_kernel_dispatch_packet run_dispatch(1,kernel_symbol_run,*completion_signal,*kernargs_run);
                    run_dispatch.set_grid_size(static_cast<uint32_t >(num_threads));
                    run_dispatch.set_workgroup_size(static_cast<uint16_t >(group_size));
                    completion_signal->store_screlease(1);
                    queue->dispatch_kernel_packet(run_dispatch);
                    completion_signal->wait_scacquire(HSA_SIGNAL_CONDITION_EQ,0);
                    flag = flag^1;
                }

                // Result is here:
                I_TYPE accum = results[flag]->template address<I_TYPE>()[0];
                this->out_queue->add( agg_finish(&event,accum) );
                // Consume event
                this->in_queue->consume_until(next_read);

                // Latency measure
                uint64_t now = static_cast<uint64_t >(std::chrono::steady_clock::now().time_since_epoch().count());
                if ( last_batch_time > 0 ) {
                    latency_accum += (now-last_batch_time);
                    latency_count++;
                }
                last_batch_time = now;
            }
        }
    }

    template<class TIN, class TOUT, class I_TYPE>
    hells_aggregate<TIN, TOUT, I_TYPE>::~hells_aggregate() {
        stop();
    }

    template<class TIN, class TOUT, class I_TYPE>
    uint64_t  hells_aggregate<TIN, TOUT, I_TYPE>::get_processing_latency() {
        return (latency_count != 0) ? latency_accum/latency_count : 0;
    }

    template<class TIN, class TOUT, class I_TYPE>
    void hells_aggregate<TIN, TOUT, I_TYPE>::stop() {
        if ( stopped )
            return;
        debug_print("%s\n", "Stopping processor thread.");
        stopped = true;
        processor.join();
        debug_print("%s\n", "Processor thread stopped.");
    }

    /////////////////////////////////////////////////////////////////
    //
    // HELLS JOIN
    //
    /////////////////////////////////////////////////////////////////


    template<class TI1, class TI2>
    hells_state<TI1, TI2>::hells_state(uint32_t window_size, const hsa::hsa_agent &agent) :
            hells_state(count_window<TI1>(window_size, agent),count_window<TI2>(window_size, agent), agent) {
    }

    template<class TI1, class TI2>
    hells_state<TI1, TI2>::hells_state(count_window<TI1> &&window_l, count_window<TI2> &&window_r, const hsa::hsa_agent &agent) :
        window_left(std::move(window_l)), window_right(std::move(window_r)){

        const uint64_t window_size = std::max(window_left.get_capacity(),window_right.get_capacity());
        const uint64_t elements_per_thread = sizeof(HELLS_JOIN_BITFIELD_TYPE) << 3;

        result_length = (window_size+elements_per_thread-1) / elements_per_thread;

        results = agent.get_fine_grained_region().allocate(result_length * sizeof(HELLS_JOIN_BITFIELD_TYPE),64);
        if ( !results )
            throw ep_exception("Could not allocate result buffer for hells join.");
    }

    template<class TI1, class TI2>
    hells_state<TI1, TI2>::hells_state(const hells_state &copy) :
            window_left(copy.window_left), window_right(copy.window_right),
            result_length(copy.result_length), probe_right(copy.probe_right),
            left(copy.left), right(copy.right), wdef(copy.wdef) {
        results = std::make_unique<hsa::hsa_buffer>(*copy.results);
    }


    template<class TI1, class TI2>
    hells_state<TI1, TI2>::hells_state(hells_state &&move) noexcept :
            window_left(std::move(move.window_left)), window_right(std::move(move.window_right)),
            result_length(move.result_length), probe_right(move.probe_right),
            left(move.left), right(move.right), wdef(move.wdef) {

    }

    template<class TI1, class TI2>
    hells_state<TI1, TI2> &hells_state<TI1, TI2>::operator=(const hells_state<TI1, TI2> &copy) {
        result_length = copy.result_length;
        left = copy.left;
        right = copy.right;
        probe_right = copy.probe_right;
        wdef = copy.wdef;
        window_left = copy.window_left;
        window_right = copy.window_right;
        results = std::make_unique<hsa::hsa_buffer>(*copy.results);

        return *this;
    }

    template<class TI1, class TI2>
    hells_state<TI1, TI2> &hells_state<TI1, TI2>::operator=(hells_state<TI1, TI2> &&move) noexcept {
        result_length = move.result_length;
        left = move.left;
        right = move.right;
        probe_right = move.probe_right;
        wdef = move.wdef;
        window_left = std::move(move.window_left);
        window_right = std::move(move.window_right);
        results = std::move(move.results);

        return *this;
    }

    template<class TI1, class TI2>
    void hells_state<TI1, TI2>::reset_results() {
        results->clear();
    }

    template<class TI1, class TI2>
    std::ostream &operator<<(std::ostream &s, const hells_state<TI1,TI2> &a) {
        s << "Hells State:" << std::endl;
        s << "  Left event  : " << a.left << std::endl;
        s << "  Right event : " << a.right << std::endl;
        s << "  Probe Right?: " << a.probe_right << std::endl;
        s << "  Results: ";

        std::unique_ptr<hsa::hsa_buffer> &buf = a.results;
        auto base = buf->address<HELLS_JOIN_BITFIELD_TYPE>();
        for ( uint32_t i = 0; i < a.result_length; i++ ) {
            std::bitset<sizeof(HELLS_JOIN_BITFIELD_TYPE)> b(base[i]);
            s << b;
        }
        s << std::endl;
        s << "  Window Left : " << a.window_left << std::endl;
        s << "  Window Right: " << a.window_right << std::endl;
        return s;
    }

    ///////////////////////////////////////////////////////////

    template<class TI1, class TI2, class TOUT>
    hells_join<TI1, TI2, TOUT>::hells_join(const std::shared_ptr<ep::event_queue<TI1>> &in_queue1,
                                           const std::shared_ptr<ep::event_queue<TI2>> &in_queue2,
                                           const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                                           uint32_t window_size, std::function<TOUT(const TI1&,const TI2&)> assemble,
                                           const hsa::hsa_platform &platform,
                                           const hsa::hsa_agent &agent, const hsa::hsa_executable &exe) :
        hells_join<TI1,TI2,TOUT>(
            in_queue1, in_queue2, out_queue, count_window<TI1>(window_size,agent), count_window<TI2>(window_size,agent),
            assemble, platform, agent, exe ){
    }

    template<class TI1, class TI2, class TOUT>
    hells_join<TI1, TI2, TOUT>::hells_join(const std::shared_ptr<event_queue<TI1>> &in_queue1,
                                           const std::shared_ptr<event_queue<TI2>> &in_queue2,
                                           const std::shared_ptr<event_queue<TOUT>> &out_queue,
                                           count_window<TI1> &&window_l, count_window<TI2> &&window_r,
                                           std::function<TOUT(const TI1 &, const TI2 &)> assemble,
                                           const hsa::hsa_platform &platform, const hsa::hsa_agent &agent,
                                           const hsa::hsa_executable &exe) :
            event_source<TOUT>(out_queue),
            binary_event_sink<TI1, TI2>(in_queue1, in_queue2),
            platform(platform),
            agent(agent),
            kernel_symbol(exe.kernel_by_name("hells_join")),
            assemble(assemble),
            elements_per_thread( sizeof(HELLS_JOIN_BITFIELD_TYPE) << 3 ),

            states(agent.get_fine_grained_region().allocate(2*sizeof(hells_state<TI1,TI2>))),
            writer(out_queue),

            flag(0),
            completion_signal(platform.create_signal(0)),
            queue(agent.create_queue(agent.queue_max_size,HSA_QUEUE_TYPE_SINGLE,UINT32_MAX,UINT32_MAX)),
            kernargs(agent.get_kernarg_region().allocate(kernel_symbol.kernarg_segment_size,kernel_symbol.kernarg_segment_alignment)),
            last_batch_time(0),
            latency_accum(0),
            latency_count(0),
            stopped(false) {


        auto sp = states->template address<hells_state<TI1,TI2>>();

        sp[0] = std::move(hells_state<TI1, TI2>(std::move(window_l),std::move(window_r),agent));
        sp[1] = sp[0];

        processor = std::make_unique<std::thread>(&hells_join<TI1, TI2, TOUT>::run, this);
    }


    template<class TI1, class TI2, class TOUT>
    hells_join<TI1, TI2, TOUT>::~hells_join() {
        stop();
    }

    template<class TI1, class TI2, class TOUT>
    void hells_join<TI1, TI2, TOUT>::run() {
        hsa_signal_value_t next_read_l = this->in_queue1->get_read_idx();
        hsa_signal_value_t next_read_r = this->in_queue2->get_read_idx();

        hsa_signal_value_t w_idx_l = -1;
        hsa_signal_value_t w_idx_r = -1;

        TI1 l;
        TI2 r;

        // Initial gathering of events
        while ( !stopped && (w_idx_l < 0 || w_idx_r < 0) ) {
            if ( w_idx_l < 0 )
                w_idx_l = this->in_queue1->wait_not_empty(100);
            if ( w_idx_r < 0 )
                w_idx_r = this->in_queue2->wait_not_empty(100);
        }

        if ( stopped )
            return;

        l = this->in_queue1->get(next_read_l);
        r = this->in_queue2->get(next_read_r);

        while ( !stopped ) {
            if (l.timestamp <= r.timestamp) {
                process_element_left(l);
                this->in_queue1->consume_until(++next_read_l);

                while ( !stopped ) {
                    w_idx_l = this->in_queue1->wait_not_empty(100);
                    if ( w_idx_l >= 0 ) {
                        l = this->in_queue1->get(next_read_l);
                        break;
                    }
                }

            }
            else {
                process_element_right(r);
                this->in_queue2->consume_until(++next_read_r);

                while ( !stopped ) {
                    w_idx_r = this->in_queue2->wait_not_empty(100);
                    if ( w_idx_r >= 0 ) {
                        r = this->in_queue2->get(next_read_r);
                        break;
                    }
                }
            }

            uint64_t now = static_cast<uint64_t >(std::chrono::steady_clock::now().time_since_epoch().count());
            if ( last_batch_time > 0 ) {
                latency_accum += (now-last_batch_time);
                latency_count++;
            }
            last_batch_time = now;
        }
    }

    template<class TI1, class TI2, class TOUT>
    void hells_join<TI1, TI2, TOUT>::process_element_left(const TI1 &event) {
        auto sp = states->template address<hells_state<TI1,TI2>>();
        hells_state<TI1,TI2> &mod = sp[flag];
        hells_state<TI1,TI2> &old = sp[flag^1];

        // Wait for kernel to complete
        completion_signal->wait_scacquire(HSA_SIGNAL_CONDITION_EQ, 0);

        // Prepare arguments
        mod.window_left.add(event);
        mod.left = event;
        mod.probe_right = true;
        mod.wdef = mod.window_right.to_struct();

        auto args = kernargs->template address<hells_join_args>();
        args->probe_right = true;
        args->event  = &mod.left;
        args->window = &mod.wdef;
        args->results = mod.results->template address<HELLS_JOIN_BITFIELD_TYPE>();


        // Compute grid size
        uint64_t num_threads = (mod.window_right.get_capacity()+elements_per_thread-1) / elements_per_thread;
        uint64_t group_size  = std::min(num_threads, static_cast<uint64_t >(agent.workgroup_max_size));

        // Dispatch kernel
        hsa::hsa_kernel_dispatch_packet dispatch_packet(1,kernel_symbol,*completion_signal,*kernargs);
        dispatch_packet.set_grid_size(static_cast<uint32_t >(num_threads));
        dispatch_packet.set_workgroup_size(static_cast<uint16_t >(group_size));
        completion_signal->store_screlease(1);
        queue->dispatch_kernel_packet(dispatch_packet);

        // Output results
        probe(old);

        // Add to old state;
        old.window_left.add(event);

        flag ^= 1;
    }

    template<class TI1, class TI2, class TOUT>
    void hells_join<TI1, TI2, TOUT>::process_element_right(const TI2 &event) {
        auto sp = states->template address<hells_state<TI1,TI2>>();
        hells_state<TI1,TI2> &mod = sp[flag];
        hells_state<TI1,TI2> &old = sp[flag^1];

        // Wait for kernel to complete
        completion_signal->wait_scacquire(HSA_SIGNAL_CONDITION_EQ, 0);

        // Prepare arguments
        mod.window_right.add(event);
        mod.right = event;
        mod.probe_right = false;
        mod.wdef = mod.window_left.to_struct();

        // Schedule kernel
        auto args = kernargs->template address<hells_join_args>();
        args->probe_right = false;
        args->event = &mod.right;
        args->window = &mod.wdef;
        args->results = mod.results->template address<HELLS_JOIN_BITFIELD_TYPE>();

        // Compute grid size
        uint64_t num_threads = (mod.window_left.get_capacity()+elements_per_thread-1) / elements_per_thread;
        uint64_t group_size = std::min(num_threads, static_cast<uint64_t >(agent.workgroup_max_size));

        // Dispatch kernel
        hsa::hsa_kernel_dispatch_packet dispatch_packet(1,kernel_symbol,*completion_signal,*kernargs);
        dispatch_packet.set_grid_size(static_cast<uint32_t >(num_threads));
        dispatch_packet.set_workgroup_size(static_cast<uint16_t >(group_size));
        completion_signal->store_screlease(1);
        queue->dispatch_kernel_packet(dispatch_packet);

        // Output results
        probe(old);

        // Add to old state;
        old.window_right.add(event);

        flag ^= 1;
    }

    template<class TI1, class TI2, class TOUT>
    void hells_join<TI1, TI2, TOUT>::probe(hells_state<TI1, TI2> &state) {
        auto base = state.results->template address<HELLS_JOIN_BITFIELD_TYPE>();

        for ( uint32_t block_idx = 0; block_idx < state.result_length; block_idx++ ) {
            const HELLS_JOIN_BITFIELD_TYPE block = base[block_idx];

            // Rough check
            if ( block == 0 )
                continue;

            // Fine check
            for ( uint32_t bit = 0; bit < elements_per_thread; bit++ ) {
                if ( (block & (1l<<bit)) != 0 ) {
                    uint32_t idx = block_idx * elements_per_thread  + bit;
                    if ( state.probe_right ) {
//                        this->out_queue->add(assemble(state.left, state.window_right.get_base()[idx]));
                        write_buffer.push_back(assemble(state.left, state.window_right.get_base()[idx]));
                        writer.write_batch(write_buffer);
                    }
                    else {
//                        this->out_queue->add(assemble(state.window_left.get_base()[idx], state.right));
                        write_buffer.push_back(assemble(state.window_left.get_base()[idx], state.right));
                        writer.write_batch(write_buffer);
                    }
                }
            }
        }
        writer.write_batch(write_buffer, true);
        state.reset_results();
    }

    template<class TI1, class TI2, class TOUT>
    uint64_t hells_join<TI1, TI2, TOUT>::get_processing_latency() {
        return (latency_count > 0) ? latency_accum/latency_count : 0;
    }

    template<class TI1, class TI2, class TOUT>
    void hells_join<TI1, TI2, TOUT>::stop() {
        if ( stopped )
            return;
        debug_print("%s\n", "Stopping processor thread.");
        stopped = true;
        processor->join();
        debug_print("%s\n", "Processor thread stopped.");
    }

    /////////////////////////////////////////////////////////////////
    //
    // INSTANTIATION
    //
    /////////////////////////////////////////////////////////////////


    template class hells_state<event, event>;
    template class hells_join<event, event, joined_event>;
    template class hells_filter<event>;
    template class hells_aggregate<event,aggregated_event,float>;

    // Latency
    template class hells_state<event_latency, event_latency>;
    template class hells_join<event_latency, event_latency, joined_event_latency>;
    template class hells_filter<event_latency>;
}