//
// Created by koerberm on 30.05.18.
//

#ifndef EPHSA_DISPATCH_OPERATORS_H
#define EPHSA_DISPATCH_OPERATORS_H

#include <thread>
#include <functional>
#include <memory>

#include "event_queue.h"
#include "operators.h"

namespace ep {

    /////////////////////////////////////////////////////////////////
    //
    // HELLS FILTER
    //
    /////////////////////////////////////////////////////////////////

    typedef struct {
        const queue_def *in_q;
        long             in_start;
        uint32_t         batch_size;
        HELLS_FILTER_BITFIELD_TYPE *results;
    } hells_filter_args;

    typedef struct {
        queue_def in_q;
    } hells_filter_aux_data;

    template<class T>
    class hells_filter : public event_source<T>, public event_sink<T> {
    public:
        hells_filter( const std::shared_ptr <ep::event_queue<T>> &in_queue,
                      const std::shared_ptr <ep::event_queue<T>> &out_queue,
                      uint32_t min_batch_size,
                      uint32_t max_batch_size,
                      const hsa::hsa_platform &platform,
                      const hsa::hsa_agent &agent,
                      const hsa::hsa_executable &exe);
        ~hells_filter();
        void stop();

        uint64_t get_processing_latency();
    private:
        const hsa::hsa_platform &platform;
        const hsa::hsa_agent &agent;
        const hsa::exe_symbol_kernel kernel_symbol;

        std::vector<T> write_buffer;
        batch_writer<T> writer;

        const uint32_t min_batch_size;
        const uint32_t max_batch_size;

        uint64_t last_batch_time;
        uint64_t latency_accum;
        uint64_t latency_count;

        bool stopped;
        std::thread processor;

        void run();
        void probe( const HELLS_FILTER_BITFIELD_TYPE *results, uint32_t result_length, hsa_signal_value_t begin, hsa_signal_value_t end);
    };

    /////////////////////////////////////////////////////////////////
    //
    // HELLS AGGREGATE
    //
    /////////////////////////////////////////////////////////////////

    template<class TIN, class I_TYPE>
    struct hells_agg_init_args_s {
        const TIN *in;
        I_TYPE    *out;
        uint64_t   element_count;
        uint32_t   elements_per_item;
    };

    template<class I_TYPE>
    struct hells_agg_args_s {
        I_TYPE   *in;
        I_TYPE   *out;
        uint64_t  element_count;
        uint32_t  elements_per_item;
    };

    template<class TIN, class TOUT, class I_TYPE>
    class hells_aggregate : public event_source<TOUT>, public event_sink<TIN> {
        typedef hells_agg_init_args_s<TIN,I_TYPE> hells_agg_init_args;
        typedef hells_agg_args_s<I_TYPE> hells_agg_run_args;
    public:
        hells_aggregate(
                const std::shared_ptr<ep::event_queue<TIN>> &in_queue,
                const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                uint64_t  window_size,
                uint32_t  elements_per_item,
                const hsa::hsa_platform &platform,
                const hsa::hsa_agent &agent,
                const hsa::hsa_executable &exe,
                std::function<TOUT(const TIN&, const I_TYPE&)> out);

        hells_aggregate(
                const std::shared_ptr<ep::event_queue<TIN>> &in_queue,
                const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                count_window<TIN> &&window,
                uint32_t  elements_per_item,
                const hsa::hsa_platform &platform,
                const hsa::hsa_agent &agent,
                const hsa::hsa_executable &exe,
                std::function<TOUT(const TIN&, const I_TYPE&)> out);

        ~hells_aggregate();
        void stop();

        uint64_t get_processing_latency();
    private:
        const hsa::hsa_platform      &platform;
        const hsa::hsa_agent         &agent;
        const hsa::exe_symbol_kernel  kernel_symbol_init;
        const hsa::exe_symbol_kernel  kernel_symbol_run;

        count_window<TIN> window;
        const std::function<TOUT(const TIN&, const I_TYPE&)> out;
        const uint32_t  elements_per_item;

        uint64_t last_batch_time;
        uint64_t latency_accum;
        uint64_t latency_count;

        bool stopped;
        std::thread processor;

        void run();
    };

    /////////////////////////////////////////////////////////////////
    //
    // HELLS JOIN
    //
    /////////////////////////////////////////////////////////////////

    template<class TI1, class TI2>
    class hells_state {
        template <class T1, class T2, class TO> friend class hells_join;
    public:
        hells_state(uint32_t window_size, const hsa::hsa_agent &agent);
        hells_state(count_window<TI1> && window_l, count_window<TI2> && window_r, const hsa::hsa_agent &agent);
        hells_state(const hells_state &copy);
        hells_state(hells_state &&move) noexcept ;

        hells_state<TI1, TI2>& operator=( const hells_state<TI1, TI2> & copy );
        hells_state<TI1, TI2>& operator=( hells_state<TI1, TI2> && move ) noexcept ;

        void reset_results();

        template<class E1, class E2>
        friend std::ostream& operator<<(std::ostream& s, const hells_state<E1,E2>& a);
        count_window<TI1> window_left;
        count_window<TI2> window_right;
        std::unique_ptr<hsa::hsa_buffer> results;
        uint64_t  result_length;
        bool probe_right;
        TI1 left;
        TI2 right;
        window_def wdef;
    };

    typedef struct {
        bool        probe_right;
        void       *event;
        window_def *window;
        HELLS_JOIN_BITFIELD_TYPE *results;
    } hells_join_args;

    template<class TI1, class TI2, class TOUT>
    class hells_join : public event_source<TOUT>, public binary_event_sink<TI1,TI2> {
    public:
        hells_join(const std::shared_ptr<ep::event_queue<TI1>> &in_queue1,
                   const std::shared_ptr<ep::event_queue<TI2>> &in_queue2,
                   const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                   uint32_t window_size,
                   std::function<TOUT(const TI1&,const TI2&)> assemble,
                   const hsa::hsa_platform &platform,
                   const hsa::hsa_agent &agent,
                   const hsa::hsa_executable &exe);

        hells_join(const std::shared_ptr<ep::event_queue<TI1>> &in_queue1,
                   const std::shared_ptr<ep::event_queue<TI2>> &in_queue2,
                   const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                   count_window<TI1> &&window_l, count_window<TI2> &&window_r,
                   std::function<TOUT(const TI1&,const TI2&)> assemble,
                   const hsa::hsa_platform &platform,
                   const hsa::hsa_agent &agent,
                   const hsa::hsa_executable &exe);

        ~hells_join();
        void stop();

        uint64_t get_processing_latency();
    private:
        const hsa::hsa_platform &platform;
        const hsa::hsa_agent &agent;
        const hsa::exe_symbol_kernel kernel_symbol;
        const std::function<TOUT(const TI1&,const TI2&)> assemble;
        const uint32_t elements_per_thread;

//        hells_state<TI1,TI2> states[2];
        std::unique_ptr<hsa::hsa_buffer> states;

        std::vector<TOUT> write_buffer;
        batch_writer<TOUT> writer;

        uint8_t  flag;
        std::shared_ptr<hsa::hsa_signal> completion_signal;
        std::unique_ptr<hsa::hsa_queue>  queue;
        std::unique_ptr<hsa::hsa_buffer> kernargs;

        uint64_t last_batch_time;
        uint64_t latency_accum;
        uint64_t latency_count;

        bool stopped;
        std::unique_ptr<std::thread> processor;

        void run();
        void process_element_left( const TI1 &event);
        void process_element_right( const TI2 &event);
        void probe( hells_state<TI1,TI2> &state );
    };

}
#endif //EPHSA_DISPATCH_OPERATORS_H
