//
// Created by koerberm on 10.04.18.
//

#include "event.h"
#include "latency/event_latency.h"
#include "event_queue.h"
#include "ep_common.h"

namespace ep {

    template<class T>
    event_queue<T>::event_queue(uint32_t capacity, const hsa::hsa_platform &platform, const hsa::hsa_agent &agent) :
            capacity(capacity), mask(capacity-1),
            reader_w_idx(0),
            reader_r_idx(0),
            writer_w_idx(0),
            writer_r_idx(0),
            buffer(agent.get_fine_grained_region().allocate(capacity* sizeof(T),64)),
            r_signal(platform.create_signal(0)),
            w_signal(platform.create_signal(0)) {

        if ( (capacity & (capacity - 1)) != 0) {
            throw ep_exception("Queue size must be a power of two");
        }
    }

    template<class T>
    event_queue<T>::~event_queue() {
        debug_print("%s\n", "Destroying queue.");
    }

    // Write related

    template<class T>
    hsa_signal_value_t  event_queue<T>::add(const T &event) {
        wait_not_full();
        auto addr = buffer->address<T>();
        addr[writer_w_idx & mask] = event;
        w_signal->add_screlease(1);
        return writer_w_idx++;
    }

    template<class T>
    hsa_signal_value_t event_queue<T>::wait_not_full(uint64_t timeout) const {
        return wait_for_space(1,timeout);
    }

    template<class T>
    hsa_signal_value_t event_queue<T>::wait_for_space(uint32_t num_elems, uint64_t timeout) const {
        if ( num_elems > capacity )
            throw ep_exception("Cannot wait for more space than capacity.");

        const hsa_signal_value_t min_read_idx = writer_w_idx - capacity + num_elems;

        if ( writer_r_idx < min_read_idx ) {
            writer_r_idx = r_signal->wait_scacquire(HSA_SIGNAL_CONDITION_GTE, min_read_idx, timeout);
        }


        auto space = writer_r_idx + capacity - writer_w_idx;
        return (space < num_elems) ? -1 : space;
    }

    // Read related

    template<class T>
    const T &event_queue<T>::get(hsa_signal_value_t  idx) const {
        if ( idx < reader_r_idx )
            throw ep_exception("Queue access out of bounds");
        else if ( idx >= reader_w_idx && idx >= (reader_w_idx=w_signal->load_scacquire()) ) {
            throw ep_exception("Queue access out of bounds");
        }
        return buffer->address<T>()[idx&mask];
    }

    template<class T>
    hsa_signal_value_t event_queue<T>::wait_not_empty(uint64_t timeout) const {
        return wait_for(1, timeout);
    }

    template<class T>
    hsa_signal_value_t event_queue<T>::wait_for(uint32_t num_elems, uint64_t timeout) const {
        if ( num_elems > capacity )
            throw ep_exception("Cannot wait for more items than capacity.");

        auto tgt = reader_r_idx+num_elems;

        if ( reader_w_idx < tgt ) {
            reader_w_idx = w_signal->wait_scacquire(HSA_SIGNAL_CONDITION_GTE, tgt, timeout);
        }
        return (reader_w_idx < tgt ) ? -1 : reader_w_idx;
    }

    template<class T>
    void event_queue<T>::consume_until(hsa_signal_value_t  until) {
        if ( until < reader_r_idx )
            throw ep_exception("Elements already consumed.");
        else if ( until > reader_w_idx && until > (reader_w_idx=w_signal->load_scacquire()) )
            throw ep_exception("Cannot consume more elements than present.");

        r_signal->store_screlease(until);
        reader_r_idx = until;
    }

    template<class T>
    T *event_queue<T>::get_base() const {
        return buffer->address<T>();
    }

    template<class T>
    queue_def event_queue<T>::to_struct() const {
        return {buffer->address<T>(), mask, capacity, {r_signal->handle.handle}, {w_signal->handle.handle} };
    }

    template<class T>
    void event_queue<T>::set_write_index(hsa_signal_value_t value) {
        w_signal->store_screlease(value);
        reader_w_idx = value;
        writer_w_idx = value;
    }

    template<class T>
    void event_queue<T>::set_read_index(hsa_signal_value_t value) {
        r_signal->store_screlease(value);
        reader_r_idx = value;
        writer_r_idx = value;
    }

    template<class T>
    hsa_signal_value_t event_queue<T>::get_write_idx() const {
        return w_signal->load_scacquire();
    }

    template<class T>
    hsa_signal_value_t event_queue<T>::get_read_idx() const {
        return r_signal->load_scacquire();
    }

    template class event_queue<uint32_t>;
    template class event_queue<event>;
    template class event_queue<joined_event>;
    template class event_queue<aggregated_event>;

    // Latency
    template class event_queue<event_latency>;
    template class event_queue<joined_event_latency>;
}
