//
// Created by koerberm on 10.04.18.
//

#ifndef EPHSA_EVENT_QUEUE_H
#define EPHSA_EVENT_QUEUE_H


#include <cstdint>
#include <memory>
#include <atomic>
#include "../hsa_platform/hsa_platform.h"
#include "../rocm/shared.h"

namespace ep {

    template<class T>
    class event_queue {
        template <class T1> friend class batch_writer;
    public:
        event_queue(uint32_t capacity, const hsa::hsa_platform &platform, const hsa::hsa_agent &agent);

        virtual ~event_queue();

        virtual hsa_signal_value_t add(const T &event);

        const T &get(hsa_signal_value_t  idx) const;

        T* get_base() const;

        void consume_until(hsa_signal_value_t until);

        hsa_signal_value_t wait_not_empty(uint64_t timeout = UINT64_MAX) const;
        hsa_signal_value_t wait_for(uint32_t num_elems, uint64_t timeout = UINT64_MAX) const;

        hsa_signal_value_t wait_not_full(uint64_t timeout = UINT64_MAX) const;
        hsa_signal_value_t wait_for_space(uint32_t num_elems, uint64_t timeout = UINT64_MAX) const;

        hsa_signal_value_t get_read_idx() const;
        hsa_signal_value_t get_write_idx() const;

        void set_write_index(hsa_signal_value_t value);
        void set_read_index(hsa_signal_value_t value);

        queue_def to_struct() const;

        const uint32_t capacity;
        const uint32_t mask;
    protected:
        mutable hsa_signal_value_t reader_w_idx;
        mutable hsa_signal_value_t reader_r_idx;

        mutable hsa_signal_value_t writer_w_idx;
        mutable hsa_signal_value_t writer_r_idx;

        std::unique_ptr<hsa::hsa_buffer> buffer;
        std::shared_ptr<hsa::hsa_signal> r_signal;
        std::shared_ptr<hsa::hsa_signal> w_signal;
    };

    template<class TOUT>
    class batch_writer {
    public:
        static const hsa_signal_value_t flush_size = 64;

        explicit batch_writer(const std::shared_ptr<ep::event_queue<TOUT>> &queue) :
                queue(queue), sync(queue->get_write_idx()) {
        }

        void write_batch( std::vector<TOUT> &items, bool flush = false ) {
            if ( items.empty() || (items.size() < flush_size &&  !flush) )
                return;

            TOUT *base = queue->get_base();
            auto mask = static_cast<hsa_signal_value_t >(queue->mask);
            auto num_items = static_cast<hsa_signal_value_t >(items.size());
            auto begin = sync.fetch_add(num_items);

            hsa_signal_value_t w_result = -1;

            while ( w_result < 0 )
                w_result = queue->wait_for_space(num_items);

            for ( hsa_signal_value_t i = 0; i < num_items; i++ ) {
                base[(begin+i)&mask] = items[i];
            }

            // Commit writes
            queue->w_signal->wait_scacquire(HSA_SIGNAL_CONDITION_EQ, begin );
            queue->set_write_index(begin+num_items);
            items.clear();
        }

    private:
        std::shared_ptr<ep::event_queue<TOUT>> queue;
        std::atomic_long sync;
    };

}
#endif //EPHSA_EVENT_QUEUE_H


