//
// Created by koerberm on 30.05.18.
//

#include <thread>
#include "hsa_operators.h"
#include "ep_common.h"
#include "event.h"
#include "latency/event_latency.h"

namespace ep {
    template<class TIN, class TOUT, class KernArgs>
    hsa_unary_operator<TIN, TOUT, KernArgs>::hsa_unary_operator(const std::shared_ptr<event_queue<TIN>> &in_queue,
                                                                      const std::shared_ptr<event_queue<TOUT>> &out_queue,
                                                                      uint32_t min_batch_size,
                                                                      uint32_t max_batch_size,
                                                                      const hsa::hsa_platform &platform,
                                                                      const hsa::hsa_agent &agent,
                                                                      hsa::exe_symbol_kernel kernel_symbol)
            :   event_source<TOUT>(out_queue),
                event_sink<TIN>(in_queue),
                completion_signal(platform.create_signal(1)),
                queue(agent.create_queue(agent.queue_max_size,HSA_QUEUE_TYPE_SINGLE)),
                kernargs(agent.get_kernarg_region().allocate(kernel_symbol.kernarg_segment_size)),
                aux_data(agent.get_fine_grained_region().allocate(sizeof(hsa_unary_operator_args),64)),
                stopped(false) {

        uint32_t wg_size = std::min<uint32_t>(agent.wave_front_size,max_batch_size);

        int64_t num_groups = max_batch_size / wg_size;
        if ( max_batch_size % wg_size != 0 ) num_groups++;

        debug_print("Unary operator: workgroup size %u, #workgroups: %ld\n", wg_size, num_groups);
        auto aux_addr = aux_data->template address<hsa_unary_operator_args>();
        auto arg_addr = kernargs->address<KernArgs>();

        aux_addr->in_q  = in_queue->to_struct();
        aux_addr->out_q = out_queue->to_struct();

        arg_addr->in_q = &aux_addr->in_q;
        arg_addr->out_q = &aux_addr->out_q;
        arg_addr->dbg_val = aux_addr->dbg;
        arg_addr->min_batch_size = min_batch_size;
        hsa::hsa_kernel_dispatch_packet p(1,kernel_symbol,*completion_signal,*kernargs);
        p.set_grid_size(static_cast<uint32_t>(num_groups)*wg_size);
        p.set_workgroup_size(static_cast<uint16_t>(wg_size));
        debug_print("%s\n", "Dispatching kernel");
        queue->dispatch_kernel_packet(p);
        debug_print("%s\n", "Finished dispatching kernel");
    }

    template<class TIN, class TOUT, class KernArgs>
    hsa_unary_operator<TIN, TOUT, KernArgs>::~hsa_unary_operator() {
        debug_print("%s\n", "Destroying op signal.");
        stop();
    }

    template<class TIN, class TOUT, class KernArgs>
    void hsa_unary_operator<TIN, TOUT, KernArgs>::stop() {
        if ( stopped )
            return;

        stopped = true;
        debug_print("%s\n", "Setting write index to term value");
        auto w = this->in_queue->get_write_idx();

        auto &aux = *aux_data->template address<hsa_unary_operator_args>();
        debug_print("Debug: (%ld,%ld)\n", aux.dbg[0], aux.dbg[1]);

        this->in_queue->set_write_index(QUEUE_TERM_IDX);
        debug_print("%s\n", "Waiting for kernel completion");
        this->completion_signal->wait_scacquire(HSA_SIGNAL_CONDITION_LT, 1);
        this->in_queue->set_write_index(w);
    }


    ////// HSA Filter Single Wavefront
    template<class T>
    hsa_filter_wf<T>::hsa_filter_wf(const std::shared_ptr<event_queue<T>> &in_queue,
                                    const std::shared_ptr<event_queue<T>> &out_queue,
                                    uint32_t min_batch_size,
                                    uint32_t max_batch_size,
                                    const hsa::hsa_platform &platform, const hsa::hsa_agent &agent,
                                    const hsa::hsa_executable &exe)
            : hsa_unary_operator<T,T,hsa_filter_wf_args>(in_queue, out_queue,
                                                            std::min(min_batch_size,agent.wave_front_size),
                                                            std::min(max_batch_size,agent.wave_front_size),
                                                            platform, agent,
                                                            exe.kernel_by_name("hsa_filter_wf") ) {

    }

    ///////////////// HSA FILTER ///////////////////////
    template<class T>
    hsa_filter<T>::hsa_filter(const std::shared_ptr<event_queue<T>> &in_queue,
                                      const std::shared_ptr<event_queue<T>> &out_queue,
                                      int64_t min_batch_size,
                                      int64_t max_batch_size,
                                      const hsa::hsa_platform &platform,
                                      const hsa::hsa_agent &agent,
                                      hsa::hsa_executable & exe) :
            event_source<T>(out_queue),
            event_sink<T>(in_queue),

            sync_1(platform.create_signal(0)),
            sync_2(platform.create_signal(0)),
            completion_signal_switch(platform.create_signal(1)),
            completion_signal_proc(platform.create_signal(1)),
            queue(agent.create_queue(agent.queue_max_size,HSA_QUEUE_TYPE_SINGLE)),
            stopped(false) {

        uint64_t num_longs_per_batch = (max_batch_size + (sizeof(uint64_t)<< 3) -1 ) / (sizeof(uint64_t)<< 3);

        results1 = agent.get_fine_grained_region().allocate(num_longs_per_batch* sizeof(uint64_t),64);
        results2 = agent.get_fine_grained_region().allocate(num_longs_per_batch* sizeof(uint64_t),64);
        aux_data = agent.get_fine_grained_region().allocate(sizeof(hsa_filter_aux_data),64);

        results1->clear();
        results2->clear();
        aux_data->clear();

        auto switch_kernel = exe.kernel_by_name("hsa_filter_switch");
        auto proc_kernel   = exe.kernel_by_name("hsa_filter_proc");

        kernargs_switch = agent.get_kernarg_region().allocate(switch_kernel.kernarg_segment_size);
        kernargs_proc   = agent.get_kernarg_region().allocate(proc_kernel.kernarg_segment_size);

        auto switch_argaddr = kernargs_switch->address<hsa_filter_switch_args>();
        auto proc_argaddr   = kernargs_proc->address<hsa_filter_proc_args>();
        auto aux_addr       = aux_data->address<hsa_filter_aux_data>();

        // Fill aux-data
        aux_addr->latency_accum     = 0;
        aux_addr->ctrl              = 0;
        aux_addr->dbg_val_proc[0]   = aux_addr->dbg_val_proc[1] = 0;
        aux_addr->dbg_val_switch[0] = aux_addr->dbg_val_switch[1] = 0;
        aux_addr->in_q              = in_queue->to_struct();
        aux_addr->out_q             = out_queue->to_struct();



        uint32_t wg_size = agent.wave_front_size;

        uint32_t elems_per_thread = sizeof(HELLS_FILTER_BITFIELD_TYPE) << 3;
        uint32_t num_threads = (max_batch_size + elems_per_thread -1) / elems_per_thread;

        int64_t num_groups = std::min(22u, (num_threads + wg_size - 1) / wg_size);



        switch_argaddr->in_q           = &aux_addr->in_q;
        switch_argaddr->out_q          = &aux_addr->out_q;
        switch_argaddr->min_batch_size =  min_batch_size;
        switch_argaddr->max_batch_size =  max_batch_size;
        switch_argaddr->num_groups     =  num_groups;
        switch_argaddr->sync_1         =  sync_1->handle;
        switch_argaddr->sync_2         =  sync_2->handle;
        switch_argaddr->result1        = results1->template address<uint64_t>();
        switch_argaddr->result2        = results2->template address<uint64_t>();
        switch_argaddr->ctrl           = &aux_addr->ctrl;
        switch_argaddr->latency_accum  = &aux_addr->latency_accum;
        switch_argaddr->dbg_val        = aux_addr->dbg_val_switch;

        hsa::hsa_kernel_dispatch_packet ps(1,switch_kernel,*completion_signal_switch,*kernargs_switch);
        ps.set_grid_size(agent.wave_front_size);
        ps.set_workgroup_size(agent.wave_front_size);

        proc_argaddr->in_q    = &aux_addr->in_q;
        proc_argaddr->sync_1  = sync_1->handle;
        proc_argaddr->sync_2  = sync_2->handle;
        proc_argaddr->result1 = results1->template address<HELLS_FILTER_BITFIELD_TYPE>();
        proc_argaddr->result2 = results2->template address<HELLS_FILTER_BITFIELD_TYPE>();
        proc_argaddr->ctrl    = &aux_addr->ctrl;
        proc_argaddr->dbg_val = aux_addr->dbg_val_proc;

        hsa::hsa_kernel_dispatch_packet pp(1,proc_kernel,*completion_signal_proc,*kernargs_proc);
        pp.set_grid_size(static_cast<uint32_t>(num_groups)*wg_size);

        if ( out_queue->capacity < agent.wave_front_size ) {
            throw ep_exception("Capacity of output-queue for dispatch_filter must be at least as big as the size of a wavefront.");
        }

        debug_print("Max batch size: %li, workgroup size %u, #workgroups: %li\n", max_batch_size, wg_size, num_groups);
        pp.set_workgroup_size(static_cast<uint16_t>(wg_size));

        debug_print("%s\n", "Dispatching kernel dispatch_filter switch");
        queue->dispatch_kernel_packet(ps);
        debug_print("%s\n", "Dispatching kernel dispatch_filter proc");
        queue->dispatch_kernel_packet(pp);
        debug_print("%s\n", "Finished dispatching kernels");
    }

    template<class T>
    hsa_filter<T>::~hsa_filter() {
        debug_print("%s\n", "Destroying HSA Filter.");
        stop();
    }

    template<class T>
    void hsa_filter<T>::stop() {
        if ( stopped )
            return;
        debug_print("%s\n", "Stopping HSA Filter.");
        stopped = true;
        auto w = this->in_queue->get_write_idx();

        auto &aux = *aux_data->template address<hsa_filter_aux_data>();


        debug_print("Sync Signals: (%ld,%ld)\n", sync_1->load_scacquire(), sync_2->load_scacquire());
        debug_print("Debug Switch: (%ld,%ld)\n", aux.dbg_val_switch[0], aux.dbg_val_switch[1]);
        debug_print("Debug Proc  : (%ld,%ld)\n", aux.dbg_val_proc[0], aux.dbg_val_proc[1]);

        this->in_queue->set_write_index(QUEUE_TERM_IDX);

        debug_print("%s\n", "Waiting for switch kernel completion");
        this->completion_signal_switch->wait_scacquire(HSA_SIGNAL_CONDITION_LT, 1);
        debug_print("%s\n", "Waiting for proc kernel completion");
        this->completion_signal_proc->wait_scacquire(HSA_SIGNAL_CONDITION_LT, 1);
        debug_print("%s\n", "Finished stopping HSA Filter.");
        this->in_queue->set_write_index(w);
    }

    ////// HSA Join

    template<class TI1, class TI2, class TOUT>
    hsa_join<TI1, TI2, TOUT>::hsa_join(const std::shared_ptr<event_queue<TI1>> &in_queue1,
                                             const std::shared_ptr<event_queue<TI2>> &in_queue2,
                                             const std::shared_ptr<event_queue<TOUT>> &out_queue,
                                             count_window<TI1> &&window_l, count_window<TI2> &&window_r,
                                             const hsa::hsa_platform &platform, const hsa::hsa_agent &agent,
                                             const hsa::hsa_executable &exe) :
        event_source<TOUT>(out_queue),
        binary_event_sink<TI1,TI2>(in_queue1,in_queue2),

        window_left(std::move(window_l)),
        window_right(std::move(window_r)),

        sync_1(platform.create_signal(0)),
        sync_2(platform.create_signal(0)),

        completion_signal_switch(platform.create_signal(1)),
        completion_signal_proc(platform.create_signal(1)),
        queue(agent.create_queue(agent.queue_max_size,HSA_QUEUE_TYPE_SINGLE)),

        stopped(false) {

            auto kernel_switch = exe.kernel_by_name("hsa_join_switch");
            auto kernel_proc   = exe.kernel_by_name("hsa_join_proc");
            kernargs_switch    = agent.get_kernarg_region().allocate(kernel_switch.kernarg_segment_size);
            kernargs_proc      = agent.get_kernarg_region().allocate(kernel_proc.kernarg_segment_size);
            aux_data           = agent.get_fine_grained_region().allocate(sizeof(hsa_join_aux_data),64);

            auto switch_argaddr = kernargs_switch->address<hsa_join_switch_args>();
            auto proc_argaddr = kernargs_proc->address<hsa_join_proc_args>();
            auto aux_addr = aux_data->template address<hsa_join_aux_data>();

            // Compute group and grid size

            // TODO: This seems to be optimal
            uint32_t num_groups = 10;
            uint32_t group_size = agent.wave_front_size;
//            uint64_t max_window = std::max(window_l.get_capacity(),window_r.get_capacity());
//            uint32_t num_groups = agent.compute_units-1;
//            uint32_t group_size = agent.wave_front_size;
//
//            while ( num_groups * group_size < max_window && group_size < 128 ) {
//                group_size = group_size + agent.wave_front_size;
//            }

            debug_print("Using grid-config (#groups/group_size): (%u,%u)\n", num_groups, group_size);


            // Prepare aux data
            aux_addr->latency_accum = 0;
            aux_addr->ctrl          = 0;
            aux_addr->w_idx         = out_queue->get_write_idx();
            aux_addr->in_q1         = in_queue1->to_struct();
            aux_addr->in_q2         = in_queue2->to_struct();
            aux_addr->out_q         = out_queue->to_struct();
            aux_addr->w_def1        = window_left.to_struct();
            aux_addr->w_def2        = window_right.to_struct();

            // Prepare switch args
            switch_argaddr->in_q1         = &aux_addr->in_q1;
            switch_argaddr->in_q2         = &aux_addr->in_q2;
            switch_argaddr->w_1           = &aux_addr->w_def1;
            switch_argaddr->w_2           = &aux_addr->w_def2;
            switch_argaddr->sync_1        =  sync_1->handle;
            switch_argaddr->sync_2        =  sync_2->handle;
            switch_argaddr->ctrl          = &aux_addr->ctrl;
            switch_argaddr->num_groups    =  num_groups;
            switch_argaddr->latency_accum = &aux_addr->latency_accum;
            switch_argaddr->dbg           =  aux_addr->dbg_val_switch;

            // Prepare proc args
            proc_argaddr->w_1    = &aux_addr->w_def1;
            proc_argaddr->w_2    = &aux_addr->w_def2;
            proc_argaddr->out_q  = &aux_addr->out_q;
            proc_argaddr->sync_1 =  sync_1->handle;
            proc_argaddr->sync_2 =  sync_2->handle;
            proc_argaddr->ctrl   = &aux_addr->ctrl;
            proc_argaddr->w_idx  = &aux_addr->w_idx;
            proc_argaddr->dbg    =  aux_addr->dbg_val_proc;


            hsa::hsa_kernel_dispatch_packet dispatch_switch(1,kernel_switch,*completion_signal_switch,*kernargs_switch);
            dispatch_switch.set_grid_size(1);
            dispatch_switch.set_workgroup_size(1);


            hsa::hsa_kernel_dispatch_packet dispatch_proc(1,kernel_proc,*completion_signal_proc,*kernargs_proc);
            dispatch_proc.set_grid_size(num_groups*group_size);
            dispatch_proc.set_workgroup_size(static_cast<uint16_t >(group_size));

            if ( out_queue->capacity < agent.wave_front_size ) {
                throw ep_exception("Capacity of output-queue for join must be at least as big as group size.");
            }

            debug_print("Window-Size: %lu, workgroup size %u, #workgroups: %u\n", window_l.get_capacity(), group_size, num_groups);
            debug_print("%s\n", "Dispatching kernel join switch");
            queue->dispatch_kernel_packet(dispatch_switch);
            debug_print("%s\n", "Dispatching kernel join proc");
            queue->dispatch_kernel_packet(dispatch_proc);
            debug_print("%s\n", "Finished dispatching kernels");
    }

    template<class TI1, class TI2, class TOUT>
    hsa_join<TI1, TI2, TOUT>::hsa_join(const std::shared_ptr<event_queue<TI1>> &in_queue1,
                                             const std::shared_ptr<event_queue<TI2>> &in_queue2,
                                             const std::shared_ptr<event_queue<TOUT>> &out_queue,
                                             uint32_t window_size, const hsa::hsa_platform &platform,
                                             const hsa::hsa_agent &agent, const hsa::hsa_executable &exe) :
            hsa_join(in_queue1,in_queue2,out_queue,
                        count_window<TI1>(window_size,agent), count_window<TI2>(window_size,agent),
                        platform,agent,exe) {
    }



    template<class TI1, class TI2, class TOUT>
    hsa_join<TI1, TI2, TOUT>::~hsa_join() {
        debug_print("%s\n", "Destroying join signal.");
        stop();
    }

    template<class TI1, class TI2, class TOUT>
    void hsa_join<TI1, TI2, TOUT>::stop() {
        if ( stopped )
            return;
        debug_print("%s\n", "Stopping HSA join.");
        stopped = true;
        auto aux_addr = aux_data->template address<hsa_join_aux_data>();
        auto w1       = this->in_queue1->get_write_idx();
        auto w2       = this->in_queue2->get_write_idx();


        debug_print("Sync Signals: (%ld,%ld)\n", sync_1->load_scacquire(), sync_2->load_scacquire());
        debug_print("Debug Switch: (%ld,%ld)\n", aux_addr->dbg_val_switch[0], aux_addr->dbg_val_switch[1]);
        debug_print("Debug Proc  : (%ld,%ld)\n", aux_addr->dbg_val_proc[0], aux_addr->dbg_val_proc[1]);

        this->in_queue1->set_write_index(QUEUE_TERM_IDX);
        this->in_queue2->set_write_index(QUEUE_TERM_IDX);

        debug_print("%s\n", "Waiting for switch kernel completion");
        this->completion_signal_switch->wait_scacquire(HSA_SIGNAL_CONDITION_LT, 1);
        debug_print("%s\n", "Waiting for proc kernel completion");
        this->completion_signal_proc->wait_scacquire(HSA_SIGNAL_CONDITION_LT, 1);
        debug_print("%s\n", "Finished stopping HSA join.");
        this->in_queue1->set_write_index(w1);
        this->in_queue2->set_write_index(w2);
        // TODO: Write-back window inserts!
    }


    //////////////////////////////////////////////////////////////7

    template<class TIN, class TOUT, class I_TYPE>
    hsa_aggregate<TIN, TOUT, I_TYPE>::hsa_aggregate(
            const std::shared_ptr<event_queue<TIN>> &in_queue,
            const std::shared_ptr<event_queue<TOUT>> &out_queue,
            uint64_t  window_size,
            const hsa::hsa_platform &platform,
            const hsa::hsa_agent &agent,
            const hsa::hsa_executable &exe) :
        hsa_aggregate<TIN,TOUT,I_TYPE>(
            in_queue,
            out_queue,
            count_window<TIN>(window_size,agent),
            platform, agent, exe) {
    }


    template<class TIN, class TOUT, class I_TYPE>
    hsa_aggregate<TIN, TOUT, I_TYPE>::hsa_aggregate(
        const std::shared_ptr<event_queue<TIN>> &in_queue,
        const std::shared_ptr<event_queue<TOUT>> &out_queue,
        count_window<TIN> &&window,
        const hsa::hsa_platform &platform,
        const hsa::hsa_agent &agent,
        const hsa::hsa_executable &exe) :

        event_source<TOUT>(out_queue),
        event_sink<TIN>(in_queue),

        window(std::move(window)),

        sync_1(platform.create_signal(0)),
        sync_2(platform.create_signal(0)),
        i_sync(platform.create_signal(0)),

        completion_signal_switch(platform.create_signal(1)),
        completion_signal_proc(platform.create_signal(1)),
        queue(agent.create_queue(agent.queue_max_size,HSA_QUEUE_TYPE_SINGLE)),
        stopped(false) {

        auto kernel_switch = exe.kernel_by_name("hsa_aggregate_switch");
        auto kernel_proc   = exe.kernel_by_name("hsa_aggregate_proc");
        kernargs_switch    = agent.get_kernarg_region().allocate(kernel_switch.kernarg_segment_size);
        kernargs_proc      = agent.get_kernarg_region().allocate(kernel_proc.kernarg_segment_size);
        aux_data           = agent.get_fine_grained_region().allocate(sizeof(hsa_aggregate_aux_data),64);

        auto switch_argaddr = kernargs_switch->address<hsa_aggregate_switch_args>();
        auto proc_argaddr   = kernargs_proc->address<hsa_aggregate_proc_args>();
        auto aux_addr       = aux_data->template address<hsa_aggregate_aux_data>();

        uint32_t num_groups = agent.compute_units-1;
        uint32_t group_size = agent.wave_front_size;

        while ( group_size < agent.workgroup_max_size &&
                num_groups * group_size < window.get_capacity() &&
                (num_groups-1) * (group_size<<1) < window.get_capacity() ) {
            group_size = group_size << 1;
        }

        num_groups = agent.compute_units-1;
        group_size = agent.wave_front_size;

        debug_print("Using grid-config (#groups/group_size): (%u,%u)\n", num_groups, group_size);

        auto grid_size = num_groups * group_size;

        // Compute capacity of output buffers
        global_accum = agent.get_fine_grained_region().allocate( num_groups * sizeof(I_TYPE) );

        // Prepare aux data
        aux_addr->in_q          = this->in_queue->to_struct();
        aux_addr->out_q         = this->out_queue->to_struct();
        aux_addr->window        = this->window.to_struct();
        aux_addr->latency_accum = 0;

        // Prepare switch args
        switch_argaddr->in_q          = &aux_addr->in_q;
        switch_argaddr->window        = &aux_addr->window;
        switch_argaddr->sync_1        =  sync_1->handle;
        switch_argaddr->sync_2        =  sync_2->handle;
        switch_argaddr->num_groups    =  num_groups;
        switch_argaddr->latency_accum = &aux_addr->latency_accum;
        switch_argaddr->dbg           =  aux_addr->dbg_val_switch;

        // Prepare proc args
        proc_argaddr->out_q        = &aux_addr->out_q;
        proc_argaddr->window       = &aux_addr->window;
        proc_argaddr->sync_1       =  sync_1->handle;
        proc_argaddr->sync_2       =  sync_2->handle;
        proc_argaddr->i_sync       =  i_sync->handle;
        proc_argaddr->global_accum =  global_accum->template address<I_TYPE>();
        proc_argaddr->dbg          =  aux_addr->dbg_val_proc;

        hsa::hsa_kernel_dispatch_packet switch_dispatch(1,kernel_switch,*completion_signal_switch,*kernargs_switch);
        switch_dispatch.set_grid_size(1);
        switch_dispatch.set_workgroup_size(1);


        hsa::hsa_kernel_dispatch_packet proc_dispatch(1,kernel_proc,*completion_signal_proc,*kernargs_proc);
        proc_dispatch.set_grid_size(static_cast<uint32_t >(grid_size));
        proc_dispatch.set_workgroup_size(static_cast<uint16_t >(group_size));

        debug_print("Window-Size: %lu, workgroup size %u, #workgroups: %u\n", window.get_capacity(), group_size, num_groups);
        debug_print("%s\n", "Dispatching kernel join switch");
        queue->dispatch_kernel_packet(switch_dispatch);
        debug_print("%s\n", "Dispatching kernel join proc");
        queue->dispatch_kernel_packet(proc_dispatch);
        debug_print("%s\n", "Finished dispatching kernels");
    }



    template<class TIN, class TOUT, class I_TYPE>
    hsa_aggregate<TIN, TOUT, I_TYPE>::~hsa_aggregate() {
        debug_print("%s\n", "Destroying HSA aggregate.");
        stop();
    }

    template<class TIN, class TOUT, class I_TYPE>
    void hsa_aggregate<TIN, TOUT, I_TYPE>::stop() {
        if ( stopped )
            return;
        debug_print("%s\n", "Stopping HSA aggregate.");
        stopped = true;
        auto aux_addr = this->aux_data->template address<hsa_aggregate_aux_data>();
        auto w1       = this->in_queue->get_write_idx();

        this->in_queue->set_write_index(QUEUE_TERM_IDX);
        std::this_thread::sleep_for(std::chrono::milliseconds(100));

        debug_print("Sync Signals: (%ld,%ld)\n", sync_1->load_scacquire(), sync_2->load_scacquire());
        debug_print("ISync Signal: (%ld)\n",     i_sync->load_scacquire());
        debug_print("Debug Switch: (%ld,%ld)\n", aux_addr->dbg_val_switch[0], aux_addr->dbg_val_switch[1]);
        debug_print("Debug Proc  : (%ld,%ld)\n", aux_addr->dbg_val_proc[0], aux_addr->dbg_val_proc[1]);

        debug_print("%s\n", "Waiting for switch kernel completion");
        this->completion_signal_switch->wait_scacquire(HSA_SIGNAL_CONDITION_LT, 1);
        debug_print("%s\n", "Waiting for proc kernel completion");
        this->completion_signal_proc->wait_scacquire(HSA_SIGNAL_CONDITION_LT, 1);
        debug_print("%s\n", "Finished stopping HSA aggregate.");
        this->in_queue->set_write_index(w1);
        // TODO: Write-back window inserts!
    }

    template class hsa_filter_wf<event>;
    template class hsa_filter<event>;
    template class hsa_join<event,event,joined_event>;
    template class hsa_aggregate<event,aggregated_event,float>;

    template class hsa_filter_wf<event_latency>;
    template class hsa_filter<event_latency>;
    template class hsa_join<event_latency,event_latency,joined_event_latency>;
}