//
// Created by koerberm on 30.05.18.
//

#ifndef EPHSA_HSA_OPERATORS_H
#define EPHSA_HSA_OPERATORS_H

#include "operators.h"

namespace ep {

    typedef struct {
        queue_def in_q;
        queue_def out_q;
        int64_t   dbg[2];

    } hsa_unary_operator_args;

    template<class TIN, class TOUT, class KernArgs>
    class hsa_unary_operator : public event_source<TOUT>, public event_sink<TIN> {
    public:
        explicit hsa_unary_operator(const std::shared_ptr<ep::event_queue<TIN>> &in_queue,
                                    const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                                    uint32_t min_batch_size,
                                    uint32_t max_batch_size,
                                    const hsa::hsa_platform &platform,
                                    const hsa::hsa_agent &agent,
                                    hsa::exe_symbol_kernel kernel_symbol);
        virtual ~hsa_unary_operator();
        void stop();
    private:
        std::shared_ptr<hsa::hsa_signal> completion_signal;
        std::unique_ptr<hsa::hsa_queue> queue;
        std::unique_ptr<hsa::hsa_buffer> kernargs;
        std::unique_ptr<hsa::hsa_buffer> aux_data;
        bool stopped;
    };

    /////////////////////////////////////////////////////////////////
    //
    // HSA FILTER SINGLE WAVEFRONT
    //
    /////////////////////////////////////////////////////////////////


    typedef struct {
        queue_def *in_q;
        queue_def *out_q;
        long min_batch_size;
        int64_t *dbg_val;
    } hsa_filter_wf_args;

    template<class T>
    class hsa_filter_wf : public hsa_unary_operator<T,T, hsa_filter_wf_args> {
    public:
        hsa_filter_wf(  const std::shared_ptr<ep::event_queue<T>> &in_queue,
                        const std::shared_ptr<ep::event_queue<T>> &out_queue,
                        uint32_t min_batch_size,
                        uint32_t max_batch_size,
                        const hsa::hsa_platform &platform,
                        const hsa::hsa_agent &agent,
                        const hsa::hsa_executable &exe);
        uint64_t get_processing_latency() { return 0; };
    };

    /////////////////////////////////////////////////////////////////
    //
    // HSA FILTER
    //
    /////////////////////////////////////////////////////////////////

    typedef struct {
        queue_def    *in_q;
        queue_def    *out_q;
        int64_t       min_batch_size;
        int64_t       max_batch_size;
        int64_t       num_groups;
        hsa_signal_t  sync_1;
        hsa_signal_t  sync_2;
        uint64_t     *result1;
        uint64_t     *result2;
        int64_t      *ctrl;
        uint64_t     *latency_accum;
        int64_t      *dbg_val;
    } hsa_filter_switch_args;

    typedef struct {
        queue_def    *in_q;
        hsa_signal_t  sync_1;
        hsa_signal_t  sync_2;
        HELLS_FILTER_BITFIELD_TYPE     *result1;
        HELLS_FILTER_BITFIELD_TYPE     *result2;
        int64_t      *ctrl;
        int64_t      *dbg_val;
    } hsa_filter_proc_args;

    typedef struct {
        queue_def in_q;
        queue_def out_q;
        int64_t   ctrl;
        uint64_t  latency_accum;
        int64_t   dbg_val_switch[2];
        int64_t   dbg_val_proc[2];
    } hsa_filter_aux_data;

    template<class T>
    class hsa_filter :  public event_source<T>, public event_sink<T> {
    public:
        hsa_filter( const std::shared_ptr<ep::event_queue<T>> &in_queue,
                    const std::shared_ptr<ep::event_queue<T>> &out_queue,
                    int64_t min_batch_size,
                    int64_t max_batch_size,
                    const hsa::hsa_platform &platform,
                    const hsa::hsa_agent &agent,
                    hsa::hsa_executable & exe );
        ~hsa_filter();

        void stop();

        uint64_t get_processing_latency() { return aux_data->template address<hsa_filter_aux_data>()->latency_accum; };
    private:
        std::unique_ptr<hsa::hsa_buffer> results1;
        std::unique_ptr<hsa::hsa_buffer> results2;


        std::shared_ptr<hsa::hsa_signal> sync_1;
        std::shared_ptr<hsa::hsa_signal> sync_2;

        std::shared_ptr<hsa::hsa_signal> completion_signal_switch;
        std::shared_ptr<hsa::hsa_signal> completion_signal_proc;
        std::unique_ptr<hsa::hsa_buffer> kernargs_switch;
        std::unique_ptr<hsa::hsa_buffer> kernargs_proc;
        std::unique_ptr<hsa::hsa_queue>  queue;

        std::unique_ptr<hsa::hsa_buffer> aux_data;
        bool stopped;
    };


    /////////////////////////////////////////////////////////////////
    //
    // HSA JOIN
    //
    /////////////////////////////////////////////////////////////////

    typedef struct {
        queue_def    *in_q1;
        queue_def    *in_q2;
        window_def   *w_1;
        window_def   *w_2;
        hsa_signal_t  sync_1;
        hsa_signal_t  sync_2;
        int64_t      *ctrl;
        int64_t       num_groups;
        uint64_t     *latency_accum;
        int64_t      *dbg;
    } hsa_join_switch_args;

    typedef struct {
        window_def *w_1;
        window_def *w_2;
        queue_def  *out_q;
        hsa_signal_t sync_1;
        hsa_signal_t sync_2;
        int64_t *ctrl;
        int64_t *w_idx;
        int64_t *dbg;
    } hsa_join_proc_args;

    typedef struct {
        queue_def  in_q1;
        queue_def  in_q2;
        queue_def  out_q;
        window_def w_def1;
        window_def w_def2;
        int64_t    ctrl;
        int64_t    w_idx;
        uint64_t   latency_accum;
        int64_t    dbg_val_switch[2];
        int64_t    dbg_val_proc[2];
    } hsa_join_aux_data;

    template<class TI1, class TI2, class TOUT>
    class hsa_join :  public event_source<TOUT>, public binary_event_sink<TI1, TI2> {
    public:
        hsa_join( const std::shared_ptr<ep::event_queue<TI1>> &in_queue1,
                  const std::shared_ptr<ep::event_queue<TI2>> &in_queue2,
                  const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                  uint32_t window_size,
                  const hsa::hsa_platform &platform,
                  const hsa::hsa_agent &agent,
                  const hsa::hsa_executable &exe);

        hsa_join( const std::shared_ptr<ep::event_queue<TI1>> &in_queue1,
                  const std::shared_ptr<ep::event_queue<TI2>> &in_queue2,
                  const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                  count_window<TI1> &&window_l, count_window<TI2> &&window_r,
                  const hsa::hsa_platform &platform,
                  const hsa::hsa_agent &agent,
                  const hsa::hsa_executable &exe);

        ~hsa_join();

        void stop();

        uint64_t get_processing_latency() { return aux_data->template address<hsa_join_aux_data>()->latency_accum; };
    private:
        count_window<TI1> window_left;
        count_window<TI2> window_right;

        std::shared_ptr<hsa::hsa_signal> sync_1;
        std::shared_ptr<hsa::hsa_signal> sync_2;

        std::shared_ptr<hsa::hsa_signal> completion_signal_switch;
        std::shared_ptr<hsa::hsa_signal> completion_signal_proc;
        std::unique_ptr<hsa::hsa_buffer> kernargs_switch;
        std::unique_ptr<hsa::hsa_buffer> kernargs_proc;
        std::unique_ptr<hsa::hsa_queue> queue;

        std::unique_ptr<hsa::hsa_buffer> aux_data;

        bool stopped;
    };

    /////////////////////////////////////////////////////////////////
    //
    // HSA AGGREGATE
    //
    /////////////////////////////////////////////////////////////////

    template<class I_TYPE>
    struct signal_aggregate_switch_args_s {
        queue_def    *in_q;
        window_def   *window;
        hsa_signal_t  sync_1;
        hsa_signal_t  sync_2;
        int64_t       num_groups;
        uint64_t     *latency_accum;
        int64_t      *dbg;
    } ;

    template<class I_TYPE>
    struct signal_aggregate_proc_args_s {
        queue_def    *out_q;
        window_def   *window;
        hsa_signal_t  sync_1;
        hsa_signal_t  sync_2;
        hsa_signal_t  i_sync;
        I_TYPE       *global_accum;
        int64_t      *dbg;
    };

    typedef struct  {
        queue_def  in_q;
        queue_def  out_q;
        window_def window;
        uint64_t   latency_accum;
        int64_t    dbg_val_switch[2];
        int64_t    dbg_val_proc[2];
    } hsa_aggregate_aux_data;

    template<class TIN, class TOUT, class I_TYPE>
    class hsa_aggregate :  public event_source<TOUT>, public event_sink<TIN> {
        typedef signal_aggregate_switch_args_s<I_TYPE> hsa_aggregate_switch_args;
        typedef signal_aggregate_proc_args_s<I_TYPE> hsa_aggregate_proc_args;
    public:
        hsa_aggregate(  const std::shared_ptr<ep::event_queue<TIN>> &in_queue,
                        const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                        uint64_t window_size,
                        const hsa::hsa_platform &platform,
                        const hsa::hsa_agent &agent,
                        const hsa::hsa_executable &exe);

        hsa_aggregate(  const std::shared_ptr<ep::event_queue<TIN>> &in_queue,
                        const std::shared_ptr<ep::event_queue<TOUT>> &out_queue,
                        count_window<TIN> &&window,
                        const hsa::hsa_platform &platform,
                        const hsa::hsa_agent &agent,
                        const hsa::hsa_executable &exe);

        ~hsa_aggregate();

        void stop();

        uint64_t get_processing_latency() { return aux_data->template address<hsa_aggregate_aux_data>()->latency_accum; };
    private:
        count_window<TIN> window;

        std::shared_ptr<hsa::hsa_signal> sync_1;
        std::shared_ptr<hsa::hsa_signal> sync_2;
        std::shared_ptr<hsa::hsa_signal> i_sync;

        std::unique_ptr<hsa::hsa_buffer> global_accum;

        std::shared_ptr<hsa::hsa_signal> completion_signal_switch;
        std::shared_ptr<hsa::hsa_signal> completion_signal_proc;
        std::unique_ptr<hsa::hsa_queue> queue;

        std::unique_ptr<hsa::hsa_buffer> kernargs_switch;
        std::unique_ptr<hsa::hsa_buffer> kernargs_proc;

        std::unique_ptr<hsa::hsa_buffer> aux_data;

        bool stopped;
    };
}

#endif //EPHSA_HSA_OPERATORS_H
