//
// Created by koerberm on 09.03.18.
//

#include "operators.h"
#include "ep_common.h"
#include "event.h"
#include "latency/event_latency.h"


namespace ep {

    template<class T>
    count_window<T>::count_window(uint64_t capacity, const hsa::hsa_agent &agent) :
            buffer(agent.get_fine_grained_region().allocate(capacity* sizeof(T), 64)),
            capacity(capacity), inserts(0) {
        debug_print("Window default constructor, space at %p)\n", buffer->address<void>());
    }

    template<class T>
    count_window<T>::count_window(const count_window<T> &copy) :
            buffer(std::make_unique<hsa::hsa_buffer>(*copy.buffer)),
            capacity(copy.capacity), inserts(copy.inserts) {
        debug_print("Window copy constructor, space at %p)\n", buffer->address<void>());
    }

    template<class T>
    count_window<T>::count_window(count_window<T> &&mov) noexcept :
            buffer(std::move(mov.buffer)),
            capacity(mov.capacity), inserts(mov.inserts) {
        debug_print("Window move constructor, space at %p)\n", buffer->address<void>());
    }

    template<class T>
    count_window<T> &count_window<T>::operator=(const count_window<T> &copy) {
        buffer = std::make_unique<hsa::hsa_buffer>(*copy.buffer);
        capacity = copy.capacity;
        inserts = copy.inserts;
        return *this;
    }

    template<class T>
    count_window<T> &count_window<T>::operator=(count_window<T> &&move) noexcept {
        buffer = std::move(move.buffer);
        capacity = move.capacity;
        inserts = move.inserts;
        return *this;
    }

    template<class T>
    count_window<T>::~count_window() {
        if ( buffer ) {
            debug_print("Freeing window space at %p\n", buffer->address<void>());
        }
    }

    template<class T>
    void count_window<T>::add(const T &event) {
        buffer->address<T>()[inserts++ % capacity] = event;
    }

    template<class T>
    window_def count_window<T>::to_struct() const {
        return {buffer->address<void>(),inserts,capacity};
    }

    template<class T>
    event_sink<T>::event_sink(const std::shared_ptr<event_queue<T>> &queue) : in_queue(queue) {
    }

    template<class T>
    event_sink<T>::~event_sink() {
        debug_print("%s\n", "Destroying sink.");
        debug_print("InQueue read index : %li\n", this->in_queue->get_read_idx());
        debug_print("InQueue write index: %li\n", this->in_queue->get_write_idx());
    }

    template<class T1, class T2>
    binary_event_sink<T1,T2>::binary_event_sink(const std::shared_ptr<event_queue<T1>> &queue1,
                                            const std::shared_ptr<event_queue<T2>> &queue2) : in_queue1(queue1), in_queue2(queue2) {

    }

    template<class T1, class T2>
    binary_event_sink<T1, T2>::~binary_event_sink() {
        debug_print("%s\n", "Destroying binary sink.");
        debug_print("InQueue1 read index : %li\n", this->in_queue1->get_read_idx());
        debug_print("InQueue1 write index: %li\n", this->in_queue1->get_write_idx());
        debug_print("InQueue2 read index : %li\n", this->in_queue2->get_read_idx());
        debug_print("InQueue2 write index: %li\n", this->in_queue2->get_write_idx());
    }

    template<class T>
    event_source<T>::event_source(const std::shared_ptr<event_queue<T>> &out_queue) : out_queue(out_queue) {
    }

    template<class T>
    event_source<T>::~event_source() {
        debug_print("%s\n", "Destroying source.");
        debug_print("OutQueue read index : %li\n", this->out_queue->get_read_idx());
        debug_print("OutQueue write index: %li\n", this->out_queue->get_write_idx());
        debug_print("%s\n", "Finished destroying source.");
    }

    /////////////////////////////////////////////////////////////////
    //
    // INSTANTIATION
    //
    /////////////////////////////////////////////////////////////////

    template class count_window<event>;

    template class event_sink<event>;
    template class event_sink<joined_event>;
    template class event_sink<aggregated_event>;

    template class binary_event_sink<event,event>;

    template class event_source<event>;
    template class event_source<joined_event>;
    template class event_source<aggregated_event>;

    // Latency
    template class count_window<event_latency>;

    template class event_sink<event_latency>;
    template class event_sink<joined_event_latency>;

    template class binary_event_sink<event_latency,event_latency>;

    template class event_source<event_latency>;
    template class event_source<joined_event_latency>;
}
