//
// Created by koerberm on 11.06.18.
//

#include <random>
#include <bitset>
#include <thread>
#include <functional>

#include "ep/operators.h"
#include "ep/cpu_operators.h"
#include "ep/cpu_parallel_operators.h"
#include "ep/dispatch_operators.h"
#include "ep/hsa_operators.h"
#include "ep/event.h"
#include "util/eval_util.h"

#define FILTER_NUM_EVENTS (16777216l)

#define EATER_MIN_BATCH 1
#define EATER_MAX_BATCH FILTER_NUM_EVENTS
#define OUT_Q_SIZE FILTER_NUM_EVENTS


#define FILTER_IN_Q_SIZE FILTER_NUM_EVENTS//4096

#define RAND_LOWER 0.0f

std::random_device rd;  //Will be used to obtain a seed for the random number engine
auto seed = rd();

template <class T>
void filter_test(uint32_t min_batch_size, uint32_t max_batch_size,
                 std::shared_ptr<ep::event_queue<event>> in_q,
                                 const hsa::hsa_platform &platform, const hsa::hsa_agent &agent,
     std::function<std::unique_ptr<ep::event_sink<T>>(const std::shared_ptr<ep::event_queue<T>> &qin,
                                                      const std::shared_ptr<ep::event_queue<T>> &qout,
                                                      uint32_t min_batch_size, uint32_t max_batch_size
     )> create_filter, std::vector<double> &results, std::vector<uint64_t > &latency) {


    auto qp = std::make_shared<ep::event_queue<event>>(OUT_Q_SIZE, platform, agent);

    auto w = in_q->get_write_idx();
    in_q->set_write_index(0);

    ep::event_eater<event> p(qp,EATER_MIN_BATCH,EATER_MAX_BATCH);
    auto filter = create_filter(in_q,qp,min_batch_size,max_batch_size);

    std::this_thread::sleep_for(std::chrono::milliseconds(100));

    auto begin= std::chrono::steady_clock::now();
    in_q->set_write_index(w);
    in_q->wait_for_space(FILTER_IN_Q_SIZE-min_batch_size, 60000000000);
    std::chrono::duration<double> duration=  std::chrono::steady_clock::now() - begin;

    results.push_back(duration.count());

    std::this_thread::sleep_for(std::chrono::milliseconds(100));

    latency.push_back(filter->get_processing_latency());

    std::cout << ", (" << p.count << "," << p.unique_val << ")" << std::flush;
}

std::shared_ptr<ep::event_queue<event>> create_input(float rand_upper, const hsa::hsa_platform &platform, const hsa::hsa_agent &agent) {
    auto qf = std::make_shared<ep::event_queue<event>>(FILTER_IN_Q_SIZE, platform, agent);
    std::mt19937 gen(seed); //Standard mersenne_twister_engine seeded with rd()
    std::uniform_real_distribution<float> dis(RAND_LOWER,rand_upper);

    for (uint64_t ts = 1; ts <= FILTER_NUM_EVENTS; ts++) {
        qf->add(create_event(ts,dis,gen));
    }
    return qf;
}

int main(int argc, char **argv) {
    (void) argc;
    (void) argv;

    hsa::hsa_platform platform;

    for ( auto &agent : platform.get_agents() ) {
        if (agent.features != HSA_AGENT_FEATURE_KERNEL_DISPATCH)
            continue;

        std::cout << "Using agent " << agent.name << std::endl;

        hsa::hsa_executable exe;

        std::vector<std::string> cos = {
                "src/rocm/cl/out/hsa_filter_normal.co",
                "src/rocm/cl/out/hsa_filter_wf_normal.co",
                "src/rocm/cl/out/hells_filter_normal.co"
        };

        for (auto &file : cos) {
            auto co = hsa::code_object_from_file(file, agent);
            try {
                exe.load_code_object(co, agent);
            } catch (const hsa::hsa_exception &e) {
                std::cout << "Error loading module (" << file << "): " << e.what() << std::endl;
                return 1;
            }
        }

        exe.freeze();


        std::function<std::unique_ptr<ep::event_sink<event>>(const std::shared_ptr<ep::event_queue<event>> &qin,
                                                             const std::shared_ptr<ep::event_queue<event>> &qout,
                                                             uint32_t min_batch_size,
                                                             uint32_t max_batch_size)> create_cpu = [](
                const std::shared_ptr<ep::event_queue<event>> &qin,
                const std::shared_ptr<ep::event_queue<event>> &qout,
                uint32_t min_batch_size, uint32_t max_batch_size) {
            return std::make_unique<ep::filter_cpu<event>>(
                    qin, qout, min_batch_size, max_batch_size, [](const event &e) { return filter_predicate(&e); }
            );
        };

        std::function<std::unique_ptr<ep::event_sink<event>>(const std::shared_ptr<ep::event_queue<event>> &qin,
                                                             const std::shared_ptr<ep::event_queue<event>> &qout,
                                                             uint32_t min_batch_size,
                                                             uint32_t max_batch_size)> create_cpu_parallel = [&](
                const std::shared_ptr<ep::event_queue<event>> &qin,
                const std::shared_ptr<ep::event_queue<event>> &qout,
                uint32_t min_batch_size, uint32_t max_batch_size) {
            uint16_t num_threads = 4;
            return std::make_unique<ep::parallel_filter<event>>(
                    qin, qout, min_batch_size, max_batch_size, num_threads, [](const event &e) { return filter_predicate(&e); },
                    platform, agent
            );
        };

        std::function<std::unique_ptr<ep::event_sink<event>>(const std::shared_ptr<ep::event_queue<event>> &qin,
                                                             const std::shared_ptr<ep::event_queue<event>> &qout,
                                                             uint32_t min_batch_size,
                                                             uint32_t max_batch_size)> create_hells = [&](
                const std::shared_ptr<ep::event_queue<event>> &qin,
                const std::shared_ptr<ep::event_queue<event>> &qout,
                uint32_t min_batch_size, uint32_t max_batch_size) {
            return std::make_unique<ep::hells_filter<event>>(
                    qin, qout, min_batch_size, max_batch_size, platform, agent, exe
            );
        };

        std::function<std::unique_ptr<ep::event_sink<event>>(const std::shared_ptr<ep::event_queue<event>> &qin,
                                                             const std::shared_ptr<ep::event_queue<event>> &qout,
                                                             uint32_t min_batch_size,
                                                             uint32_t max_batch_size)> create_hsa_wave = [&](
                const std::shared_ptr<ep::event_queue<event>> &qin,
                const std::shared_ptr<ep::event_queue<event>> &qout,
                uint32_t min_batch_size, uint32_t max_batch_size) {
            return std::make_unique<ep::hsa_filter_wf<event>>(
                    qin, qout, min_batch_size, max_batch_size, platform, agent, exe
            );
        };

        std::function<std::unique_ptr<ep::event_sink<event>>(const std::shared_ptr<ep::event_queue<event>> &qin,
                                                             const std::shared_ptr<ep::event_queue<event>> &qout,
                                                             uint32_t min_batch_size,
                                                             uint32_t max_batch_size)> create_hsa = [&](
                const std::shared_ptr<ep::event_queue<event>> &qin,
                const std::shared_ptr<ep::event_queue<event>> &qout,
                uint32_t min_batch_size, uint32_t max_batch_size) {
            return std::make_unique<ep::hsa_filter<event>>(
                    qin, qout, min_batch_size, max_batch_size, platform, agent, exe
            );
        };

        std::vector<double> cpu;
        std::vector<double> cpu_mt;
        std::vector<double> hells;
        std::vector<double> hsa;
        std::vector<double> hsa_wf;

        std::vector<uint64_t> cpu_l;
        std::vector<uint64_t> cpu_mt_l;
        std::vector<uint64_t> hells_l;
        std::vector<uint64_t> hsa_l;
        std::vector<uint64_t> hsa_wf_l;

        std::vector<uint32_t> batch_sizes = {/*1,2,4,8,16,32,*/64,128,256,512,1024,2048,4096, 4096 << 1, 4096 << 2 };
        std::vector<float> sels = {1/*,2,4*/,10,100,1000,10000,100000};

        for (auto upper : sels) {
            cpu.clear();
            cpu_l.clear();
            cpu_mt.clear();
            cpu_mt_l.clear();
            hells.clear();
            hells_l.clear();
            hsa.clear();
            hsa_l.clear();
            hsa_wf.clear();
            hsa_wf_l.clear();

            std::cout << "Selectivity: " << (1.0f / upper) << ", creating input data..." << std::flush;
            auto in_q = create_input(upper,platform,agent);
            std::cout << " Done!" << std::endl;


            for (auto batch : batch_sizes) {
                std::cout << batch << std::flush;

                filter_test(batch, batch, in_q, platform, agent, create_cpu, cpu, cpu_l);
                in_q->set_read_index(0);

                filter_test(batch, batch, in_q, platform, agent, create_cpu_parallel, cpu_mt, cpu_mt_l);
                in_q->set_read_index(0);

                filter_test(batch, batch, in_q, platform, agent, create_hells, hells, hells_l);
                in_q->set_read_index(0);

                filter_test(batch, batch, in_q, platform, agent, create_hsa, hsa, hsa_l);
                in_q->set_read_index(0);

//                filter_test(batch, batch, in_q, platform, agent, create_hsa_wave, hsa_wf, hsa_wf_l);
//                in_q->set_read_index(0);
                std::cout << std::endl;
            }

            std::cout << "Batches ; " << batch_sizes << std::endl;
            std::cout << "CPU     ; " << cpu << std::endl;
            std::cout << "CPU-MT  ; " << cpu_mt << std::endl;
            std::cout << "Dispatch; " << hells << std::endl;
            std::cout << "HSA     ; " << hsa << std::endl;
//            std::cout << "HSA_WF  ; " << hsa_wf << std::endl;

            std::cout << "Latency CPU   : " << cpu_l << std::endl;
            std::cout << "Latency CPU-MT: " << cpu_mt_l << std::endl;
            std::cout << "Latency HELLS : " << hells_l << std::endl;
            std::cout << "Timings HSA   : " << hsa_l << std::endl;
//            std::cout << "Latency HSA_WF: " << hsa_wf_l << std::endl;

            std::cout << "==================================================" << std::endl;
        }
        std::cout << "==========================================================" << std::endl;
    }
    return 0;
}