//
// Created by koerberm on 11.06.18.
//

#include <random>
#include <bitset>
#include <thread>
#include <functional>

#include "ep/operators.h"
#include "ep/cpu_operators.h"
#include "ep/dispatch_operators.h"
#include "ep/hsa_operators.h"
#include "ep/latency/latency_addons.h"
#include "util/eval_util.h"
#include "ep/latency/event_latency.h"

#define FILTER_MAX_BATCH 4096
#define FILTER_IN_Q_SIZE 8192
#define FILTER_NUM_EVENTS (FILTER_IN_Q_SIZE<<7)

#define RAND_LOWER 0.0f

std::random_device rd;  //Will be used to obtain a seed for the random number engine
auto seed = rd();

template <class T>
void filter_test(float rand_upper, uint32_t min_batch_size, uint32_t max_batch_size,
                                 const hsa::hsa_platform &platform, const hsa::hsa_agent &agent,
     std::function<std::unique_ptr<ep::event_sink<T>>(const std::shared_ptr<ep::event_queue<T>> &qin,
                                                      const std::shared_ptr<ep::event_queue<T>> &qout,
                                                      uint32_t min_batch_size, uint32_t max_batch_size
     )> create_filter, std::vector<double> &results, std::vector<ep::latency_stats> &latency) {

    auto qf = std::make_shared<ep::event_queue_latency<event_latency>>(FILTER_IN_Q_SIZE, platform, agent,false);
    auto qp = std::make_shared<ep::event_queue_latency<event_latency>>(FILTER_NUM_EVENTS, platform, agent,true);

    std::mt19937 gen(seed); //Standard mersenne_twister_engine seeded with rd()
    std::uniform_real_distribution<float> dis(RAND_LOWER,rand_upper);

    ep::latency_measure<event_latency> p(qp);
    auto filter = create_filter(qf,qp,min_batch_size,max_batch_size);

    std::this_thread::sleep_for(std::chrono::milliseconds(100));

    std::vector<event_latency> in;
    for (uint64_t ts = 1; ts <= FILTER_NUM_EVENTS; ts++) {
        in.push_back(create_event_latency(ts,dis,gen));
    }


    auto begin= std::chrono::steady_clock::now();
    for ( auto &e : in ) {
        qf->add(e);
    }
    qf->wait_for_space(FILTER_IN_Q_SIZE-min_batch_size, 60000000000);
    std::chrono::duration<double> duration=  std::chrono::steady_clock::now() - begin;
    qp->wait_for_space(FILTER_NUM_EVENTS, 60000000000);

    results.push_back(duration.count());
    latency.push_back(p.get_stats());
    std::this_thread::sleep_for(std::chrono::milliseconds(100));
}

int main(int argc, char **argv) {
    (void) argc;
    (void) argv;

    hsa::hsa_platform platform;

    for ( auto &agent : platform.get_agents() ) {
        if (agent.features != HSA_AGENT_FEATURE_KERNEL_DISPATCH)
            continue;

        std::cout << "Using agent " << agent.name << std::endl;

        hsa::hsa_executable exe;

        std::vector<std::string> cos = {
                "src/rocm/cl/out/hsa_filter_latency.co",
                "src/rocm/cl/out/hells_filter_latency.co"
        };

        for (auto &file : cos) {
            auto co = hsa::code_object_from_file(file, agent);
            try {
                exe.load_code_object(co, agent);
            } catch (const hsa::hsa_exception &e) {
                std::cout << "Error loading module (" << file << "): " << e.what() << std::endl;
                return 1;
            }
        }
        exe.freeze();


        std::function<std::unique_ptr<ep::event_sink<event_latency>>(
                const std::shared_ptr<ep::event_queue<event_latency>> &qin,
                const std::shared_ptr<ep::event_queue<event_latency>> &qout,
                uint32_t min_batch_size, uint32_t max_batch_size)> create_cpu = [](
                const std::shared_ptr<ep::event_queue<event_latency>> &qin,
                const std::shared_ptr<ep::event_queue<event_latency>> &qout,
                uint32_t min_batch_size, uint32_t max_batch_size) {
            return std::make_unique<ep::filter_cpu<event_latency>>(
                    qin, qout, min_batch_size, max_batch_size,
                    [](const event_latency &e) { return filter_predicate(&e); }
            );
        };

        std::function<std::unique_ptr<ep::event_sink<event_latency>>(
                const std::shared_ptr<ep::event_queue<event_latency>> &qin,
                const std::shared_ptr<ep::event_queue<event_latency>> &qout,
                uint32_t min_batch_size, uint32_t max_batch_size)> create_hells = [&](
                const std::shared_ptr<ep::event_queue<event_latency>> &qin,
                const std::shared_ptr<ep::event_queue<event_latency>> &qout,
                uint32_t min_batch_size, uint32_t max_batch_size) {
            return std::make_unique<ep::hells_filter<event_latency>>(
                    qin, qout, min_batch_size, max_batch_size, platform, agent, exe);
        };

        std::function<std::unique_ptr<ep::event_sink<event_latency>>(
                const std::shared_ptr<ep::event_queue<event_latency>> &qin,
                const std::shared_ptr<ep::event_queue<event_latency>> &qout,
                uint32_t min_batch_size, uint32_t max_batch_size)> create_hsa = [&](
                const std::shared_ptr<ep::event_queue<event_latency>> &qin,
                const std::shared_ptr<ep::event_queue<event_latency>> &qout,
                uint32_t min_batch_size, uint32_t max_batch_size) {
            return std::make_unique<ep::hsa_filter_wf<event_latency>>(
                    qin, qout, min_batch_size, max_batch_size, platform, agent, exe
            );
        };

        std::vector<double> cpu;
        std::vector<double> hells;
        std::vector<double> hsa;

        std::vector<ep::latency_stats> cpu_l;
        std::vector<ep::latency_stats> hells_l;
        std::vector<ep::latency_stats> hsa_l;

        std::vector<uint32_t> batch_sizes = {1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096};
        std::vector<float> sels = {/*1,2,4,*/10, 100, 1000};

        for (auto upper : sels) {
            cpu.clear();
            cpu_l.clear();
            cpu.clear();
            hells.clear();
            hells_l.clear();
            hsa.clear();
            hsa_l.clear();


            std::cout << "Selectivity: " << (1.0f / upper) << std::endl;
            std::cout << "Min Batch size: ";
            for (auto batch : batch_sizes) {
                std::cout << batch << std::flush << ", ";
                filter_test(upper, batch, FILTER_MAX_BATCH, platform, agent, create_cpu, cpu, cpu_l);
                filter_test(upper, batch, FILTER_MAX_BATCH, platform, agent, create_hells, hells, hells_l);
                filter_test(upper, batch, FILTER_MAX_BATCH, platform, agent, create_hsa, hsa, hsa_l );
            }
            std::cout << " Done!" << std::endl;

            std::cout << "Batches      : " << batch_sizes << std::endl;
            std::cout << "Timings CPU  : " << cpu << std::endl;
            std::cout << "Timings HELLS: " << hells << std::endl;
            std::cout << "Timings HSA  : " << hsa << std::endl;

            std::cout << "Latency CPU  : " << cpu_l << std::endl;
            std::cout << "Latency HELLS: " << hells_l << std::endl;
            std::cout << "Latency HSA  : " << hsa_l << std::endl;

            std::cout << "==================================================" << std::endl;
        }
        std::cout << "==========================================================" << std::endl;
    }


    return 0;
}