//
// Created by koerberm on 05.03.18.
//

#include <iostream>
#include "hsa_platform/hsa_platform.h"

void qcb(hsa_status_t stat, hsa_queue_t * q, void * data ) {
    (void) data;
    const char* msg;
    hsa_status_string(stat,&msg);
    std::cout << "Error on queue: " << q->id << ": " << msg << std::endl;
}

int main(int argc, char **argv) {
    (void) argc;
    (void) argv;
    std::cout << "Initializing HSA Platform..." << std::endl;

    try {
        hsa::hsa_platform platform;
        std::cout << platform << std::endl;

        for ( auto &agent : platform.get_agents() ) {
            if ( agent.features != HSA_AGENT_FEATURE_KERNEL_DISPATCH )
                continue;

            std::cout << "==========================================================" << std::endl;
            std::cout << "Using agent: " << agent.name << std::endl;

            std::cout << "Creating queue." << std::endl;
            auto queue = agent.create_queue(agent.queue_max_size, HSA_QUEUE_TYPE_SINGLE, UINT32_MAX, UINT32_MAX, qcb);

            std::cout << "Loading code object." << std::endl;
            auto co = hsa::code_object_from_file("src/rocm/cl/out/vector_copy.co",agent);

            std::cout << "Creating exe." << std::endl;
            hsa::hsa_executable exe;

            std::cout << "Loading code object." << std::endl;
            exe.load_code_object(co, agent);

            std::cout << "Exe before freeze." << std::endl;
            std::cout << exe << std::endl;
            std::cout << "Freezing exe." << std::endl;
            exe.freeze();
            std::cout << "Exe after freeze." << std::endl;
            std::cout << exe << std::endl;

            std::cout << "Obtaining kernarg region." << std::endl;
            auto &kernarg_segment = agent.get_kernarg_region();
            std::cout << "Obtaining kernel symbol." << std::endl;
            auto &kernel = exe.get_kernel_symbols()[0];

            struct {
                void *in;
                void *out;
            } args = {nullptr, nullptr};

            uint16_t size = 1024;

            std::cout << "Allocating argument buffers." << std::endl;
            auto buf_in = agent.get_fine_grained_region().allocate(size * sizeof(int));
            auto buf_out = agent.get_fine_grained_region().allocate(size * sizeof(int));

            args.in = buf_in->address<void>();
            args.out = buf_out->address<void>();

            for (uint16_t i = 0; i < size; i++) {
                static_cast<int *>(args.in)[i] = i;
            }


            std::cout << "Populating argument buffer." << std::endl;
            auto kernargs = kernarg_segment.allocate(kernel.kernarg_segment_size, kernel.kernarg_segment_alignment);
            memcpy(kernargs->address<void>(), &args, sizeof(args));


            std::cout << "Creating completion signal." << std::endl;
            auto completion_signal = platform.create_signal(1);

            std::cout << "Creating dispatch packet." << std::endl;
            hsa::hsa_kernel_dispatch_packet p(1, kernel, *completion_signal, *kernargs);
            p.set_grid_size(size);
            p.set_workgroup_size(256);

            std::cout << "Dispatching kernel." << std::endl;
            queue->dispatch_kernel_packet(p);

            std::cout << "Waiting for completion." << std::endl;

            completion_signal->wait_scacquire(HSA_SIGNAL_CONDITION_LT, 1, platform.timestamp_frequency * 5);

            std::cout << "Kernel completed." << std::endl;

/*        size_t csize = 1024 * sizeof(uint32_t);
        std::cout << "Allocating " << csize << " bytes in coarse grained memory region: " << std::endl;

        auto & r = agent.get_coarse_grained_region();

        std::cout << r << std::endl;


        auto cbuf = r.allocate(csize);
        cbuf->assign(agent);

        std::cout << "Erasing allocated memory" << std::endl;

        cbuf->address<uint32_t>()[0] = 42;

        //memset(cbuf->address<char>(),0,16);

        std::cout << "End of program." << std::endl;*/


            std::cout << "In : ";
            for (int i = 0; i < size; i++) {
                std::cout << reinterpret_cast<int *>(args.in)[i] << ", ";
            }
            std::cout << std::endl;

            std::cout << "Out: ";
            for (int i = 0; i < size; i++) {
                std::cout << reinterpret_cast<int *>(args.out)[i] << ", ";
            }
            std::cout << std::endl;

            std::cout << "==========================================================" << std::endl;
        }

    } catch ( const hsa::hsa_exception& e) {
        std::cerr << e.what() << std::endl;
    }

    return 0;
}