//
// Created by koerberm on 06.03.18.
//

#include "hsa_platform_core.h"
#include "hsa_platform_util.h"
#include "hsa_platform_programs.h"
#include <hsa/hsa_ext_amd.h>

#define check(status) \
if ((status) != HSA_STATUS_SUCCESS) { \
    throw hsa::hsa_exception(status); \
}

namespace hsa {

    /**
     * hsa_isa_info_call_convention
     */
    hsa_isa_info_call_convention::hsa_isa_info_call_convention(uint32_t wavefront_size,
                                                               uint32_t wavefronts_per_compute_unit) :
            wavefront_size(wavefront_size),
            wavefronts_per_compute_unit(wavefronts_per_compute_unit) {
    }

    std::ostream &operator<<(std::ostream &s, const hsa::hsa_isa_info_call_convention &cc) {
        s <<  "Call Convention: [Wavefront Size: " << cc.wavefront_size;
        s <<  ", Wavefronts per Compute Unit: " << cc.wavefronts_per_compute_unit << "]";
        return s;
    }

    /*
     * isa_info
     */
    hsa_isa_info::hsa_isa_info(hsa_isa_t handle) : handle(handle) {
        if ( handle.handle == 0 )
            return;

        uint32_t name_len;
        uint32_t  cc_count;

        hsa_status_t stat = hsa_isa_get_info(handle, HSA_ISA_INFO_NAME_LENGTH, 0, &name_len); check(stat);

        if ( name_len > 0 ) {
            name.resize(name_len-1,0);
            stat = hsa_isa_get_info(handle, HSA_ISA_INFO_NAME, 0, &(name[0]));
            check(stat);
        }

        stat = hsa_isa_get_info(handle, HSA_ISA_INFO_CALL_CONVENTION_COUNT, 0, &cc_count); check(stat);

        for ( uint32_t i = 0; i < cc_count; i++ ) {
            uint32_t  wf_size;
            uint32_t  wf_cu;
            stat = hsa_isa_get_info(handle, HSA_ISA_INFO_CALL_CONVENTION_INFO_WAVEFRONT_SIZE, i, &wf_size); check(stat);
            stat = hsa_isa_get_info(handle, HSA_ISA_INFO_CALL_CONVENTION_INFO_WAVEFRONTS_PER_COMPUTE_UNIT, i, &wf_cu); check(stat);
            call_conventions.emplace_back(hsa_isa_info_call_convention(wf_size,wf_cu));
        }
    }

    std::ostream &operator<<(std::ostream &s, const hsa::hsa_isa_info &a) {
        s << "ISA Info (" << a.handle.handle  << "): "<< std::endl;
        s << "  Name: " << a.name << std::endl;
        s << "  Call Conventions (" << a.call_conventions.size() << "):" << std::endl;
        for ( auto &cc : a.call_conventions ) {
            s << "    " << cc << std::endl;
        }
        return s;
    }

    /*
     * agent
     */

    std::ostream &operator<<(std::ostream &s, const hsa::hsa_agent &a) {
        s << "Agent (" << a.handle.handle << "):" << std::endl;
        s << "  Name         : " << a.name << std::endl;
        s << "  Vendor Name  : " << a.vendor_name << std::endl;
        s << "  Device Type  : " << map_device_type(a.device) << std::endl;
        s << "  Version      : " << a.version_major << "." << a.version_minor << std::endl;
        s << "  Machine Model: " << map_machine_model(a.machine_model) << std::endl;
        s << "  Profile      : " << map_profile(a.profile) << std::endl;
        s << "  Features     : " << map_agent_feature(a.features) << std::endl;
        s << "  Compute Units: " << a.compute_units << std::endl;

        s << "  Memory Regions (" << a.memory_regions.size() << "):" << std::endl;
        for ( auto &mr : a.memory_regions ) {
            s << mr << std::endl;
        }


        /*
         * Hardware Related
         */

        s << "  Hardware Infos:" << std::endl;
        s << "    Cache Sizes (bytes): " << std::endl;
        s << "      L1: " << a.cache_sizes[0] << std::endl;
        s << "      L2: " << a.cache_sizes[1] << std::endl;
        s << "      L3: " << a.cache_sizes[2] << std::endl;
        s << "      L4: " << a.cache_sizes[3] << std::endl;
        s << "    Numa Node ID                            : " << a.node << std::endl;
        s << "    Default Float Rounding Mode             : " << map_rounding_mode(a.default_float_rounding_mode) << std::endl;
        s << "    Base Profile Default Float Rounding Mode: " << map_rounding_mode(a.base_profile_default_float_rounding_mode) << std::endl;

        /*
         * Queue Related
         */

        if (a.features != 0) {
            s << "  Queue Infos:" << std::endl;
            s << "    Queue Type    : " << map_queue_type(a.queue_type) << std::endl;
            s << "    Max #Queues   : " << a.queues_max << std::endl;
            s << "    Queue Min Size: " << a.queue_min_size << std::endl;
            s << "    Queue Max Size: " << a.queue_max_size << std::endl;

            s << a.isa_info << std::endl;
        }

        /*
         * Kernel Agent Related
         */
        if ( (a.features & HSA_AGENT_FEATURE_KERNEL_DISPATCH)  != 0 ) {
            s << "  Kernel Dispatch Infos: " << std::endl;
            s << "    Wavefront Size     : " << a.wave_front_size << std::endl;
            s << "    Workgroup Max Size : " << a.workgroup_max_size << std::endl;
            s << "    Workgroup Max Dim  : " << a.workgroup_max_dim[0] << "x" << a.workgroup_max_dim[1] << "x" << a.workgroup_max_dim[2] << std::endl;
            s << "    Grid Max Size      : " << a.grid_max_size << std::endl;
            s << "    Grid Max Dim       : " << a.grid_max_dim.x << "x" << a.grid_max_dim.x << "x" << a.grid_max_dim.z << std::endl;
            s << "    Fast F16 Operations: " << (a.fast_f16_operation ? "true" : "false") << std::endl;
            s << "    FBarrier Max Size  : " << a.fbarrier_max_size;
        }

        return s;
    }

    hsa_agent::hsa_agent(hsa_agent_t agent) : handle(agent) {
        hsa_status_t stat;
        name.resize(63,0);
        stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_NAME, &name[0]);
        check(stat);
        name.resize(strlen(&name[0]),0);

        vendor_name.resize(63,0);
        stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_VENDOR_NAME, &vendor_name[0]);
        check(stat);
        vendor_name.resize(strlen(&vendor_name[0]),0);
        stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_VERSION_MAJOR, &version_major); check(stat);
        stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_VERSION_MINOR, &version_minor); check(stat);
        stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_DEVICE, &device); check(stat);
        stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_FEATURE, &features); check(stat);
        stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_MACHINE_MODEL, &machine_model); check(stat);
        stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_PROFILE, &profile); check(stat);
        stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_EXTENSIONS, extensions); check(stat);

        /** Hardware related */
        stat = hsa_agent_get_info(handle, (hsa_agent_info_t)HSA_AMD_AGENT_INFO_COMPUTE_UNIT_COUNT, &compute_units); check(stat);
        stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_CACHE_SIZE, cache_sizes); check(stat);
        stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_NODE, &node); check(stat);
        stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_DEFAULT_FLOAT_ROUNDING_MODE, &default_float_rounding_mode); check(stat);
        stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_BASE_PROFILE_DEFAULT_FLOAT_ROUNDING_MODES, &base_profile_default_float_rounding_mode); check(stat);

        stat = hsa_agent_iterate_regions(handle,regions_callback,&memory_regions); check(stat);

        if ( features != 0 ) {
            /** Queue related */
            stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_QUEUE_TYPE, &queue_type); check(stat);
            stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_QUEUES_MAX, &queues_max); check(stat);
            stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_QUEUE_MIN_SIZE, &queue_min_size); check(stat);
            stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_QUEUE_MAX_SIZE, &queue_max_size); check(stat);


            /** ISA */
            hsa_isa_t isa_handle = {0};
            stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_ISA, &isa_handle); check(stat);
            isa_info = hsa_isa_info(isa_handle);

            if ( (features & HSA_AGENT_FEATURE_KERNEL_DISPATCH) != 0 ) {
                /** Kernel agent only */
                stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_WAVEFRONT_SIZE, &wave_front_size); check(stat);
                stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_WORKGROUP_MAX_SIZE, &workgroup_max_size); check(stat);
                stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_WORKGROUP_MAX_DIM, workgroup_max_dim); check(stat);
                stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_GRID_MAX_SIZE, &grid_max_size); check(stat);
                stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_GRID_MAX_DIM, &grid_max_dim); check(stat);
                stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_FBARRIER_MAX_SIZE, &fbarrier_max_size); check(stat);
                stat = hsa_agent_get_info(handle, HSA_AGENT_INFO_FAST_F16_OPERATION, &fast_f16_operation); check(stat);
            }
        }
    }

    const std::vector<hsa_memory_region> &hsa_agent::get_memory_regions() const {
        return memory_regions;
    }

    const hsa_memory_region &hsa_agent::get_kernarg_region() const {
        for ( auto &r : memory_regions ) {
            if ( r.is_kernarg() )
                return r;
        }
        throw hsa_exception("Could not find kernarg memory region");
    }

    const hsa_memory_region &hsa_agent::get_fine_grained_region() const {
        for ( auto &r : memory_regions ) {
            if ( r.is_fine_grained() )
                return r;
        }
        throw hsa_exception("Could not find fine grained memory region");
    }

    const hsa_memory_region &hsa_agent::get_coarse_grained_region() const {
        for ( auto &r : memory_regions ) {
            if ( r.is_coarse_grained() )
                return r;
        }
        throw hsa_exception("Could not find corase grained memory region");
    }

    hsa_status_t hsa_agent::regions_callback(hsa_region_t region, void *data) {
        auto v = (std::vector<hsa_memory_region>*) data;
        v->emplace_back( hsa_memory_region(region) );
        return HSA_STATUS_SUCCESS;
    }

    std::unique_ptr<hsa_queue> hsa_agent::create_queue(uint32_t size, hsa_queue_type_t type, uint32_t priv_segment_size,
                                                       uint32_t group_segment_size,
                                                       void (*callback)(hsa_status_t, hsa_queue_t *, void *),
                                                       void *data) const {
        hsa_queue_t *queue;
        hsa_status_t stat;
        stat = hsa_queue_create(handle,size,type,callback,data,priv_segment_size,group_segment_size,&queue); check(stat);
        return std::make_unique<hsa_queue>(queue);
    }

    /*
     * platform
     */
    hsa_platform::hsa_platform() {
        hsa_status_t stat = hsa_init(); check(stat);

        stat = hsa_system_get_info(HSA_SYSTEM_INFO_VERSION_MAJOR, &version_major); check(stat);
        stat = hsa_system_get_info(HSA_SYSTEM_INFO_VERSION_MINOR, &version_minor); check(stat);
        stat = hsa_system_get_info(HSA_SYSTEM_INFO_TIMESTAMP_FREQUENCY, &timestamp_frequency); check(stat);
        stat = hsa_system_get_info(HSA_SYSTEM_INFO_SIGNAL_MAX_WAIT, &signal_max_wait); check(stat);
        stat = hsa_system_get_info(HSA_SYSTEM_INFO_ENDIANNESS, &endianness); check(stat);
        stat = hsa_system_get_info(HSA_SYSTEM_INFO_MACHINE_MODEL, &machine_model); check(stat);
        stat = hsa_system_get_info(HSA_SYSTEM_INFO_EXTENSIONS, extensions); check(stat);
        stat = hsa_iterate_agents(agent_callback, &agents); check(stat);
    }

    hsa_platform::~hsa_platform() {
        hsa_shut_down();
    }

    hsa_status_t hsa_platform::agent_callback(hsa_agent_t agt, void *data) {
        auto agents = (std::vector<hsa_agent> *) data;
        agents->push_back(hsa_agent(agt));
        return HSA_STATUS_SUCCESS;
    }

    std::ostream &operator<<(std::ostream &s, const hsa::hsa_platform &p) {
        s << "HSA Platform Info:" << std::endl;
        s << "  Version            : " << p.version_major << "." << p.version_minor << std::endl;
        s << "  Timestamp frequency: " << p.timestamp_frequency << "Hz" << std::endl;
        s << "  Signal Max Wait    : " << p.signal_max_wait << " (times frequency)" << std::endl;
        s << "  Endianess          : " << map_endianness(p.endianness) << std::endl;
        s << "  Machine Model      : " << map_machine_model(p.machine_model) << std::endl;
        s << "  Agents (" << p.agents.size() << "):" << std::endl;
        for ( auto &a : p.agents ) {
            s << a << std::endl;
        }
        return s;
    }

    const std::vector<hsa_agent> &hsa_platform::get_agents() const {
        return agents;
    }

//        No finalization in ROCm 2.x
//    const hsa_finalizer &hsa_platform::get_finalizer() const {
//        std::lock_guard<std::mutex> lock(mtx);
//        if ( !finalizer ) {
//            bool supported;
//            hsa_status_t stat = hsa_system_extension_supported(HSA_EXTENSION_FINALIZER, HSA_FINALIZER_VERSION_MAJOR, HSA_FINALIZER_VERSION_MINOR, &supported);
//            check(stat);
//            if ( !supported )
//                throw hsa_exception("Finalizer extension not supported.");
//            finalizer = std::make_unique<hsa_finalizer>();
//        }
//        return *finalizer;
//    }

    std::shared_ptr<hsa_signal>
    hsa_platform::create_signal(hsa_signal_value_t value, std::vector<std::reference_wrapper<const hsa_agent>> consumers) const {

        hsa_signal_t signal = {0};

        hsa_agent_t *agents = nullptr;
        if ( !consumers.empty() ) {
            agents = (hsa_agent_t*) malloc( consumers.size() * sizeof(hsa_agent_t) );
            for ( size_t i = 0; i < consumers.size(); i++ )
                agents[i] = consumers[i].get().handle;
        }

        if ( consumers.size() > UINT32_MAX )
            throw hsa_exception("Too many signal consumers");

        hsa_status_t stat;
        stat = hsa_signal_create(value, static_cast<uint32_t >(consumers.size()), agents, &signal);

        if ( agents ) {
            free(agents);
        }

        check(stat);

        return std::make_shared<hsa_signal>(signal);
    }

    uint64_t hsa_platform::get_timestamp() const {
        uint64_t res;
        hsa_system_get_info(HSA_SYSTEM_INFO_TIMESTAMP,&res);
        return res;
    }

    /*
     * Memory region
     */

    hsa_memory_region::hsa_memory_region(hsa_region_t region) : handle(region) {
        hsa_status_t stat;
        stat = hsa_region_get_info(handle, HSA_REGION_INFO_SEGMENT, &segment); check(stat);
        stat = hsa_region_get_info(handle, HSA_REGION_INFO_GLOBAL_FLAGS, &global_flags); check(stat);
        stat = hsa_region_get_info(handle, HSA_REGION_INFO_SIZE, &size); check(stat);
        stat = hsa_region_get_info(handle, HSA_REGION_INFO_ALLOC_MAX_SIZE, &max_alloc_size); check(stat);
        stat = hsa_region_get_info(handle, HSA_REGION_INFO_RUNTIME_ALLOC_ALLOWED, &runtime_alloc_allowed); check(stat);
        stat = hsa_region_get_info(handle, HSA_REGION_INFO_RUNTIME_ALLOC_ALIGNMENT, &runtime_alloc_alignment); check(stat);
        stat = hsa_region_get_info(handle, HSA_REGION_INFO_RUNTIME_ALLOC_GRANULE, &runtime_alloc_granule); check(stat);
    }

    std::ostream &operator<<(std::ostream &s, const hsa_memory_region &a) {
        s << "Memory Region (" << a.handle.handle << "):" << std::endl;
        s << "  Segment: " << map_memory_segment(a.segment) << std::endl;
        if (a.segment == HSA_REGION_SEGMENT_GLOBAL) {
            s << "  Global Flags:" << std::endl;
            s << "    KernArg       : " << ((a.global_flags & HSA_REGION_GLOBAL_FLAG_KERNARG) != 0) << std::endl;
            s << "    Fine Grained  : " << ((a.global_flags & HSA_REGION_GLOBAL_FLAG_FINE_GRAINED) != 0) << std::endl;
            s << "    Coarse Grained: " << ((a.global_flags & HSA_REGION_GLOBAL_FLAG_COARSE_GRAINED) != 0) << std::endl;
        }
        s << "  Size                   : " << a.size << std::endl;
        s << "  Max Alloc Size         : " << a.max_alloc_size << std::endl;
        s << "  Runtime Alloc          : " << a.runtime_alloc_allowed << std::endl;
        s << "  Runtime Alloc Granule  : " << a.runtime_alloc_granule << std::endl;
        s << "  Runtime Alloc Alignment: " << a.runtime_alloc_alignment;
        return s;
    }

    std::unique_ptr<hsa_buffer> hsa_memory_region::allocate(size_t size) const {
        return std::make_unique<hsa_buffer>(handle, size);
    }

    std::unique_ptr<hsa_buffer> hsa_memory_region::allocate(size_t size, size_t alignment) const {
        return std::make_unique<hsa_buffer>(handle, size, alignment);
    }

    hsa_buffer::hsa_buffer(hsa_region_t region, size_t size) : handle(region), real_size(size), aligned_size(size) {
        hsa_status_t stat;
        stat = hsa_memory_allocate(handle,real_size,&alloced); check(stat);
        ::memset(alloced,0,real_size);
        aligned = alloced;
    }

    hsa_buffer::hsa_buffer(hsa_region_t region, size_t size, size_t alignment) : handle(region), real_size(size+alignment), aligned_size(size) {
        hsa_status_t stat;
        stat = hsa_memory_allocate(handle,real_size,&alloced); check(stat);
        ::memset(alloced,0,real_size);
        aligned = std::align(alignment, size, alloced, real_size);
        if ( !aligned ) {
            hsa_memory_free(alloced);
            throw hsa_exception("Unable to allocate aligned memory.");
        }
    }

    hsa_buffer::hsa_buffer(const hsa_buffer &copy) : handle(copy.handle), real_size(copy.real_size), aligned_size(copy.aligned_size) {
        hsa_status_t stat;
        stat = hsa_memory_allocate(handle,real_size,&alloced); check(stat);
        aligned = std::align(real_size-aligned_size, aligned_size, alloced, real_size);
        if ( !aligned ) {
            hsa_memory_free(alloced);
            throw hsa_exception("Unable to allocate aligned memory.");
        }
        ::memcpy(aligned,copy.aligned,aligned_size);
    }

    hsa_buffer::~hsa_buffer() {
        hsa_memory_free(alloced);
    }

    void hsa_buffer::clear() {
        ::memset(alloced,0,real_size);
    }

    void hsa_buffer::assign(const hsa_agent &agent, hsa_access_permission_t perm) {
        hsa_status_t stat;
        stat = hsa_memory_assign_agent(aligned, agent.handle, perm); check(stat);
    }

    /*
     * Queue
     */

    hsa_queue::~hsa_queue() {
        hsa_queue_destroy(handle);
    }

    hsa_queue::hsa_queue(hsa_queue_t *handle) : handle(handle), queue_mask(handle->size - 1) {

    }

    uint64_t hsa_queue::get_read_index() const {
        return hsa_queue_load_read_index_acquire(handle);
    }

    uint64_t hsa_queue::get_write_index() const {
        return hsa_queue_load_write_index_acquire(handle);
    }

    void hsa_queue::dispatch_kernel_packet(const hsa_kernel_dispatch_packet &packet, bool ring_doorbell) const {
        uint64_t w_idx = hsa_queue_add_write_index_relaxed(handle,1);

        uint64_t w_masked = w_idx & queue_mask;

        // Wait for space
        while (w_idx - hsa_queue_load_read_index_acquire(handle) >= handle->size);


        hsa_kernel_dispatch_packet_t* dispatch_packet = &(((hsa_kernel_dispatch_packet_t*)(handle->base_address))[w_masked]);

        // Initialize packet
        ::memset(((uint8_t*) dispatch_packet) + 4, 0, sizeof(hsa_kernel_dispatch_packet_t) - 4);


        // Packet setup
        uint16_t setup = 0;
        setup |= packet.dimensions << HSA_KERNEL_DISPATCH_PACKET_SETUP_DIMENSIONS;

        // Header
        uint16_t header = 0;
        header |= HSA_FENCE_SCOPE_SYSTEM << HSA_PACKET_HEADER_ACQUIRE_FENCE_SCOPE;
        header |= HSA_FENCE_SCOPE_SYSTEM << HSA_PACKET_HEADER_RELEASE_FENCE_SCOPE;
        header |= HSA_PACKET_TYPE_KERNEL_DISPATCH << HSA_PACKET_HEADER_TYPE;

        // Populate
        dispatch_packet->setup                = setup;
        dispatch_packet->workgroup_size_x     = packet.workgroup_size_x;
        dispatch_packet->workgroup_size_y     = packet.workgroup_size_y;
        dispatch_packet->workgroup_size_z     = packet.workgroup_size_z;
        dispatch_packet->grid_size_x          = packet.grid_size_x;
        dispatch_packet->grid_size_y          = packet.grid_size_y;
        dispatch_packet->grid_size_z          = packet.grid_size_z;
        dispatch_packet->kernel_object        = packet.kernel_symbol.kernel_object;
        dispatch_packet->kernarg_address      = packet.kernargs.address<void>();
        dispatch_packet->private_segment_size = packet.kernel_symbol.private_segment_size;
        dispatch_packet->group_segment_size   = packet.kernel_symbol.group_segment_size;
        dispatch_packet->completion_signal    = packet.completion_signal.handle;

        //packet_store_release((uint32_t*) dispatch_packet, header, setup);

        __atomic_store_n(&(dispatch_packet->header), header, __ATOMIC_RELEASE);

        if (ring_doorbell) {
            hsa_signal_store_relaxed(handle->doorbell_signal, w_idx);
        }
    }

    /*
     * Signal
     */
    hsa_signal::hsa_signal(hsa_signal_t signal) : handle(signal) {

    }

    hsa_signal::~hsa_signal() {
        hsa_signal_destroy(handle);
    }



    hsa_signal_value_t
    hsa_signal::wait_relaxed(hsa_signal_condition_t condition, hsa_signal_value_t comp, uint64_t timeout_hint,
                             hsa_wait_state_t wait_hint) const {
        return hsa_signal_wait_relaxed(handle,condition,comp,timeout_hint,wait_hint);
    }

    hsa_signal_value_t
    hsa_signal::wait_scacquire(hsa_signal_condition_t condition, hsa_signal_value_t comp, uint64_t timeout_hint,
                               hsa_wait_state_t wait_hint) const {
        return hsa_signal_wait_scacquire(handle,condition,comp,timeout_hint,wait_hint);
    }

    hsa_signal_value_t hsa_signal::load_relaxed() const {
        return hsa_signal_load_relaxed(handle);
    }

    hsa_signal_value_t hsa_signal::load_scacquire() const {
        return hsa_signal_load_scacquire(handle);
    }

    void hsa_signal::add_relaxed(hsa_signal_value_t v) {
        hsa_signal_add_relaxed(handle,v);
    }

    void hsa_signal::add_screlease(hsa_signal_value_t v) {
        hsa_signal_add_scacq_screl(handle,v);
    }

    void hsa_signal::store_relaxed(hsa_signal_value_t v) {
        hsa_signal_store_relaxed(handle,v);
    }

    void hsa_signal::store_screlease(hsa_signal_value_t v) {
        hsa_signal_store_screlease(handle,v);
    }
}
