//
// Created by koerberm on 06.03.18.
//

#ifndef EPHSA_HSA_PLATFORM_CORE_H
#define EPHSA_HSA_PLATFORM_CORE_H

#include <vector>
#include <iostream>
#include <memory>
#include <mutex>
#include <hsa/hsa.h>

namespace hsa {

    class hsa_platform;
//        No finalization in ROCm 2.x
//    class hsa_finalizer;
    class hsa_agent;
    class hsa_kernel_dispatch_packet;

    class hsa_queue {
        friend class hsa_agent;
    public:
        explicit hsa_queue(hsa_queue_t *handle);
        ~hsa_queue();

        hsa_queue(const hsa_queue&) = delete;
        const hsa_queue& operator=(const hsa_queue&) = delete;

        void dispatch_kernel_packet( const hsa_kernel_dispatch_packet &packet, bool ring_doorbell = true ) const;
        uint64_t get_read_index() const;
        uint64_t get_write_index() const;


    private:
        hsa_queue_t *handle;
        const uint32_t queue_mask;
    };

    class hsa_signal {
        friend class hsa_queue;
    public:
        explicit hsa_signal(hsa_signal_t signal);
        ~hsa_signal();

        void add_relaxed(hsa_signal_value_t v);

        void add_screlease(hsa_signal_value_t v);

        void store_relaxed(hsa_signal_value_t v);

        void store_screlease(hsa_signal_value_t v);

        hsa_signal_value_t  load_relaxed() const;

        hsa_signal_value_t  load_scacquire() const;

        hsa_signal_value_t wait_relaxed(hsa_signal_condition_t condition, hsa_signal_value_t comp,
                                        uint64_t timeout_hint = UINT64_MAX,
                                        hsa_wait_state_t wait_hint = HSA_WAIT_STATE_BLOCKED) const;

        hsa_signal_value_t wait_scacquire(hsa_signal_condition_t condition, hsa_signal_value_t comp,
                                          uint64_t timeout_hint = UINT64_MAX,
                                          hsa_wait_state_t wait_hint = HSA_WAIT_STATE_BLOCKED) const;

//    private:
        hsa_signal_t handle;
    };

    class hsa_buffer {
    public:
        hsa_buffer(hsa_region_t region, size_t size);
        hsa_buffer(hsa_region_t region, size_t size, size_t alignment);
        hsa_buffer( const hsa_buffer &copy );

        ~hsa_buffer();

        template<typename T>
        T* address() const {
            return reinterpret_cast<T*>(aligned);
        }

        void clear();

        void assign( const hsa_agent &agent, hsa_access_permission_t perm = HSA_ACCESS_PERMISSION_RW);

    private:
        hsa_region_t handle;
        size_t real_size;
        size_t aligned_size;
        void *alloced;
        void *aligned;
    };

    class hsa_memory_region {
    public:
        hsa_region_segment_t segment;
        size_t size;
        size_t max_alloc_size;
        bool runtime_alloc_allowed;
        size_t runtime_alloc_granule;
        size_t runtime_alloc_alignment;

        explicit hsa_memory_region(hsa_region_t region);
        ~hsa_memory_region() = default;

        bool is_kernarg() const { return (global_flags & HSA_REGION_GLOBAL_FLAG_KERNARG) != 0; };
        bool is_fine_grained() const { return (global_flags & HSA_REGION_GLOBAL_FLAG_FINE_GRAINED) != 0; };
        bool is_coarse_grained() const { return (global_flags & HSA_REGION_GLOBAL_FLAG_COARSE_GRAINED) != 0; };
        std::unique_ptr<hsa_buffer> allocate( size_t size ) const;
        std::unique_ptr<hsa_buffer> allocate( size_t size, size_t alignment ) const;

        friend std::ostream& operator<<(std::ostream& s, const hsa_memory_region& a);
    private:
        hsa_region_t handle;
        uint32_t global_flags;
    };

    class hsa_isa_info_call_convention {
    public:
        uint32_t wavefront_size;
        uint32_t wavefronts_per_compute_unit;

        hsa_isa_info_call_convention(uint32_t wavefront_size, uint32_t wavefronts_per_compute_unit);

        friend std::ostream& operator<<(std::ostream& s, const hsa_isa_info_call_convention& a);
    };

    class hsa_isa_info {
        friend class hsa_program;
    public:
        std::string name;
        std::vector<hsa_isa_info_call_convention> call_conventions;

        explicit hsa_isa_info(hsa_isa_t handle = {0});
        friend std::ostream& operator<<(std::ostream& s, const hsa_isa_info& a);
    private:
        hsa_isa_t handle;
    };

    class hsa_agent {
        friend class hsa_executable;
        friend class hsa_platform;
        friend class hsa_program;
        friend class hsa_buffer;
    public:
        std::string name;
        std::string vendor_name;
        uint16_t version_major;
        uint16_t version_minor;
        hsa_device_type_t device;
        hsa_agent_feature_t features;
        hsa_machine_model_t machine_model;
        hsa_profile_t profile;

        uint32_t wave_front_size;
        uint32_t compute_units;
        uint16_t workgroup_max_dim[3];
        uint32_t workgroup_max_size;
        hsa_dim3_t grid_max_dim;
        uint32_t grid_max_size;

        uint32_t queues_max;
        uint32_t queue_min_size;
        uint32_t queue_max_size;
        hsa_queue_type_t queue_type;

        uint32_t cache_sizes[4];
        hsa_isa_info isa_info;

        uint32_t node;
        hsa_default_float_rounding_mode_t default_float_rounding_mode;
        hsa_default_float_rounding_mode_t base_profile_default_float_rounding_mode;
        bool fast_f16_operation;
        uint32_t  fbarrier_max_size;

        uint8_t extensions[128];

        explicit hsa_agent(hsa_agent_t handle);

        const hsa_memory_region& get_kernarg_region() const;
        const hsa_memory_region& get_fine_grained_region() const;
        const hsa_memory_region& get_coarse_grained_region() const;
        std::unique_ptr<hsa_queue> create_queue(uint32_t size, hsa_queue_type_t type,
                                                uint32_t priv_segment_size = UINT32_MAX,
                                                uint32_t group_segment_size = UINT32_MAX,
                                                void (*callback)(hsa_status_t status, hsa_queue_t *source, void *data) = nullptr,
                                                void *data = nullptr) const;
        const std::vector<hsa_memory_region>& get_memory_regions() const;

        friend std::ostream& operator<<(std::ostream& s, const hsa_agent& a);

    private:
        hsa_agent_t handle;
        std::vector<hsa_memory_region> memory_regions;
        static hsa_status_t regions_callback(hsa_region_t region, void* data);
    };

    class hsa_platform {
        friend class hsa_agent;
    public:
        uint16_t version_major;
        uint16_t version_minor;
        uint64_t timestamp_frequency;
        uint64_t signal_max_wait;
        uint8_t  extensions[128];
        hsa_endianness_t endianness;
        hsa_machine_model_t machine_model;

        hsa_platform();
        ~hsa_platform();
        hsa_platform(const hsa_platform&) = delete;
        const hsa_platform& operator=(const hsa_platform&) = delete;

        uint64_t get_timestamp() const;
        const std::vector<hsa_agent>& get_agents() const;
//        No finalization in ROCm 2.x
//        const hsa_finalizer& get_finalizer() const;
        std::shared_ptr<hsa_signal> create_signal(hsa_signal_value_t, std::vector<std::reference_wrapper<const hsa_agent>> consumers = {}) const;
        friend std::ostream& operator<<(std::ostream& s, const hsa_platform& p);
    private:
        mutable std::mutex mtx;
        std::vector<hsa_agent> agents;
//        No finalization in ROCm 2.x
//        mutable std::unique_ptr<hsa_finalizer> finalizer;

        static hsa_status_t agent_callback(hsa_agent_t agent, void *data);
    };

}

#endif //EPHSA_HSA_PLATFORM_CORE_H
