//
// Created by koerberm on 09.03.18.
//

#include <random>
#include <bitset>
#include <thread>
#include "ep/operators.h"
#include "ep/cpu_operators.h"
#include "ep/cpu_parallel_operators.h"
#include "ep/dispatch_operators.h"
#include "ep/hsa_operators.h"
#include "ep/event.h"
#include "util/eval_util.h"

#define EATER_MIN_BATCH 1
#define EATER_MAX_BATCH 16777216
#define OUT_Q_SIZE 16777216

#define RAND_LOWER 0.0f

std::random_device rd;  //Will be used to obtain a seed for the random number engine
auto seed = rd();

////////////////////////////////////////////////////////////////
//
// JOIN
//
////////////////////////////////////////////////////////////////

template <class T1, class T2, class TO>
void join_test(uint64_t num_events, float rand_upper, uint64_t window_size,
               const hsa::hsa_platform &platform, const hsa::hsa_agent &agent,
               std::function<std::unique_ptr<ep::binary_event_sink<T1,T2>>(
                       const std::shared_ptr<ep::event_queue<T1>> &in1,
                       const std::shared_ptr<ep::event_queue<T2>> &in2,
                       const std::shared_ptr<ep::event_queue<TO>> &out,
                       ep::count_window<event> &&window_l, ep::count_window<event> &&window_r)> create_op,
               std::vector<double> &results, std::vector<uint64_t> &latency) {

    uint32_t queue_size = 1;
    while ( queue_size < num_events/2 )
        queue_size = queue_size << 1;

    auto ql = std::make_shared<ep::event_queue<event>>(queue_size, platform, agent);
    auto qr = std::make_shared<ep::event_queue<event>>(queue_size, platform, agent);
    auto qo = std::make_shared<ep::event_queue<joined_event>>(OUT_Q_SIZE, platform, agent);

    std::mt19937 gen(seed); //Standard mersenne_twister_engine seeded with rd()
    std::uniform_real_distribution<float> dis(RAND_LOWER,rand_upper);

    ep::event_eater<joined_event> p(qo,EATER_MIN_BATCH,EATER_MAX_BATCH);

    // Pre-Fill Windows
    ep::count_window<event> window_l(window_size, agent);
    ep::count_window<event> window_r(window_size, agent);

    uint64_t ts = 1;
    for ( ; ts <= 2*window_size; ts+=2) {
        window_l.add(create_event(ts,dis,gen));
        window_r.add(create_event(ts+1,dis,gen));
    }

    for ( ; ts <= 2*window_size+num_events; ts+=2) {
        ql->add(create_event(ts,dis,gen));
        qr->add(create_event(ts+1,dis,gen));
    }
    auto wl = ql->get_write_idx();
    auto wr = qr->get_write_idx();
    ql->set_write_index(0);
    qr->set_write_index(0);

    auto join = create_op(ql,qr,qo,std::move(window_l),std::move(window_r));

    std::this_thread::sleep_for(std::chrono::milliseconds(100));

    auto begin= std::chrono::steady_clock::now();
    ql->set_write_index(wl);
    qr->set_write_index(wr);
    ql->wait_for_space(queue_size, 600000000000);
    std::chrono::duration<double> duration =  std::chrono::steady_clock::now() - begin;
    join->stop();
    std::this_thread::sleep_for(std::chrono::milliseconds(100));
    qo->wait_for_space(OUT_Q_SIZE, 6000000000);
    results.push_back(duration.count());
    latency.push_back(join->get_processing_latency());
    std::cout << ", (" << p.count << "," << p.unique_val << ")" << std::flush;
}

int main(int argc, char **argv) {
    (void) argc;
    (void) argv;


    hsa::hsa_platform platform;

    for ( auto &agent : platform.get_agents() ) {
        if (agent.features != HSA_AGENT_FEATURE_KERNEL_DISPATCH)
            continue;

        std::cout << "Using agent " << agent.name << std::endl;

        hsa::hsa_executable exe;

        std::vector<std::string> cos = {
                "src/rocm/cl/out/hells_join_normal.co",
                "src/rocm/cl/out/hsa_join_normal.co"
        };

        for (auto &file : cos) {
            auto co = hsa::code_object_from_file(file, agent);
            try {
                exe.load_code_object(co, agent);
            } catch (const hsa::hsa_exception &e) {
                std::cout << "Error loading module (" << file << "): " << e.what() << std::endl;
                return 1;
            }
        }

        exe.freeze();


        // Operator create functions

        std::function<std::unique_ptr<ep::binary_event_sink<event, event>>(
                const std::shared_ptr<ep::event_queue<event>> &in1,
                const std::shared_ptr<ep::event_queue<event>> &in2,
                const std::shared_ptr<ep::event_queue<joined_event>> &out,
                ep::count_window<event> &&window_l, ep::count_window<event> &&window_r)> create_cpu = [&](
                const std::shared_ptr<ep::event_queue<event>> &in1,
                const std::shared_ptr<ep::event_queue<event>> &in2,
                const std::shared_ptr<ep::event_queue<joined_event>> &out,
                ep::count_window<event> &&window_l, ep::count_window<event> &&window_r) {
            return std::make_unique<ep::join_cpu<event, event, joined_event>>(
                    in1, in2, out, std::move(window_l), std::move(window_r),
                    [](const event &el, const event &er) { return join_predicate(&el, &er); },
                    [](const event &el, const event &er) { return compose(&el, &er); });
        };

        std::function<std::unique_ptr<ep::binary_event_sink<event, event>>(
                const std::shared_ptr<ep::event_queue<event>> &in1,
                const std::shared_ptr<ep::event_queue<event>> &in2,
                const std::shared_ptr<ep::event_queue<joined_event>> &out,
                ep::count_window<event> &&window_l, ep::count_window<event> &&window_r)> create_cpu_mt = [&](
                const std::shared_ptr<ep::event_queue<event>> &in1,
                const std::shared_ptr<ep::event_queue<event>> &in2,
                const std::shared_ptr<ep::event_queue<joined_event>> &out,
                ep::count_window<event> &&window_l, ep::count_window<event> &&window_r) {

            uint16_t num_threads = 4;

            return std::make_unique<ep::parallel_join<event, event, joined_event>>(
                    in1, in2, out, std::move(window_l), std::move(window_r), num_threads,
                    [](const event &el, const event &er) { return join_predicate(&el, &er); },
                    [](const event &el, const event &er) { return compose(&el, &er); },
                    platform, agent);
        };

        std::function<std::unique_ptr<ep::binary_event_sink<event, event>>(
                const std::shared_ptr<ep::event_queue<event>> &in1,
                const std::shared_ptr<ep::event_queue<event>> &in2,
                const std::shared_ptr<ep::event_queue<joined_event>> &out,
                ep::count_window<event> &&window_l, ep::count_window<event> &&window_r)> create_hells = [&](
                const std::shared_ptr<ep::event_queue<event>> &in1,
                const std::shared_ptr<ep::event_queue<event>> &in2,
                const std::shared_ptr<ep::event_queue<joined_event>> &out,
                ep::count_window<event> &&window_l, ep::count_window<event> &&window_r) {
            return std::make_unique<ep::hells_join<event, event, joined_event>>(
                    in1, in2, out, std::move(window_l), std::move(window_r),
                    [](const event &el, const event &er) { return compose(&el, &er); },
                    platform, agent, exe);
        };

        std::function<std::unique_ptr<ep::binary_event_sink<event, event>>(
                const std::shared_ptr<ep::event_queue<event>> &in1,
                const std::shared_ptr<ep::event_queue<event>> &in2,
                const std::shared_ptr<ep::event_queue<joined_event>> &out,
                ep::count_window<event> &&window_l, ep::count_window<event> &&window_r)> create_hsa = [&](
                const std::shared_ptr<ep::event_queue<event>> &in1,
                const std::shared_ptr<ep::event_queue<event>> &in2,
                const std::shared_ptr<ep::event_queue<joined_event>> &out,
                ep::count_window<event> &&window_l, ep::count_window<event> &&window_r) {

            auto res = std::make_unique<ep::hsa_join<event, event, joined_event>>
                    (in1, in2, out, std::move(window_l), std::move(window_r), platform, agent, exe);

            return res;
        };


        std::vector<double> cpu;
        std::vector<double> cpu_mt;
        std::vector<double> hells;
        std::vector<double> hsa;

        std::vector<uint64_t> cpu_l;
        std::vector<uint64_t > cpu_mt_l;
        std::vector<uint64_t> hells_l;
        std::vector<uint64_t> hsa_l;

        std::vector<uint32_t> window_sizes = {4096, 4096 << 1, 4096 << 2, 4096 << 3, 4096 << 4, 4096 << 5, 4096 << 6,
                                              4096 << 7, 4096 << 8};
        std::vector<float> sels = {1,/*2,4,*/10, 100,1000,10000, 100000};

        for (auto upper : sels) {
            cpu.clear();
            cpu_l.clear();
            cpu_mt.clear();
            cpu_mt_l.clear();
            hells.clear();
            hells_l.clear();
            hsa.clear();
            hsa_l.clear();
            std::cout << "Selectivity: " << (1.0f / upper) << std::endl;
            for (auto window : window_sizes) {
                uint32_t factor = (*(--window_sizes.end())/window);
                uint64_t num_events = 1000;

                std::cout << window << "/" << num_events << std::flush;
                join_test(num_events, upper, window, platform, agent, create_cpu, cpu, cpu_l);
                join_test(num_events, upper, window, platform, agent, create_cpu_mt, cpu_mt, cpu_mt_l);
                join_test(num_events, upper, window, platform, agent, create_hells, hells, hells_l);
                join_test(num_events, upper, window, platform, agent, create_hsa, hsa, hsa_l);
                std::cout << std::endl;
            }

            std::cout << "Window sizes  ; " << window_sizes << std::endl;
            std::cout << "CPU   ; " << cpu << std::endl;
            std::cout << "CPU-MT; " << cpu_mt << std::endl;
            std::cout << "HELLS ; " << hells << std::endl;
            std::cout << "HSA   ; " << hsa << std::endl;

            std::cout << "Latency CPU  : " << cpu_l << std::endl;
            std::cout << "Latency CPU-MT " << cpu_mt_l << std::endl;
            std::cout << "Latency HELLS: " << hells_l << std::endl;
            std::cout << "Latency HSA  : " << hsa_l << std::endl;

            std::cout << "==================================================" << std::endl;
        }
        std::cout << "==========================================================" << std::endl;
    }

    return 0;
}
