#include "../../hsa_helper.h"

void populate_queue_info( const queue_def *q, queue_info *info ) {
        info->read_idx  = get_queue_read_idx(q);
        info->write_idx = get_queue_write_idx(q);
}

bool wait_for_elements( const queue_def *q, queue_info *info, const hsa_signal_value_t num_elements ) {
        const hsa_signal_value_t min_write_idx = info->read_idx + num_elements;
        if ( info->write_idx < min_write_idx ) {
            info->write_idx = hsa_signal_wait(q->w_signal, HSA_SIGNAL_CONDITION_GTE, min_write_idx, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);
        }
        return info->write_idx == QUEUE_TERM_IDX;
}

__kernel void hsa_aggregate_switch(
        __constant queue_def  *in_q,
        __global   window_def *w,
        hsa_signal_t sync_1,
        hsa_signal_t sync_2,
        long num_groups,
        __global ulong *latency_accum,
        __global long *dbg
) {

    bool term = false;
    bool signal_switch = false;

    // Inititalize queue-info
    const queue_def  queue_in = *in_q;
    queue_info       queue_in_info;

    window_def window  = *w;

    ulong local_latency = 0;
    ulong last_begin = 0;
    ulong num_turns = 0;

    // Get current queue states
    populate_queue_info(&queue_in, &queue_in_info);

    T current;

    while ( !term ) {
        signal_switch = !signal_switch;

        term = wait_for_elements( &queue_in, &queue_in_info, 1 );
        if ( term ) {
            continue;
        }

        // Wait for groups to settle
        hsa_coord_wait(signal_switch, sync_1, sync_2, num_groups);

        // Retrieve next element
        current = *((T*) get_queue_element(&queue_in, queue_in_info.read_idx, sizeof(T)));
        T *window_address = get_window_element( &window, window.num_inserts, sizeof(T) );


        const ulong now = timestamp();
        if ( last_begin != 0 ) {
            num_turns++;
            local_latency += (now-last_begin);
        }
        last_begin = now;


        // Populate window
        *window_address = current;

        // Update window
        window.num_inserts++;

        hsa_coord_fire(signal_switch, sync_1, sync_2);

        // Tell event was consumed
        add_to_queue_read_idx( &queue_in, 1);
        queue_in_info.read_idx++;
    }

    // Wait for processors to settle and signal termination
    hsa_coord_wait(signal_switch, sync_1, sync_2, num_groups);

    hsa_signal_store(sync_1, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);
    hsa_signal_store(sync_2, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);

    // Write back window state
    *w = window;
    *latency_accum = (num_turns > 0) ? (local_latency/num_turns) * DEVICE_TIMER_FACTOR : 0;
}

////////////////////////////////////////////////////////////////////////////////////////7


AGG_TYPE proc_internal(
    ulong element_count,
    const T *in,
    __local  AGG_TYPE *local_accum,
    __global AGG_TYPE *global_accum,
    hsa_signal_t i_sync) {

    const size_t global_id   = get_global_linear_id();
    const size_t global_size = get_global_linear_size();

    const size_t local_id    = get_local_linear_id();
    const size_t local_size  = get_local_linear_size();
    const size_t group_waves = (local_size + WAVE_SIZE - 1) / WAVE_SIZE;

    const size_t num_groups  = get_linear_num_groups();

    size_t elements_per_thread = (element_count + global_size - 1) / global_size;

    // Initial round
    AGG_TYPE  accum = agg_init(&in[global_id]);

    for ( int i = global_size + global_id; i < element_count; i+=global_size ) {
        agg_merge(&accum, agg_init(&in[i]));
    }

    accum = agg_wf_merge(accum);

    if ( lane_id() == 0 ) {
        local_accum[local_id / WAVE_SIZE] = accum;
    }
    barrier(0);

    // Initial step complete, perform per group reduction
    if ( local_id == 0 ) {
        for ( size_t idx = 1; idx < group_waves; idx++ ) {
            agg_merge(&accum, local_accum[idx]);
        }
        global_accum[get_linear_group_id ()] = accum;
        hsa_signal_add(i_sync, 1, HSA_MEMORY_ORDER_SEQ_CST);
    }


    // Final step
    if ( global_id == 0 ) {
        hsa_signal_wait(i_sync, HSA_SIGNAL_CONDITION_EQ, num_groups, -1, HSA_MEMORY_ORDER_RELAXED);

        for ( size_t idx = 1; idx < num_groups; idx++ ) {
            agg_merge(&accum, global_accum[idx]);
        }
        hsa_signal_store(i_sync, 0, HSA_MEMORY_ORDER_SEQ_CST);
    }
    return accum;
}


AGG_TYPE proc_internal_sg(
    ulong element_count,
    const T *in,
    __local AGG_TYPE *local_accum) {

    const size_t global_id   = get_global_linear_id();
    const size_t global_size = get_global_linear_size();
    const size_t num_waves   = (global_size + WAVE_SIZE - 1) / WAVE_SIZE;

    size_t elements_per_thread = (element_count + global_size - 1) / global_size;

    // Initial round
    AGG_TYPE  accum = agg_init(&in[global_id]);

    for ( int i = global_size + global_id; i < element_count; i+=global_size ) {
        agg_merge(&accum, agg_init(&in[i]));
    }

    accum = agg_wf_merge(accum);

    if ( lane_id() == 0 ) {
        local_accum[global_id / WAVE_SIZE] = accum;
    }
    barrier(0);

    if ( global_id == 0 ) {
        for ( size_t idx = 1; idx < num_waves; idx++ ) {
            agg_merge(&accum, local_accum[idx]);
        }
    }
    return accum;
}



__kernel void hsa_aggregate_proc(
__constant queue_def  *out_q,
__constant window_def *w,
hsa_signal_t sync_1,
hsa_signal_t sync_2,
hsa_signal_t i_sync,
__global AGG_TYPE *global_accum,
__global int64_t *dbg
) {
    __local hsa_signal_value_t bcast[1];
    __local AGG_TYPE local_accum[1024/WAVE_SIZE];

    window_def       window         = *w;
    const queue_def  queue_out      = *out_q;
    queue_info       queue_out_info;

    const T *in = (T*) window.base;

    const bool single_group = get_linear_num_groups() == 1;
    bool signal_switch      = false;
    bool term               = false;

    AGG_TYPE result;

    populate_queue_info(&queue_out, &queue_out_info);

    // Determine the minimum read_idx of the output queue must reach
    hsa_signal_value_t min_read_idx = (queue_out_info.write_idx-queue_out.capacity);

    // Signal steady state
    hsa_group_ready(sync_1,sync_2);

    while ( !term ) {
        signal_switch = !signal_switch;

        term = hsa_group_wait( signal_switch, sync_1, sync_2, bcast );
        // Terminate?
        if ( term ) {
            continue;
        }

        ulong element_count = min(window.num_inserts,window.size);


        if ( single_group ) {
            result = proc_internal_sg(element_count, in, local_accum);
        }
        else {
            result = proc_internal(  element_count, in, local_accum, global_accum, i_sync );
        }

        // Populate result
        if ( get_global_linear_id()  == 0 ) {
            // Wait for space
            min_read_idx++;
            if ( queue_out_info.read_idx < min_read_idx ) {
                queue_out_info.read_idx = hsa_signal_wait(queue_out.r_signal, HSA_SIGNAL_CONDITION_GTE, min_read_idx, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);
            }

            // Write result
            TOUT *out = (TOUT*) get_queue_element(&queue_out, queue_out_info.write_idx, sizeof(TOUT));
            T current = *((T*)get_window_element( &window, window.num_inserts, sizeof(T)));
            *out = agg_finish(&current, result);
            queue_out_info.write_idx++;
            add_to_queue_write_idx( &queue_out, 1 );
        }

        // Update window
        window.num_inserts++;
        hsa_group_complete( signal_switch, sync_1, sync_2 );
    }
}

