#include "../../hsa_helper.h"

void populate_queue_info( const queue_def *q, queue_info *info ) {
        info->read_idx  = get_queue_read_idx(q);
        info->write_idx = get_queue_write_idx(q);
}

bool wait_for_elements( const queue_def *q, queue_info *info, const hsa_signal_value_t num_elements ) {
        const hsa_signal_value_t min_write_idx = info->read_idx + num_elements;
        if ( info->write_idx < min_write_idx ) {
            if ( lane_id() == 0 ) {
                info->write_idx = hsa_signal_wait(q->w_signal, HSA_SIGNAL_CONDITION_GTE, min_write_idx, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);
            }
            info->write_idx = broadcast_u64(info->write_idx,0);
        }
        return info->write_idx == QUEUE_TERM_IDX;
}

inline void write_block(
    const ulong block,
    const long base_read_idx,
    __private const queue_def *out,
    __private queue_info *info_out,
    __private const queue_def *in) {

    const uint num_results_wave =  __builtin_popcountl(block);
    const uint total_results     = __ockl_wfred_add_i32(num_results_wave);

    if ( total_results == 0 )
        return;

    hsa_signal_value_t min_read_idx = (info_out->write_idx-out->capacity);
    int    my_read_offset    = -1;
    uint   result_counter = 0;

    for ( int i = 0; i < WAVE_SIZE; i++ ) {
        const uint current = broadcast_u32(num_results_wave,i);

        if ( result_counter + current > WAVE_SIZE ) {
            // Make sure there is space in output queue
            min_read_idx += result_counter;
            if ( info_out->read_idx < min_read_idx ) {
                info_out->read_idx = hsa_signal_wait(out->r_signal, HSA_SIGNAL_CONDITION_GTE, min_read_idx, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);
            }

            if ( my_read_offset >= 0 ) {
                T *item  = ((T*) get_queue_element( in, base_read_idx + my_read_offset, sizeof(T) ));
                T *dest = (T*) get_queue_element( out, info_out->write_idx+active_lane_id(), sizeof(T) );
                do_write(dest,item);
            }
            if ( lane_id() == 0 ) {
                add_to_queue_write_idx(out,result_counter);
            }

            info_out->write_idx += result_counter;
            my_read_offset = -1;
            result_counter = 0;
        }


        if ( current > 0 ) {
            const long cblock = broadcast_u64(block,i);
            const bool passed = (cblock & (1l << lane_id()));

            int tmp_read_idx = -1;
            int to_lane = 63;


            if ( passed ) {
                to_lane = result_counter+active_lane_id();
                tmp_read_idx = i*64 + lane_id();
            }

            tmp_read_idx = __builtin_amdgcn_ds_permute(to_lane<<2,tmp_read_idx);

            if ( lane_id() >= result_counter )
                my_read_offset = tmp_read_idx;


            result_counter += current;
        }
    }

    if ( result_counter >0 ) {
        // Make sure there is space in output queue
        min_read_idx += result_counter;
        if ( info_out->read_idx < min_read_idx ) {
            info_out->read_idx = hsa_signal_wait(out->r_signal, HSA_SIGNAL_CONDITION_GTE, min_read_idx, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);
        }

        if ( my_read_offset >= 0 ) {
            T *item  = ((T*) get_queue_element( in, base_read_idx + my_read_offset, sizeof(T) ));
            T *dest = (T*) get_queue_element( out, info_out->write_idx+active_lane_id(), sizeof(T) );
            do_write(dest,item);
        }
        if ( lane_id() == 0 ) {
            add_to_queue_write_idx(out,result_counter);
        }

        info_out->write_idx += result_counter;
    }

}

__kernel void hsa_filter_switch(
    __constant queue_def *in_q,
    __constant queue_def *out_q,
    long min_batch_size,
    long max_batch_size,
    long num_groups,
    hsa_signal_t sync_1,
    hsa_signal_t sync_2,
    __global ulong *res1,
    __global ulong *res2,
    __global long *ctrl,
    __global ulong *latency_accum,
    __global long *dbg ) {

    bool term = false;
    bool signal_switch = false;

    // Inititalize queue-info
    const queue_def  q_in  = *in_q;
    queue_info info_in;
    const queue_def  q_out  = *out_q;
    queue_info info_out;

    // results
    ulong *results[2];
    results[0] = res1;
    results[1] = res2;

    const uint num_result_values = (max_batch_size+(sizeof(ulong)<<3)-1)/(sizeof(ulong)<<3);

    ulong local_latency = 0;
    ulong last_begin = 0;
    ulong num_turns = 0;

    populate_queue_info(&q_in, &info_in);
    populate_queue_info(&q_out, &info_out);

    term = wait_for_elements( &q_in, &info_in, min_batch_size );

    long next_read = info_in.read_idx;

    while ( !term ) {
        signal_switch = !signal_switch;

        if ( lane_id() == 0 ) {
            hsa_coord_wait(signal_switch, sync_1, sync_2, num_groups);
        }
        barrier(0);

        const long read_idx_base = info_in.read_idx;

        info_in.read_idx = broadcast_u64(next_read,0);

        term = wait_for_elements( &q_in, &info_in, min_batch_size );

        if (  lane_id() == 0 ) {
            // After next schedule, read-index will be this value
            next_read = min(info_in.read_idx+max_batch_size, info_in.write_idx);
            *ctrl = next_read;
            hsa_coord_fire(signal_switch, sync_1, sync_2);
        }


        for ( uint i = 0; i < num_result_values; i+= 64 ) {
            const ulong block =  (i+lane_id() < num_result_values) ? results[!signal_switch][i+lane_id()] : 0l;
            write_block(block,read_idx_base,&q_out,&info_out,&q_in);
        }

        if (  lane_id() == 0 ) {
            hsa_signal_store(q_in.r_signal, next_read, HSA_MEMORY_ORDER_SEQ_CST);
            const ulong now = timestamp();
            if ( last_begin != 0 ) {
                num_turns++;
                local_latency += (now-last_begin);
            }
            last_begin = now;
        }
    }

    if ( lane_id() == 0 ) {
        // Wait for processors to settle and signal termination
        hsa_coord_wait(!signal_switch, sync_1, sync_2, num_groups);
        hsa_signal_store(sync_1, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);
        hsa_signal_store(sync_2, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);

        // Write back window state
        *latency_accum = (num_turns > 0) ? (local_latency/num_turns) * DEVICE_TIMER_FACTOR : 0;
    }
}

__kernel void hsa_filter_proc(
    __constant queue_def *in_q,
    hsa_signal_t sync_1,
    hsa_signal_t sync_2,
    __global HELLS_FILTER_BITFIELD_TYPE *res1,
    __global HELLS_FILTER_BITFIELD_TYPE *res2,
    __global long *ctrl,
    __global long *dbg ) {

    __local hsa_signal_value_t bcast[1];

    const size_t global_id   = get_global_linear_id();
    const size_t global_size = get_global_linear_size();
    const size_t group_id    = get_linear_group_id();
    const size_t group_size  = get_local_linear_size();
    const size_t num_groups  = get_linear_num_groups();

    // Inititalize queue-info
    const queue_def  q_in   = *in_q;
    queue_info info_in;

    // results
    HELLS_FILTER_BITFIELD_TYPE *results[2];
    results[0] = res1;
    results[1] = res2;


    populate_queue_info(&q_in, &info_in);

    bool signal_switch = false;
    bool term          = false;

    T item;

    // Items per complete loop with all groups
    const uint elements_per_thread = sizeof(HELLS_FILTER_BITFIELD_TYPE) << 3;
    const long items_per_loop  = global_size * elements_per_thread;

    // Accum per thread
    HELLS_FILTER_BITFIELD_TYPE bitfield;

    // Signal steady state
    hsa_group_ready(sync_1,sync_2);

    while ( !term ) {
        signal_switch = !signal_switch;

        term = hsa_group_wait( signal_switch, sync_1, sync_2, bcast );
        // Terminate?
        if ( term ) {
            continue;
        }

        const long up_to       = *ctrl;
        const long num_elems   = up_to - info_in.read_idx;
        const long loops_per_group = (num_elems + items_per_loop - 1) / items_per_loop;

        hsa_signal_value_t loop_id = group_id;
        for ( int i = 0; i < loops_per_group; i++ ) {
            bitfield = 0;
            const long base_idx = info_in.read_idx + (i * global_size * elements_per_thread) + global_id*elements_per_thread;
            for ( int j = 0; j < elements_per_thread && base_idx + j < up_to; j++ ) {
                item = *((T*) get_queue_element(&q_in, base_idx+j, sizeof(T)));
                bitfield |= (filter_predicate(&item) << j);
            }
            results[signal_switch][global_size*i+global_id] = bitfield;
        }

        // Synchronize queue info
        info_in.read_idx = up_to;
        hsa_group_complete( signal_switch, sync_1, sync_2 );
    }
}
