#include "../../hsa_helper.h"

void populate_queue_info( const queue_def *q, queue_info *info ) {
        if ( lane_id() == 0 ) {
            info->read_idx  = get_queue_read_idx(q);
            info->write_idx = get_queue_write_idx(q);
        }

        // Broadcast
        info->read_idx  = broadcast_u64(info->read_idx,0);
        info->write_idx = broadcast_u64(info->write_idx,0);
}

void wait_for_elements( const queue_def *q, queue_info *info, const hsa_signal_value_t num_elements ) {
        const hsa_signal_value_t min_write_idx = info->read_idx + num_elements;
        if ( info->write_idx < min_write_idx ) {
            if ( lane_id() == 0 ) {
                info->write_idx = hsa_signal_wait(q->w_signal, HSA_SIGNAL_CONDITION_GTE, min_write_idx, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);
            }

            // Broadcast
            info->write_idx = broadcast_u64(info->write_idx,0);
        }
}

void wait_for_space( const queue_def *q, queue_info *info, const hsa_signal_value_t num_elements  ) {
    if ( active_lane_id() == 0 ) {
        const hsa_signal_value_t min_read_idx = info->write_idx - q->capacity + num_elements;
        if ( info->read_idx < min_read_idx ) {
            info->read_idx = hsa_signal_wait(q->r_signal, HSA_SIGNAL_CONDITION_GTE, min_read_idx, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);
        }
    }
}

__kernel void hsa_filter_wf( __constant queue_def *in_q, __constant queue_def *out_q, long min_batch_size, __global long *dbg ) {

    const long  wg_size = get_local_size(0);
    bool term = false;

    // Inititalize queue-info
    const queue_def  q_in  = *in_q;
    const queue_def  q_out = *out_q;
    queue_info info_in;
    queue_info info_out;

    T element;
    bool check = false;
    long num_elements;
    long num_results;

    // Get current queue states
    populate_queue_info(&q_in, &info_in);
    populate_queue_info(&q_out, &info_out);

    while ( !term ) {
        wait_for_elements( &q_in, &info_in, min_batch_size );

        // Terminate
        if ( info_in.write_idx == QUEUE_TERM_IDX ) {
            term = true;
        }
        // Process
        else {
            num_elements = min(info_in.write_idx-info_in.read_idx, wg_size);
            check = false;

            // Process only available elements
            if (  lane_id() < num_elements ) {
                // Get event
                const hsa_signal_value_t my_read_idx = info_in.read_idx + lane_id();
                element = *((T*) get_queue_element(&q_in, my_read_idx, sizeof(T)));

                // Check condition
                check = filter_predicate(&element);
            }

            // To update out_q write_idx for all work-items
            num_results = lane_count(check);

            if ( num_results > 0 ) {
                // Wait for space in out-queue
                wait_for_space(&q_out, &info_out, num_results);

                // Write event
                if ( check ) {
                    T *out = ((T*) get_queue_element(&q_out, info_out.write_idx + active_lane_id(), sizeof(T)));
                    do_write(out,&element);
                }

                // Update indexes
                info_out.write_idx += num_results;
                // Notify queues
                if ( lane_id() == 0 ) {
                    add_to_queue_write_idx( &q_out, num_results );
                }
            }
            barrier(0);

            // Update indexes
            info_in.read_idx += num_elements;

            // Notify queues
            if ( lane_id() == 0 ) {
                add_to_queue_read_idx( &q_in, num_elements );
            }
        }
    }
}
