#include "../../hsa_helper.h"

void populate_queue_info( const queue_def *q, queue_info *info ) {
        info->read_idx  = get_queue_read_idx(q);
        info->write_idx = get_queue_write_idx(q);
}

bool wait_for_elements( const queue_def *q, queue_info *info, const hsa_signal_value_t num_elements ) {
        const hsa_signal_value_t min_write_idx = info->read_idx + num_elements;
        if ( info->write_idx < min_write_idx ) {
            info->write_idx = hsa_signal_wait(q->w_signal, HSA_SIGNAL_CONDITION_GTE, min_write_idx, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);
        }
        return info->write_idx == QUEUE_TERM_IDX;
}

__kernel void hsa_join_switch(
        __constant queue_def *in_q1,
        __constant queue_def *in_q2,
        __global window_def * w_1,
        __global window_def * w_2,
        hsa_signal_t sync_1,
        hsa_signal_t sync_2,
        __global long *ctrl,
        long num_groups,
        __global ulong *latency_accum,
        __global long *dbg
) {

    bool term = false;
    bool signal_switch = false;

    // Inititalize queue-info
    const queue_def  q_left  = *in_q1;
    const queue_def  q_right = *in_q2;

    queue_info info_left;
    queue_info info_right;
    
    window_def window_left  = *w_1;
    window_def window_right = *w_2;

    T1 left;
    T2 right;

    ulong local_latency = 0;
    ulong last_begin = 0;
    ulong num_turns = 0;

    // Get current queue states
    populate_queue_info(&q_left, &info_left);
    populate_queue_info(&q_right, &info_right);


    // Wait for one event at every queue
    term = wait_for_elements( &q_left, &info_left, 1 );
    term = term || wait_for_elements( &q_right, &info_right, 1 );

    if ( !term ) {
        left  = *((T1*) get_queue_element(&q_left, info_left.read_idx, sizeof(T1)));
        right = *((T2*) get_queue_element(&q_right, info_right.read_idx, sizeof(T2)));
    }
    
    while ( !term ) {
        signal_switch = !signal_switch;

        hsa_coord_wait(signal_switch, sync_1, sync_2, num_groups);

        const ulong now = timestamp();
        if ( last_begin != 0 ) {
            num_turns++;
            local_latency += (now-last_begin);
        }
        last_begin = now;
        
       if ( left.timestamp <= right.timestamp) {
            // Store event in window
            T1 *window_address = get_window_element( &window_left, window_left.num_inserts, sizeof(T1) );
            *window_address = left;
            
            // Update window state
            window_left.num_inserts++;
            
            // Set control flag (left side processing)
            *ctrl = window_left.num_inserts;
            
            // Tell event was consumed
            add_to_queue_read_idx( &q_left, 1);
            info_left.read_idx++;
            
            
            hsa_coord_fire(signal_switch, sync_1, sync_2);
            
            // Load new left event
            term = wait_for_elements( &q_left, &info_left, 1 );
            if ( !term ) {
                left  = *((T1*) get_queue_element(&q_left, info_left.read_idx, sizeof(T1)));
            }
        }
        else {
            // Store event in window
            T2 *window_address = get_window_element( &window_right, window_right.num_inserts, sizeof(T2) );
            *window_address = right;
            
            // Update window state
            window_right.num_inserts++;
            
            // Set control flag (left side processing)
            *ctrl = -(window_right.num_inserts);
            
            // Tell event was consumed
            add_to_queue_read_idx( &q_right, 1);
            info_right.read_idx++;
            
            // Tell processors to begin
            hsa_coord_fire(signal_switch, sync_1, sync_2);
            
            // Load new left event
            term = wait_for_elements( &q_right, &info_right, 1 );
            if ( !term ) {
                right  = *((T2*) get_queue_element(&q_right, info_right.read_idx, sizeof(T2)));
            }
        }
    }

    // Wait for processors to settle and signal termination
    hsa_coord_wait(!signal_switch, sync_1, sync_2, num_groups);
    hsa_signal_store(sync_1, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);
    hsa_signal_store(sync_2, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);

    // Write back window state
    *w_1 = window_left;
    *w_2 = window_right;
    *latency_accum = (num_turns > 0) ? (local_latency/num_turns) * DEVICE_TIMER_FACTOR : 0;
}

inline void write_wave( __private const T1 *l, __private const T2 *r, __global long *w_idx, __private const queue_def *q, __private queue_info *info ) {
    const bool passed = join_predicate(l,r);
    const uint num_results_wave = lane_count(passed);

    if ( num_results_wave > 0 ) {

        hsa_signal_value_t write_idx = -1;

        // Reserve space
        if ( active_lane_id() == 0 ) {
            write_idx = atom_add(w_idx, num_results_wave);
            
            // Wait for space to be available
            const hsa_signal_value_t min_read_idx = (write_idx+num_results_wave-q->capacity);

            if ( info->read_idx < min_read_idx )
                info->read_idx = hsa_signal_wait(q->r_signal, HSA_SIGNAL_CONDITION_GTE, min_read_idx, QUEUE_TERM_IDX, HSA_MEMORY_ORDER_SEQ_CST);
        }
        write_idx      = broadcast_u64(write_idx,0);
//        info->read_idx = broadcast_u64(info->read_idx,0);

        // Write results
        if ( passed ) {
            TOUT *out = get_queue_element( q, write_idx+active_lane_id(), sizeof(TOUT) );
            do_write(out,l,r);
        }

        // Update out-queue write index
        if ( active_lane_id() == 0 ) {
            update_queue_write_idx(q, write_idx, write_idx + num_results_wave);
        }
    }
}


__kernel void hsa_join_proc(
__constant window_def *w_1,
__constant window_def *w_2,
__constant queue_def  *out_q,
hsa_signal_t sync_1,
hsa_signal_t sync_2,
__global int64_t *ctrl,
__global int64_t *w_idx,
__global int64_t *dbg
) {
    __local hsa_signal_value_t bcast[1];
    const size_t global_id = get_global_linear_id();
    const size_t grid_size = get_global_linear_size();
    
    const queue_def   q_out        = *out_q;
    queue_info  q_info;
    window_def window_left  = *w_1;
    window_def window_right = *w_2;
    
    bool signal_switch = false;
    bool term          = false;

    populate_queue_info(&q_out, &q_info);

    // Signal steady state
    hsa_group_ready(sync_1,sync_2);

    while ( !term ) {
        signal_switch = !signal_switch;

        term = hsa_group_wait( signal_switch, sync_1, sync_2, bcast );
        // Terminate?
        if ( term ) {
            continue;
        }

        const long which = *ctrl;
        T1 left;
        T2 right;

        // New left event
        if ( which >= 0 ) {
            left = *((T1*) get_window_element(&window_left, which-1, sizeof(T1)));
            // Update local window information -- DO NOT REMOVE
            window_left.num_inserts++;

            for ( size_t i = global_id; i < min(window_right.num_inserts,window_right.size); i+=grid_size ) {
                // Check if in range
                right = *((T2*) get_window_element(&window_right, i, sizeof(T2)));
                write_wave( &left, &right, w_idx, &q_out, &q_info);
            }
        }
        // New right event
        else {
            right = *((T2*) get_window_element(&window_right, -(which+1), sizeof(T2)));
            window_right.num_inserts++;

            for ( size_t i = global_id; i < min(window_left.num_inserts,window_left.size); i+=grid_size ) {
                // Check if in range
                left = *((T1*) get_window_element(&window_left, i, sizeof(T1)));
                write_wave( &left, &right, w_idx, &q_out, &q_info);
            }
        }
        hsa_group_complete( signal_switch, sync_1, sync_2 );
    }
}
