//
// Created by koerberm on 24.05.18.
//

#include <random>
#include <bitset>
#include <thread>
#include <vector>
#include "hsa_platform/hsa_platform.h"
#include "util/cl_util.h"
#include "util/eval_util.h"

void signal_test(const hsa::hsa_platform &platform, const hsa::hsa_agent &agent, const hsa::hsa_executable &exe) {
    auto start = platform.create_signal(2);
    auto end   = platform.create_signal(0);
    auto completion_signal = platform.create_signal(1);
    auto queue = agent.create_queue(agent.queue_min_size,HSA_QUEUE_TYPE_SINGLE,UINT32_MAX,UINT32_MAX);

    completion_signal->store_relaxed(1);

    uint64_t value = 0;

    // Args
    struct args_t {
        uint64_t * val;
        hsa_signal_t handle_in;
        hsa_signal_t handle_out;
    } args;

    args.val = &value;
    args.handle_in = start->handle;
    args.handle_out = end->handle;

    auto kernel_symbol = exe.kernel_by_name("signal_test_kernel");
    auto kernarg_region = agent.get_kernarg_region();
    auto kernargs = kernarg_region.allocate(kernel_symbol.kernarg_segment_size,kernel_symbol.kernarg_segment_alignment);

    memcpy(kernargs->address<char>(), &args, sizeof(args));

    int sleep = 2;

    // Dispatch
    hsa::hsa_kernel_dispatch_packet p(1,kernel_symbol,*completion_signal,*kernargs);
    p.set_grid_size(1);
    p.set_workgroup_size(1);

    std::cout << "Signal wait iterations: " << value << std::endl;
    std::cout << "Dispatching kernel" << std::endl;
    queue->dispatch_kernel_packet(p);

    std::cout << "Waiting " << sleep << " seconds" << std::endl;
    std::this_thread::sleep_for(std::chrono::seconds(sleep));

    std::cout << "Triggering signal" << std::endl;
    start->store_relaxed(1);

    std::cout << "Waiting for kernel termination" << std::endl;
    completion_signal->wait_relaxed(HSA_SIGNAL_CONDITION_LT, 1);

    std::cout << "Signal wait iterations: " << value << std::endl;
}

void signal_loop_test(const hsa::hsa_platform &platform, const hsa::hsa_agent &agent, const hsa::hsa_executable &exe,
                      uint16_t num_groups, uint64_t num_loops, std::vector<double> &results) {
    auto start = platform.create_signal(0);
    auto end   = platform.create_signal(0);
    auto completion_signal = platform.create_signal(1);
    auto queue = agent.create_queue(agent.queue_max_size,HSA_QUEUE_TYPE_SINGLE,UINT32_MAX,UINT32_MAX);

    completion_signal->store_relaxed(1);

    uint32_t group_size = agent.workgroup_max_size;

    // Args
    struct args_t {
        hsa_signal_t handle_in;
        hsa_signal_t handle_out;
        uint64_t num_loops;
    } args;

    args.handle_in = start->handle;
    args.handle_out = end->handle;
    args.num_loops = num_loops;

    auto kernel_symbol = exe.kernel_by_name("signal_loop_test_kernel");
    auto kernarg_region = agent.get_kernarg_region();
    auto kernargs = kernarg_region.allocate(kernel_symbol.kernarg_segment_size,kernel_symbol.kernarg_segment_alignment);

    memcpy(kernargs->address<char>(), &args, sizeof(args));

    start->store_relaxed(num_groups);
    end->store_relaxed(0);

    // Dispatch
    hsa::hsa_kernel_dispatch_packet p(1,kernel_symbol,*completion_signal,*kernargs);
    p.set_grid_size(group_size*num_groups);
    p.set_workgroup_size(group_size);
//    p.set_grid_size(num_groups);
//    p.set_workgroup_size(1);

    queue->dispatch_kernel_packet(p);

    auto begin= std::chrono::steady_clock::now();
    bool failed = false;
    for ( uint64_t i = 0; i < num_loops && !failed; i++ ) {
        if ( i % 2 == 0 ) {
            start->store_relaxed(0);
            if ( end->wait_relaxed(HSA_SIGNAL_CONDITION_GTE, num_groups, 1000000000 ) < num_groups )
                failed = true;
        }
        else {
            end->store_relaxed(0);
            if ( start->wait_relaxed(HSA_SIGNAL_CONDITION_GTE, num_groups, 1000000000 ) < num_groups )
                failed = true;
        }
//        std::cout << "Signal values (" << start->load_scacquire() << ", " << end->load_scacquire() << ")" << std::endl;
    }
    completion_signal->wait_relaxed(HSA_SIGNAL_CONDITION_LT, 1, 1000000000);
    std::cout << " " << (failed? "FAILED" : "SUCCESS") << std::flush;

    std::chrono::duration<double> last=  std::chrono::steady_clock::now() - begin;
    results.push_back(last.count());
    //std::cout << "time: " << last.count() << std::endl;
}

void dispatch_loop_test(const hsa::hsa_platform &platform, const hsa::hsa_agent &agent, const hsa::hsa_executable &exe, uint16_t num_groups, uint64_t num_loops, std::vector<double> &results) {
    auto completion_signal = platform.create_signal(0);
    auto queue = agent.create_queue(agent.queue_max_size,HSA_QUEUE_TYPE_SINGLE,UINT32_MAX,UINT32_MAX);

    uint32_t group_size = agent.workgroup_max_size;

    auto kernel_symbol = exe.kernel_by_name("noop_kernel");
    auto kernarg_region = agent.get_kernarg_region();
    auto kernargs = kernarg_region.allocate(kernel_symbol.kernarg_segment_size,kernel_symbol.kernarg_segment_alignment);

    auto begin= std::chrono::steady_clock::now();
    for ( uint64_t i = 0; i < num_loops; i++ ) {
        hsa::hsa_kernel_dispatch_packet p(1,kernel_symbol,*completion_signal,*kernargs);
        p.set_grid_size(group_size*num_groups);
        p.set_workgroup_size(group_size);
        completion_signal->store_relaxed(1);
        queue->dispatch_kernel_packet(p);
        completion_signal->wait_relaxed(HSA_SIGNAL_CONDITION_LT, 1);
    }
    std::chrono::duration<double> last=  std::chrono::steady_clock::now() - begin;
    results.push_back(last.count());
    //std::cout << "time: " << last.count() << std::endl;
}

void ocl_loop_test(cl::Device &device, uint16_t num_groups, uint64_t num_loops, std::vector<double> &results) {

    cl::Context context(device);

    auto group_size = device.getInfo<CL_DEVICE_MAX_WORK_GROUP_SIZE>();

    // Create command queue.
    cl::CommandQueue queue(context, device);

    auto program = cl_util::program_from_file("src/rocm/cl/testing/noop.cl", context);

    cl_util::build_program(program, device);

    cl::Kernel kernel(program, "noop_kernel");
    cl::Event event;
    auto begin = std::chrono::steady_clock::now();
    for (uint64_t i = 0; i < num_loops; i++) {
        queue.enqueueNDRangeKernel(kernel, cl::NullRange, group_size * num_groups, group_size, NULL, &event);
        event.wait();
    }
    std::chrono::duration<double> last = std::chrono::steady_clock::now() - begin;
    results.push_back(last.count());
}

int main(int argc, char **argv) {
    (void) argc;
    (void) argv;

    hsa::hsa_platform platform;

    for ( auto &agent : platform.get_agents() ) {
        if ( agent.features != HSA_AGENT_FEATURE_KERNEL_DISPATCH )
            continue;

        std::cout << "Using agent " << agent.name << std::endl;

        hsa::hsa_executable exe;

        std::vector<std::string> modules = {
            "src/rocm/cl/out/noop.co",
            "src/rocm/cl/out/signal.co",
            "src/rocm/cl/out/signal_loop.co"
        };

        for ( auto &file : modules ) {
            try {
                auto co = hsa::code_object_from_file(file,agent);
                exe.load_code_object(co,agent);
            } catch ( const hsa::hsa_exception &e ) {
                std::cout << "Error loading module (" << file << "): " << e.what() << std::endl;
                return 1;
            }
        }
        exe.freeze();

        auto q = agent.create_queue(agent.queue_min_size,HSA_QUEUE_TYPE_SINGLE);

//        std::cout << "===========================================" << std::endl;
//
//        std::cout << "Testing signal for busy waiting..." << std::endl;
//        signal_test(platform, agent, exe);


        std::cout << "===========================================" << std::endl;

        std::cout << "Testing signal synchronization..." << std::endl;

        uint64_t num_loops = 1000000;

        std::vector<double> results_ocl_dispatch;
        std::vector<double> results_hsa_dispatch;
        std::vector<double> results_hsa_signal;

        std::vector<cl::Device> devices;
        cl_util::get_gpu_devices(devices);

        auto &cl_device = devices[1];

//        cl_platform_id platforms[10];
//        cl_uint num_platforms;
//
//        clGetPlatformIDs(10,platforms,&num_platforms);
//
//        for ( uint i = 0; i < num_platforms; i++ ) {
//            cl_device_id devices2[10];
//            cl_uint num_devices;
//
//            clGetDeviceIDs(platforms[i], CL_DEVICE_TYPE_GPU, 10, devices2,&num_devices);
//
//            for ( int j = 0; j < num_devices; j++ ) {
//                char name[256];
//                size_t name_length;
//                cl_int error = clGetDeviceInfo(devices2[j],CL_DEVICE_BOARD_NAME_AMD,256,name,&name_length);
//                name[name_length] = '\0';
//                std::cout << "Found board: " << name << std::endl;
//            }
//
//        }


        std::cout << "CL-Device: " <<  cl_device.getInfo<CL_DEVICE_NAME>() << std::endl;

        for (uint16_t groups = 1; groups <= 16; groups++) {
            std::cout << "#Groups:" << groups << std::flush;
            dispatch_loop_test(platform, agent, exe, groups, num_loops, results_hsa_dispatch);
            signal_loop_test(platform, agent, exe, groups, num_loops, results_hsa_signal);
            ocl_loop_test(cl_device,groups,num_loops, results_ocl_dispatch);
            std::cout  << std::endl;
        }
        std::cout << "Timings OpenCL dispatch: " << results_ocl_dispatch << std::endl;
        std::cout << "Timings HSA dispatch   : " << results_hsa_dispatch << std::endl;
        std::cout << "Timings HSA signal     : " << results_hsa_signal << std::endl;

        std::cout << "==========================================================" << std::endl;
    }
    return 0;
}